
/*
 cc -g -c ../as/asqueue.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#include "../s_tools/smem.h"
#include "../s_tools/stcp.h"

#include "../as/asevent.h"
#include "../as/asinput.h"
#include "../as/asqueue.h"


/* ------------------------------- AsqueuE --------------------------- */


int Asqueue_new(queue,agent,type,progadr,flag)
struct AsqueuE **queue;
struct AsagenT *agent;
int type;
char *progadr;
int flag;
{
 struct AsqueuE *q;
 int ret;

 *queue= q= TSOB_FELD(struct AsqueuE,1);
 if(q==NULL)
   return(-1);
 q->agent= agent;
 q->count= 0;
 q->fetch_loop_limit= Asqueue_fetch_loop_limiT;
 q->jam_limit= Asqueue_jam_limiT;
 q->no_waiting= 0;
 q->first= NULL;
 q->last= NULL;
 q->input= NULL;
 ret= Asinput_new(&(q->input),q,type,progadr,0);
 if(ret<=0){
   Asqueue_destroy(queue,0);
   return(-1);
 }
 return(1);
}


int Asqueue_destroy(queue,flag)
struct AsqueuE **queue;
int flag;
{

 if(*queue==NULL)
   return(0);
 Asqueue_cancel(*queue,0);
 if((*queue)->input!=NULL)
   Asinput_destroy(&((*queue)->input),0);
 Smem_freE((char *) *queue);
 *queue= NULL;
 return(1);
}


int Asqueue_default_kbd(queue,flag)
struct AsqueuE *queue;
int flag;
{
 int ret;

 ret= Asinput_default_kbd(queue->input,flag);
 return(ret);
}


int Asqueue_count_services(queue,flag)
struct AsqueuE *queue;
int flag;
/*
 bit0= count internal input as a service
*/
{
 int ret;

 ret= Asinput_count_services(queue->input,flag&1);
 return(ret);
}


int Asqueue_append(queue,in_event,flag)
struct AsqueuE *queue;
struct AsevenT **in_event;                 /* *in_event may be set to NULL */
int flag;
/*
 bit0= insert at head of queue rather than at tail
*/
{
 struct AsevenT *event;

 event= *in_event;
 if(event->type==1) { /* control event */
   if(strcmp(event->command,"CANCEL")==0)   
     Asqueue_cancel(queue,0);
   Asevent_destroy(in_event,0);
   return(2);
 }
 event->prev= event->next= NULL;
 if(queue->first==NULL) {
   queue->first= event;
   queue->last= event;
 } else if(flag&1){
   event->next= queue->first;
   queue->first->prev= event;
   queue->first= event;
 } else {
   event->prev= queue->last;
   queue->last->next= event;
   queue->last= event;
 }
 (queue->count)++;
 return(1);
}


int Asqueue_event_is_pending(queue,limit_for_3,flag)
struct AsqueuE *queue;
int limit_for_3;
int flag;
/*
flag:
 bit0= do not update the queue before testing
 bit1= reply 3 if there are at least limit_for_3 events pending
return:
  <0 = error
   0 = no event pending
   1 = event pending
   2 = quite urgent token has arrived
   3 = at least limit_for_3 events pending (with flag bit1)
   5 = immediate end of event loop demanded
*/
{
 int ret,count= 0;
 struct AsevenT *event;

 if(!(flag&1)) {
   ret= Asqueue_update(queue,1);
   if(ret<=0)
     return(ret);
   if(ret==2 || ret==5)
     return(ret);
 }
 if(limit_for_3>0 && (flag&2)) {
   for(event= queue->first;event!=NULL;event= event->next) {
     count++;
     if(count>=limit_for_3)
       return(3);
   }
 }
 return(queue->first!=NULL);
}


int Asqueue_advance(queue,event,flag)
struct AsqueuE *queue;
struct AsevenT **event;
int flag;
/*
flag:
 bit0= do not update the queue before advancing
 bit1= quick update, do not wait
return:
  <0 = error
   0 = no event pending
   1 = ok, event taken from queue
   5 = immediate end of event loop demanded
*/
{
 int ret;

 if(!(flag&1)) {
   ret= Asqueue_update(queue,(flag&2)>>1);
   if(ret<=0)
     return(ret);
   if(ret==2) {
     /* >>> how to handle quite urgent events ? */;
   } 
   if(ret==5)
     return(ret);
 }
 Asevent_destroy(event,0);
 *event= queue->first;
 if((*event)==NULL)
   return(0);
 queue->first= (*event)->next;
 if(queue->first!=NULL) {
   queue->first->prev= NULL;
 } else {
   queue->last= NULL;
 }
 (queue->count)--;
 (*event)->prev= NULL;
 (*event)->next= NULL;
 return(1);
}


int Asqueue_cancel(queue,flag)
struct AsqueuE *queue;
int flag;
{
 struct AsevenT *event,*next;
 event= queue->first;
 while(event!=NULL) {
   next= event->next;
   Asevent_destroy(&event,0);
   event= next;
 }
 queue->first= queue->last= NULL;
 queue->count= 0;
 return(1);
}


int Asqueue_input_to_events(queue,flag)
struct AsqueuE *queue;
int flag;
/*
 bit0= insert as first event of the queue rather than as last one
*/
{
 int ret,i,start,stop,step;
 struct AsevenT *event;
 char *replyline= NULL;

 if(queue->input->reply_type==1) {
   ret= Asinput_get_replyline(queue->input,&replyline,0);
   if(ret<=0)
     goto ex;
   if(strcmp(replyline,"CANCEL")==0) {
     Asqueue_cancel(queue,0);
     /* to notify possible servers, a normal event is generated here */
     ret= Asinput_set_replyline(queue->input,"-queue_canceled",0);
     if(ret<=0)
       goto ex;
     queue->input->reply_type= 0;
   } else {
     Asagent_stderr(queue->agent,"Program error: unknown control event: ",0);
     Asagent_stderr(queue->agent,replyline,1);
     ret= 0;
     goto ex;
   }
 }

 if(flag&1) {
   start= queue->input->reply_count-1;
   stop= -1;
   step= -1;
 } else {
   start= 0;
   stop= queue->input->reply_count;
   step= 1;
 }
 for(i=start;i!=stop;i+= step){
   ret= Asevent_new(&event,queue->input->reply[i],
                     queue->input->from_connection,queue->input->username,
                     (queue->input->reply_type==1));
   if(ret<=0)
     goto ex;
   ret= Asqueue_append(queue,&event,flag&1);
   if(ret<=0)
     goto ex;
 }

 ret= 1;
ex:
 Sregex_string(&replyline,NULL,0);
 return(ret);
}


int Asqueue_update(queue,flag)
struct AsqueuE *queue;
int flag;
/*
flag:
 bit0= this is not part of an interactive call. do not wait
return:
 <=0 = error
   1 = ok
   2 = quite urgent token arrived
   5 = immediate end of event loop demanded
*/
{
 int ret,followup,quite_urgent_event= 0,queue_jam= 0,i;

 followup= (flag&1)||(queue->no_waiting)||Asqueue_event_is_pending(queue,0,1);
 for(i=0;i<queue->fetch_loop_limit;i++){
   queue_jam= (queue->count>=queue->jam_limit)<<6;
   ret= Asinput_fetch_token(queue->input,(!!followup)|queue_jam);
   queue->no_waiting= 0;
   followup= 1;
   if(ret<0)
     goto ex;
   if(ret==0)
 break;
   else if(ret==2)
    quite_urgent_event= 1;
   else if(ret==3) {
     queue->no_waiting= 1;
 break;
   } else if(ret==5) {
     /* demand immediate end of event loop */
     goto ex;
   }
   ret= Asqueue_input_to_events(queue,0);
   if(ret<0)
     goto ex;
   if(quite_urgent_event)
 break;
 }
 ret= 1+quite_urgent_event;
ex:
 return(ret);
}


int Asqueue_synthetic(queue,text,from_connection,username,flag)
struct AsqueuE *queue;
char *text;
int from_connection;
char *username;
int flag;
/*
 bit0= non ESC : use raw_reply without interpretation
 bit1= ESC : use raw_reply without interpretation
 bit2= insert as first event of the queue rather than as last one
*/
{
 int ret;

 queue->input->from_connection= from_connection;
 if(Sregex_string(&(queue->input->username),username,0)<=0)
   {ret= -1; goto ex;}
 ret= Asinput_set_reply(queue->input,text,flag&3);
 if(ret<=0)
   goto ex;
 ret= Asqueue_input_to_events(queue,(flag>>2)&1);
 if(ret<0)
   goto ex;
 ret= 1;
ex:
 return(ret);
}


int Asqueue_fetch_token(queue,reply,flag)
struct AsqueuE *queue;
char **reply;
int flag;
/*
 bit0= set blocking mode according to bit1 (halfdelay else, if possible)
 bit1= blocking, non blocking else
 bit2= do not convert reply character by prog
 bit3= this looks for requested line input
 bit4= only look for modes 3 and 4 (stdout,stderr pipes)

return:
 -1 = fatal error
  0 = no token read
  1 = token read
*/
{
 int ret;
 ret= Asinput_fetch_token(queue->input,(flag&31)|32); /* no conn mode 5 or 6 */
 if(ret<=0)
   goto ex;
 ret= Asinput_get_replyline(queue->input,reply,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Asqueue_fetch_line(queue,reply,flag)
struct AsqueuE *queue;
char **reply;
int flag;
{
 int ret;
 
 ret= Asinput_fetch_line(queue->input,0);
 if(ret<=0)
   goto ex;
 ret= Asinput_get_replyline(queue->input,reply,0);
 if(ret<=0)
   goto ex;
ex:;
 return(ret);
}


int Asqueue_start_tcpserv(queue,servername,portnumber,external_s_mask,
                           type,flag)
struct AsqueuE *queue;
char *servername;
int portnumber;
int external_s_mask;
int flag;
{
 int ret;

 ret= Asinput_start_tcpserv(queue->input,servername,portnumber,external_s_mask,
                           type,flag);
 return(ret);
}


int Asqueue_get_tcpserv(queue,type,serv,flag)
struct AsqueuE *queue;
int type;
struct StcpserV **serv;
int flag;
{
 if(type==1)
   *serv= queue->input->char_serv;
 else if(type==2)
   *serv= queue->input->line_serv;
 else {
   *serv= NULL;
   return(-1);
 }
 return(1);
}


int Asqueue_new_inconn(queue,type,filename,fd,keep_pipe_alive_fd,flag)
struct AsqueuE *queue;
int type;
char *filename;
int fd;
int keep_pipe_alive_fd;
int flag;
{
 int ret;

 ret= Asinput_new_inconn(queue->input,type,filename,fd,keep_pipe_alive_fd,
                          flag);
 return(ret);
}


int Asqueue_sconn_to_inconn(queue,type,sconn,
                            app_handle,App_input_handler,app_reaction,flag)
struct AsqueuE *queue;
int type;
struct SconN *sconn;
char *app_handle;
int (*App_input_handler)();
int app_reaction;
int flag;
{
 int ret;

 ret= Asinput_sconn_to_inconn(queue->input,type,sconn,
                              app_handle,App_input_handler,app_reaction,flag);
 return(ret);
}


int Asqueue_destroy_inconn(queue,fd,flag)
struct AsqueuE *queue;
int fd;                   /* -2 = all */
int flag;
/*
 bit0= (together with fd= -2) also destroy first (dummy) connection of input
*/
{
 int ret;

 ret= Asinput_destroy_connection(queue->input,fd,flag&1);
 return(ret);
}

 
int Asqueue_find_sconn(queue,sconn,id,fd,flag)
struct AsqueuE *queue;
struct SconN **sconn;
char *id;
int fd;
int flag;
{
 int ret= 0;

 ret= Asinput_find_sconn(queue->input,sconn,id,fd,flag);
 return(ret);
}


int Asqueue_set_internal(queue,value,flag)
struct AsqueuE *queue;
int value;
int flag;
{
 int ret;

 ret= Asinput_set_internal(queue->input,value,0);
 return(ret);
}


int Asqueue_set_bundle_endmark(queue,fd,text,flag)
struct AsqueuE *queue;
int fd;
char *text;
int flag;
{
 int ret;

 ret= Asinput_set_bundle_endmark(queue->input,fd,text,0);
 return(ret);
}


int Asqueue_remove_fd_events(queue,fd,flag)
struct AsqueuE *queue;
int fd;
int flag;
/*
 bit0= do not remove but mark as connection orphans
*/
{
 struct AsevenT *current,*prev,*next;
 int ret;
 
 for(current= queue->first; current!=NULL; current= next) {
   prev= current->prev;
   next= current->next;
   if(current->connection == fd) {
     if(flag&1) {
       current->connection= -2;
     } else {
       if(current==queue->first)
         queue->first= next;
       if(current==queue->last)
         queue->last= prev;
       Asevent_destroy(&current,1);
       queue->count--;
     }
   }
 }
 
 ret= 1;
ex:
 return(ret);
}
