
/*
 cc -g -c imgdb.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgsamp.h"
#include "imgexempt.h"

#include "imgdb.h"


/* -------------------------------- ImgdB ------------------------------ */


int Imgdb_new(imdb,imp,mapname,spoints,flag)
struct ImgdB **imdb;
struct ImgmaP *imp;
char *mapname;
int spoints;
int flag;
{
 int ret;
 struct ImgdB *m= NULL;

 *imdb= m= TSOB_FELD(struct ImgdB,1);
 if(*imdb==NULL)
   return(-1);
 m->imp= imp;
 strcpy(m->mapname,mapname);
 m->spoints= spoints;
 m->write_dbs= 0;
 m->read_dbs= 0;
 m->imdbf= NULL;
 m->imdbs= NULL;
 m->errmsg[0]= 0;
 m->query_token= NULL;
 m->debug_mode= 2;

 ret= Imgdbf_new(&(m->imdbf),m,0);
 if(ret<=0)
   goto failed;
 m->write_dbs= 1;
 m->read_dbs= 1;

 ret= Imgdb_read_conf(m,0);
 if(ret<=0)
   goto failed;

 return(1);
failed:;
 if(m!=NULL)
   if(m->errmsg[0]!=0)
     Imgmap_set_errmsg(imp,m->errmsg,0);
 Imgdb_destroy(imdb,0);
 return(-1);
}


int Imgdb_destroy(imdb,flag)
struct ImgdB **imdb;
int flag;
{
 struct ImgdB *m;
 
 m= *imdb;
 if(m==NULL)
   return(0);
 if(m->query_token!=NULL)
   Imgdb_end_query(m,0);
 Imgdbf_destroy(&(m->imdbf),0);
 Imgdbs_destroy(&(m->imdbs),0);
 Smem_freE((char *) m);
 *imdb= NULL;
 return(1);
}


int Imgdb_copy(source,target,flag)
struct ImgdB *source,*target;
int flag;
{
 int ret;

 strcpy(target->mapname,source->mapname);
 target->spoints= source->spoints;
 target->write_dbs= source->write_dbs;
 target->read_dbs= source->read_dbs;
 ret= Imgdbf_copy(source->imdbf,target->imdbf,0);
 if(ret<=0)
   return(ret);
 if(source->imdbs!=NULL && target->imdbs!=NULL) {
   ret= Imgdbs_copy(source->imdbs,target->imdbs,0);
   if(ret<=0)
     return(ret);
 }
 target->debug_mode= source->debug_mode;
 return(1);
}


int Imgdb_get_spoints(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 return(imdb->spoints);
}


int Imgdb__make_mapname(raw_mapname,spoints,mapname,flag)
char *raw_mapname;
int spoints;
char *mapname;
int flag;
/*
 bit0-7: mode
         0= raw mapname 
         1= name of list of file names
         2= name of list of samples 
         3= name of -list_doubles cursor file
         4= name of break file
         5= name of file with server address (pubno)
         6= name of exempt list file
         7= name of db configuration file
         8= name of SQL password file
*/
{
 int modus;

 modus= flag&255;
 if(modus==0) {
   strcpy(mapname,raw_mapname);
 } else if (modus==1) {
   sprintf(mapname,"%s_nam",raw_mapname);
 } else if (modus==2) {
   sprintf(mapname,"%s_smp%ld",raw_mapname,spoints);
 } else if (modus==3) {
   sprintf(mapname,"%s_list_doubles_pos",raw_mapname);
 } else if (modus==4) {
   sprintf(mapname,"%s_break",raw_mapname);
 } else if (modus==5) {
   sprintf(mapname,"%s_pubno",raw_mapname);
 } else if (modus==6) {
   sprintf(mapname,"%s_exempt",raw_mapname);
 } else if (modus==7) {
   sprintf(mapname,"%s_conf",raw_mapname);
 } else if (modus==8) {
   sprintf(mapname,"%s_sql_pwd",raw_mapname);
 } else {
   return(0);
 }
 /* always consider Imgdbf_garbage_coll() when extending this list */
 return(1);
}


int Imgdb_get_mapname(imdb,mapname,flag)
struct ImgdB *imdb;
char *mapname;
int flag;
/*
 bit0-7: mode see Imgdb__make_mapname()
*/
{
 int ret;

 ret= Imgdb__make_mapname(imdb->mapname,imdb->spoints,mapname,flag&255);
 return(ret);
}


int Imgdb__mapname_highcode(flag)
int flag;
{
 return(8);
}


int Imgdb_set_dbs(imdb,text,flag)
struct ImgdB *imdb;
char *text;
int flag;
/*
 bit0= read_dbs (else write_dbs)
*/
{
 int disable= 0,*dbs;

 if(flag&1) {
   dbs= &(imdb->read_dbs);
   *dbs= 0;
 } else {
   dbs= &(imdb->write_dbs);
 }
 if(text[0]=='-')
   disable= 1;
 else if(text[0]!='+')
   *dbs= 0;
 if(strstr(text,":file:")!=NULL) {
   if(disable)
    (*dbs)&= ~1;
   else
    (*dbs)|= 1;
 }
 if(strstr(text,":sql:")!=NULL) {
   if(disable)
    (*dbs)&= ~2;
   else {
    if(imdb->imdbs==NULL) {

      /* >>> create new ImgdbS or refuse */;
      return(0);

    }
    (*dbs)|= 2;
   }
 }
 return(1);
}


int Imgdb_get_dbs(imdb,text,flag)
struct ImgdB *imdb;
char text[320];
int flag;
/*
 bit0= read_dbs (else write_dbs)
*/
{
 int dbs;

 if(flag&1)
   dbs= imdb->read_dbs;
 else 
   dbs= imdb->write_dbs;
 text[1]= ':';
 text[1]= 0;
 if(dbs&1)
   strcat(text,"file:");
 if(dbs&2)
   strcat(text,"sql:");
 return(1);
}


int Imgdb_reset_errmsg(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 imdb->errmsg[0]= 0;
 return(1);
}


int Imgdb_get_errmsg(imdb,msg,flag)
struct ImgdB *imdb;
char *msg;
int flag;
{
 strcpy(msg,imdb->errmsg);
 return(1);
}


int Imgdb__is_writemode(mode,flag)
char *mode;
int flag;
{
 if(strchr(mode,'a') || strchr(mode,'w') || strchr(mode,'+'))
   return(1);
 return(0);
}


int Imgdb_read_conf(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 int ret,i,count= 0,use_dbs= 0;
 char filename[SfileadrL],**names= NULL,**contents= NULL;
 char dbtype[81],varname[81],mirroring[81];
 struct stat stbuf;
 char *Imgdbs__dbtype();

 if(imdb->imdbs!=NULL)
   {ret= 2; goto ex;}

 /* look for sql parameter file */
 ret= Imgdb_get_mapname(imdb,filename,7);
 if(ret<=0)
   goto ex;
 if(stat(filename,&stbuf)==-1) 
   {ret= 2; goto ex;}
 ret= Sfile_read_vars(filename,&names,&contents,&count,0);
 if(ret<=0)
   goto ex;

 /* look for defined names and interpret content */
 varname[0]= 0;

 /* mandatory setting */
 strcpy(varname,"dbtype");
 ret= Sfile_find_var(varname,names,contents,count,dbtype,sizeof(dbtype),0);
 if(ret<=0)
   goto ex;
 varname[0]= 0;

 /* optional */
 ret= Sfile_find_var("mirroring",names,contents,count,mirroring,
                     sizeof(mirroring),0);
 if(ret<0)
   goto ex;

 /* dbtype specific configurations */
 if(strcmp(dbtype,"files")==0 || strcmp(dbtype,"file")==0) {
   /* currently no configuration of the files storage system */;
   imdb->read_dbs= 1;
   imdb->write_dbs= 1;
 } else {
   if(strcmp(dbtype,Imgdbs__dbtype(0))!=0) {
     sprintf(imdb->errmsg,"dbtype mismatch. requested: %s , available: files",
                          dbtype);
     if(strcmp("null",Imgdbs__dbtype(0))!=0)
       sprintf(imdb->errmsg+strlen(imdb->errmsg),",%s",Imgdbs__dbtype(0));
     ret= 0; goto ex;
   }
   ret= Imgdbs_read_conf(&(imdb->imdbs),imdb,names,contents,count,0);
   if(ret<=0)
     goto ex;
   imdb->read_dbs= 2;
   imdb->write_dbs= 2;
 }

 /* eventually set mirroring (read from files, write both) */
 if(strcmp(mirroring,"on")==0 || strcmp(mirroring,"1")==0) {
   imdb->read_dbs= 1;
   imdb->write_dbs|= 1;
 }

 ret= 1;
ex:;
 if(ret==0 && varname[0]!=0) {
   sprintf(imdb->errmsg,"Required variable missing or content too long.  %s",
                        varname); 
 }
 /* release possibly allocated variable list */
 Sfile_read_vars(filename,&names,&contents,&count,1<<15);
 return(ret);
}


int Imgdb_open(imdb,mode,flag)
struct ImgdB *imdb;
char *mode;
int flag;
/*
 bit0= eventually close database connection and re-open
       (else: let connection untouched)
 bit14= create database if missing
*/
{
 int ret= 0,dbs;

 if(Imgdb__is_writemode(mode,0)) {
   imdb->is_writemode= 1;
   dbs= imdb->write_dbs;
 } else {
   imdb->is_writemode= 0;
   dbs= imdb->read_dbs;
 }
 if(dbs&1) {
   ret= Imgdbf_fopen(imdb->imdbf,mode,1|flag&(1<<14));
   if(ret<=0)
     goto ex;
   ret= Imgdbf_fopen(imdb->imdbf,mode,2|flag&(1<<14));
   if(ret<=0) {
     ret= Imgdbf_fclose(imdb->imdbf,1);
     goto ex;
   }
 }
 if(dbs&2) {
   if(Imgdbs_is_open(imdb->imdbs,0)) {
     if(flag&1) 
       Imgdbs_close(imdb->imdbs,0);
     else
       {ret= 2; goto ex;}
   }
   ret= Imgdbs_open(imdb->imdbs,flag&(1<<14));
   if(ret<=0)
     goto ex;
 }
 strcpy(imdb->mode,mode);
 ret= 1;
ex:;
 return(ret);
}


int Imgdb_close(imdb,flag)
struct ImgdB *imdb;
int flag;
/*
 bit0= soft close : flush and possibly close (depends on data base type)
*/
{
 /* close all possible connections (they know wether they are open) */
 Imgdbf_fclose(imdb->imdbf,2);
 Imgdbf_fclose(imdb->imdbf,1);
 if(imdb->imdbs!=NULL && !(flag&1))
   Imgdbs_close(imdb->imdbs,0);
 return(1);
}


int Imgdb_start_query(imdb,samp,name,flag)
struct ImgdB *imdb;
struct ImgsamplE *samp;
char *name;
int flag;
/*
 bit0= prepare for writing (else assume readonly) (anyway use read_dbs)
 bit1= sample contains image data for possible pre-selection
 bit2= name contains fileaddress for possible pre-selection
 bit3= caching is fully active
 bit4= prepare for possible intermediate queries
*/
{
 int ret= 0,dbs;
 char mode[8];

 if(imdb->query_token!=NULL) {
   ret= Imgdb_end_query(imdb,0);
   if(ret<=0)
     goto ex; 
 }
 imdb->is_writemode= 0;
 dbs= imdb->read_dbs;
 if(dbs&1) {                     /* dbf does not pre-select names or samples */
   if(flag&1)
     strcpy(mode,"r+b");
   else
     strcpy(mode,"rb");
   ret= Imgdbf_fopen(imdb->imdbf,mode,1|flag&(1<<14));
   if(ret<=0)
     goto ex;
   ret= Imgdbf_fopen(imdb->imdbf,mode,2|flag&(1<<14));
   if(ret<=0) {
     ret= Imgdbf_fclose(imdb->imdbf,1);
     goto ex;
   }
   ret= 1; goto ex;
 }
 if(dbs&2) { 
   ret= Imgdbs_init_search_query(imdb->imdbs,samp,name,
                                 (!!(flag&8)) | (flag&(2|4|16)));
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 if(ret>0) {
   strcpy(imdb->mode,mode);
   if(Sregex_string(&(imdb->query_token),"imgdb_query_token",0)<=0)
     ret= -1;
 }
 return(ret);
}


int Imgdb_end_query(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 if(imdb->read_dbs&1)
   Imgdb_close(imdb,1);
 if(imdb->read_dbs&2)
   Imgdbs_end_query(imdb->imdbs,1);
 Sregex_string(&(imdb->query_token),NULL,0);
 return(1);
}


int Imgdb_is_open(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 int ret= 0,dbs;

 if(imdb->is_writemode)
   dbs= imdb->write_dbs;
 else
   dbs= imdb->read_dbs;
 ret= 0;
 if(dbs&1) 
   ret= Imgdbf_is_open(imdb->imdbf,0);
 if(dbs&2) 
   ret|= Imgdbs_is_open(imdb->imdbs,0);
 return(ret);
}


int Imgdb_db_exists(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 int ret= 0,dbs;

 if(imdb->is_writemode)
   dbs= imdb->write_dbs;
 else
   dbs= imdb->read_dbs;
 ret= 0;
 if(dbs&1) 
   ret|= Imgdbf_db_exists(imdb->imdbf,0);
 if(dbs&2) 
   ret|= Imgdbs_db_exists(imdb->imdbs,0);
 return(ret);
}


int Imgdb_activate_searchpos(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_activate_searchpos(imdb->imdbf,0);
   if(ret<=0)
     return(ret);
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_activate_searchpos(imdb->imdbs,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Imgdb_preserve_searchpos(imdb,flag)
struct ImgdB *imdb;
int flag;
/*
 bit15= reset
*/
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_preserve_searchpos(imdb->imdbf,flag);
   if(ret<=0)
     return(ret);
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_preserve_searchpos(imdb->imdbs,0);
   if(ret<=0)
     return(ret);
 }
 return(ret);
}


int Imgdb_address_samp(imdb,samp,flag)
struct ImgdB *imdb;
struct ImgsamplE *samp;
int flag;
{
 int ret= 0,dbs;

 if(imdb->is_writemode && (flag&1)) 
   dbs= imdb->write_dbs;
 else
   dbs= imdb->read_dbs;
 if(dbs&1) { 
   ret= Imgdbf_address_samp(imdb->imdbf,samp,flag&1);
   if(ret<=0)
     goto ex;
 }
 if(dbs&2)
   ret= 1;
ex:;
 return(ret);
}


int Imgdb_write_samp(imdb,writesample,sadr,flag)
struct ImgdB *imdb;
struct ImgsamplE *writesample;
int flag;
/*
 bit15= mark written sample as completely invalid
*/
{
 int ret= 0;

 /* Write SQL to obtain any DBMS generated id. Then eventually write to file */
 if(imdb->write_dbs&2) {
   ret= Imgdbs_write_samp(imdb->imdbs,writesample,0,flag&(1<<15));
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&1) {
   ret= Imgdbf_write_samp(imdb->imdbf,writesample,0,(flag&(1<<15))|1);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_record(imdb,filename,writesample,flag)
struct ImgdB *imdb;
char *filename;
struct ImgsamplE *writesample;
int flag;
{
 int ret=0,i;
 char name[SfileadrL];

 Imgdb_reset_errmsg(imdb,0);
 /* Write SQL to obtain the DBMS generated id. Then eventually write to file */
 if(imdb->write_dbs&2) {
   ret= Imgdbs_record(imdb->imdbs,filename,writesample,0);
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&1) {
   ret= Imgdbf_record(imdb->imdbf,filename,writesample,0);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_rename(imdb,samp,new_name,flag)
struct ImgdB *imdb;
struct ImgsamplE *samp;
char *new_name;
int flag;
{
 int ret;

 if(imdb->write_dbs&2) {
   ret= Imgdbs_rename(imdb->imdbs,samp,new_name,0);
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&1) {
   ret= Imgdbf_rename(imdb->imdbf,samp,new_name,0);
   if(ret<=0)
     goto ex;
 }

ex:;
 return(ret);
}


int Imgdb_read_samp(imdb,readsample,flag)
struct ImgdB *imdb;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample
*/
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_read_samp(imdb->imdbf,readsample,flag&1);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_read_samp(imdb->imdbs,readsample,4|(flag&1));
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_read_nam(imdb,readsample,flag)
struct ImgdB *imdb;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample
*/
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_read_nam(imdb->imdbf,readsample,flag);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2)
   ret= 1;
ex:;
 return(ret);
}


int Imgdb_del_back(imdb,samp,flag)
/* delete sample and everything that has a higher address */
struct ImgdB *imdb;
struct ImgsamplE *samp;
int flag;
{
 int ret= 0;

 if(imdb->write_dbs&1) {
   ret= Imgdbf_del_back(imdb->imdbf,samp,0);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_del_back(imdb->imdbs,samp,0);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_rm_sample(imdb,samp,flag)
struct ImgdB *imdb;
struct ImgsamplE *samp;
int flag;
/*
 bit0= keep files open if search is successful
 bit1= no need to reopen files
*/
{
 int ret= 0;

 if(imdb->write_dbs&1) {
   ret= Imgdbf_rm_sample(imdb->imdbf,samp,flag&(1|2));
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&2) {
   ret= Imgdbs_rm_sample(imdb->imdbs,samp,0);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_add_exempt(imdb,filename,lookalikes,flag)
struct ImgdB *imdb;
char *filename;
struct LstrinG **lookalikes;
int flag;
{
 int ret= 0;

 if(imdb->write_dbs&1) {
   ret= Imgdbf_add_exempt(imdb->imdbf,filename,lookalikes,0);
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&2) {
   ret= Imgdbs_add_exempt(imdb->imdbs,filename,lookalikes,0);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_exemptlist_is_changed(imdb,list_seal,seal,flag)
struct ImgdB *imdb;
unsigned long list_seal;
unsigned long *seal;
int flag;
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_exemptlist_is_changed(imdb->imdbf,list_seal,seal,flag);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_exemptlist_is_changed(imdb->imdbs,list_seal,seal,flag);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_count_exempts(imdb,count,flag)
struct ImgdB *imdb;
int *count;
int flag;
/*
 bit0= do not add 1 for security
*/
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_count_exempts(imdb->imdbf,count,flag);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_count_exempts(imdb->imdbs,count,flag);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_read_exempt(imdb,listname,filename,lstring,flag)
struct ImgdB *imdb;
char *listname;
char filename[SfileadrL];
struct LstrinG **lstring;
int flag;
/*
 bit0= resume reading
*/
/*
 return:  1=success
          0=everything is read (or cannot open file)
         -1=fatal failure
*/
{
 int ret= 0;

 if(imdb->read_dbs&1) {
   ret= Imgdbf_read_exempt(imdb->imdbf,listname,filename,lstring,flag&1);
   if(ret<=0)
     goto ex;
 }
 if(imdb->read_dbs&2) {
   ret= Imgdbs_read_exempt(imdb->imdbs,listname,filename,lstring,flag&1);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret); 
}


int Imgdb_garbage_coll(imdb,flag)
struct ImgdB *imdb;
int flag;
{
 int ret= 2;

 if(imdb->write_dbs&1) {
   ret= Imgdbf_garbage_coll(imdb->imdbf,0);
   if(ret<=0)
     goto ex;
 }
 if(imdb->write_dbs&2) {
   ret= Imgdbs_garbage_coll(imdb->imdbs,0);
   if(ret<=0)
     goto ex;
 }
ex:;
 return(ret);
}


int Imgdb_sample_is_valid(imdb,samp,flag)
struct ImgdB *imdb;
struct ImgsamplE *samp;
int flag;
/*
 bit0= check for writability (else validation after read)
 bit1= in case of SQL : insist in valid id
*/
{
 int ret= 0,dbs;

 if(flag&1) 
   dbs= imdb->write_dbs;
 else
   dbs= imdb->read_dbs;
 if(dbs&1) {
   if(samp->invalid&Imgsamp_valid_requireD)
     return(0);
   else
     return(1);
 }
 if(dbs&2) {
   if(samp->invalid&Imgsamp_invalid_coorD)
     return(0);
   if((flag&1)&&!(flag&2))
     return(1);
   if(samp->id>0)
     return(1);
   return(0);
 }  
 return(0);
}


