
/*
 cc -g -c imgmap.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgrgb.h"
#include "imgsamp.h"
#include "imgstd.h"
#include "imgexempt.h"

#include "imgmap.h"



/* ----------------------------- ImgmapcachE ------------------------------ */


int Imgmapcache_new(cache,mapname,flag)
struct ImgmapcachE **cache;
char *mapname;
int flag;
{
 *cache= TSOB_FELD(struct ImgmapcachE,1);
 if(*cache==NULL)
    return(-1);
 (*cache)->refcount= 0;
 strcpy((*cache)->mapname,mapname);
 (*cache)->start_samp= NULL;
 (*cache)->last_samp= NULL;
 return(1);
}


int Imgmapcache_destroy(del_cache,flag)
struct ImgmapcachE **del_cache;
int flag;
{
 struct ImgsamplE *samp,*next,*prev;
 struct ImgmapcachE *cache;
 int removed=0;

 cache= *del_cache;
 if(cache==NULL)
   return(0);
 cache->refcount--;
 if(cache->refcount>0)
   return(1);
 /* this should be much better while full usage of smem.c */
 for(samp= cache->last_samp;samp!=NULL;samp= prev) {
   prev= samp->prev;
   Imgsample_destroy(&samp,0);
 }

/*
 for(samp= cache->start_samp;samp!=NULL;samp= next) {
   next= samp->next;
   Imgsample_destroy(&samp,0);
 }
*/

 Smem_freE((char *) cache);
 *del_cache= NULL;
 return(2);
}


/* -------------------------------- ImgmaP ------------------------------ */


int Imgmap_new(imp,icmp,mapname,flag)
struct ImgmaP **imp;
struct ImgcmP *icmp;
char *mapname;
int flag;
/*
 bit0= do not create sub objects (pseudo non-initialized)
*/
{
 int ret;
 struct ImgmaP *m;

 *imp= m= TSOB_FELD(struct ImgmaP,1);
 if(m==NULL)
   return(-1);
 ret= Imgmap_init(*imp,icmp,mapname,flag&1);
 if(ret<=0)
   goto failed;
 return(1);
failed:;
 Imgmap_destroy(imp,0);
 return(-1);
}


int Imgmap_destroy(imp,flag)
struct ImgmaP **imp;
int flag;
{
 struct ImgmaP *m;
 
 m= *imp;
 if(m==NULL)
   return(0);
 Imgmap_deactivate(m,2|4);
 Smem_freE((char *) m);
 *imp= NULL;
 return(1);
}


int Imgmap_init(imp,icmp,mapname,flag)
struct ImgmaP *imp;
struct ImgcmP *icmp;
char *mapname;
int flag;
/*
 bit0= do not create sub objects (pseudo non-initialized)
*/
{
 int ret;

 imp->icmp= icmp;
 imp->mapname[0]= 0;
 imp->imdb= NULL;
 imp->disabled= 0;
 imp->guarding= 0;
 imp->caching= 0;
 imp->cache= NULL;
 imp->exemptlist= NULL;
 imp->use_exemptlist= 0;
 imp->searchsamp= NULL;
 imp->spoints= Imgmap_sample_sizE;
 imp->matcher= NULL;
 imp->errmsg[0]= 0;
 if(!(flag&1)) {
   if(mapname[0])
     Imgmap_set_mapname(imp,mapname,0);
   ret= Imgmatch_new(&(imp->matcher),0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Imgmap_deactivate(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= do not close
 bit1= destroy ImgdB and ImgmatcH
 bit2= finally destroy all attached objects
*/
{
 if(!(flag&1)) {
   Imgmap_end_query(imp,0);
   if(flag&2) {
     Imgdb_destroy(&(imp->imdb),0);
   } else
     Imgmap_close(imp,0);
 }
 if(flag&4) {;
   Imgmatch_destroy(&(imp->matcher),0);
 }
 Imgexemptlist_destroy(&(imp->exemptlist),0);
 Imgmap_delete_cache(imp,0);
 imp->use_exemptlist= 0;
 return(1);
}


int Imgmap_copy(source,target,flag)
struct ImgmaP *source,*target;
int flag;
/*
 bit0= also copy link to cache
 bit1= target is not initialized yet
 bit2= do not copy reference to exempt list
*/
{
 int ret;

 if(flag&2)
   Imgmap_init(target,source->icmp,source->mapname,0);
 ret= Imgdb_copy(source->imdb,target->imdb,0);
 if(ret<=0)
   goto ex;
 strcpy(target->mapname,source->mapname);
 target->disabled= source->disabled;
 target->guarding= source->guarding;
 target->spoints= source->spoints;
 ret= Imgmatch_copy(source->matcher,target->matcher,0);
 if(ret<=0)
   goto ex;
 strcpy(target->errmsg,source->errmsg);
 if(flag&1){
   target->caching= source->caching;
   Imgmap_attach_cache(target,source->cache,0);
   target->searchsamp= source->searchsamp;
 }else{
   target->caching= 0;
   target->cache= NULL;
   target->searchsamp= NULL;
 }
 if(flag&4) {
   target->exemptlist= NULL;
   target->use_exemptlist;
 } else {
   Imgexemptlist_destroy(&(target->exemptlist),0);
   Imgexemptlist_refer(source->exemptlist,&(target->exemptlist),0);
   target->use_exemptlist= source->use_exemptlist;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_get_icmp(imp,icmp,flag)
struct ImgmaP *imp;
struct ImgcmP **icmp;
int flag;
{
 *icmp= imp->icmp;
 return(1);
}


int Imgmap_get_matcher(imp,matcher,flag)
struct ImgmaP *imp;
struct ImgmatcH **matcher;
int flag;
{
 *matcher= imp->matcher;
 return(1);
}


int Imgmap_set_eps(imp,color_eps,bw_eps,flag)
struct ImgmaP *imp;
double color_eps,bw_eps;
int flag;
{
 int ret= 1;
 ret= Imgmatch_set_eps(imp->matcher,color_eps,bw_eps,0);
 return(ret);
}


int Imgmap_set_spoints(imp,spoints,flag)
struct ImgmaP *imp;
int spoints;
int flag;
{
 imp->spoints= spoints;
 return(1);
}


int Imgmap_get_spoints(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->spoints);
}


int Imgmap_set_mapname(imp,mapname,flag)
struct ImgmaP *imp;
char *mapname;
int flag;
/*
 bit0= do not create database if missing
 bit1= do not open, just set name
*/
{
 int ret,hflag;

 if(strcmp(imp->mapname,mapname)==0)
   return(2);
 Imgmap_deactivate(imp,2);
 strcpy(imp->mapname,mapname);
 if(flag&2)
   return(1);
 hflag= 1|2;
 if(!(flag&1))
    hflag|= (1<<14);
 ret= Imgmap_open(imp,"r+",hflag);
 if(ret<=0)
   return(ret);
 if(ret!=2) {
   ret= Imgmap_close(imp,1);
   if(ret<=0)
     return(ret);
 }
/* 
 ret= Imgdb_new(&(imp->imdb),imp,mapname,imp->spoints,0);
 if(ret<=0)
   return(ret);
*/
 if(imp->caching&1) 
   Imgmap_load_cache(imp,0);
 return(1);
}


int Imgmap_get_mapname(imp,mapname,flag)
struct ImgmaP *imp;
char *mapname;
int flag;
/*
 bit0-7: mode, see Imgdb_get_mapname()
*/
{
 int ret;

 if(imp->imdb==NULL) {
   ret= Imgdb__make_mapname(imp->mapname,Imgmap_sample_sizE,mapname,flag&255);
 } else {
   ret= Imgdb_get_mapname(imp->imdb,mapname,flag&255);
 }
 return(ret);
}


int Imgmap_set_disabled(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->disabled= !!value;
 return(1);
}


int Imgmap_get_disabled(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->disabled);
}


int Imgmap_set_use_exemptlist(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->use_exemptlist= !!value;
 return(1);
}


int Imgmap_get_use_exemptlist(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->use_exemptlist);
}


int Imgmap_set_guarding(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->guarding= !!value;
 return(1);
}


int Imgmap_get_guarding(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->guarding);
}


int Imgmap_get_caching(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->caching);
}


int Imgmap_get_searchsample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE **samp;
int flag;
/*
 bit0= obtain last found sample (there has to a successful search before)
*/
{
 int ret;

 *samp= NULL;
 if((imp->caching&3)!=3)
   return(2);
 *samp= imp->searchsamp;
 if(flag&1) {
   if((*samp)==NULL) 
     *samp= imp->cache->last_samp;
   else
     *samp= (*samp)->prev;
 }
 return(1);
}  


int Imgmap_reset_errmsg(imp,flag)
struct ImgmaP *imp;
int flag;
{
 imp->errmsg[0]= 0;
 return(1);
}


int Imgmap_set_errmsg(imp,text,flag)
struct ImgmaP *imp;
char *text;
int flag;
{
 if(strlen(text)>=sizeof(imp->errmsg))
   return(0);
 strcpy(imp->errmsg,text);
 return(1);
}


char *Imgmap_get_errmsg(imp,msg,flag)
struct ImgmaP *imp;
char *msg;
int flag;
/*
 bit0= return a reference to imp->errmsg and leave msg untouched
       (do the same if msg==NULL)
*/
{
 if((flag&1)||msg==NULL)
   return(imp->errmsg);
 strcpy(msg,imp->errmsg);
 return(msg);
}


int Imgmap_set_match_par(imp,search_method,
                         color_eps,bw_eps,color_max_misses,bw_max_misses,flag)
struct ImgmaP *imp;
int search_method;
double color_eps,bw_eps;
int color_max_misses,bw_max_misses;
int flag;
{
 int ret= 1;

 ret= Imgmatch_set_match_par(imp->matcher,search_method,
                         color_eps,bw_eps,color_max_misses,bw_max_misses,0);
 return(ret);
}


int Imgmap_temp_sample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE **samp;
int flag;
{
 int ret;

 Imgmap_reset_errmsg(imp,0);
 ret= Imgsample_new(samp,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(**samp));
   return(-1);
 }
 (*samp)->spoints= imp->spoints;
 return(1);
}


int Imgmap_assert_db(imp,mode,flag)
/* check wether db is accessible and ensure existence of db object */
struct ImgmaP *imp;
char *mode;
int flag;
/*
 bit1= return(2) if disabled
*/
{
 char filename[SfileadrL];
 int ret;

 if(imp->disabled) {
disabled:;   
   Imgmap_get_mapname(imp,filename,0);
   sprintf(imp->errmsg,
           "Access to guarded map file '%s' is disabled.",filename);
   if(flag&2)
     return(2);
   return(0);
 }
 if(imp->imdb==NULL) {
   ret= Imgdb_new(&(imp->imdb),imp,imp->mapname,imp->spoints,0);
   if(ret<=0)
     return(ret);
 }
 /* check for published server address */
 if(!imp->guarding) {
   Imgmap_get_mapname(imp,filename,5);
   if(Sfile_size(filename,1)!=-1)
     goto disabled;
 }
 return(1);
}


int Imgmap_open(imp,mode,flag)
struct ImgmaP *imp;
char *mode;
int flag;
/*
 bit0= eventually close database connection and re-open 
       (else: let connection untouched)
 bit1= return(2) if disabled (i.e. guarded map)
for Imgdb_open() :
 bit14= create database if missing
*/
{
 char filename[SfileadrL];
 int typ,ret;

 ret= Imgmap_assert_db(imp,mode,flag&2);
 if(ret<=0)
   return(ret);
 if(ret==2)
   return(2);
 if(!(flag&1)) {
   if(Imgdb_is_open(imp->imdb,0))
     return(2);
 }
 ret= Imgdb_open(imp->imdb,mode,flag&(1|(1<<14)));
 if(ret<=0)
   Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
 return(ret);
}


int Imgmap_close(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= soft close : flush and possibly close (depends on data base type)
*/
{
 int ret;

 if(imp->imdb==NULL)
   return(2);
 Imgdb_close(imp->imdb,flag&1);
 if(ret<=0)
   Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
 return(1);
}


int Imgmap_start_query(imp,sample,name,flag)
struct ImgmaP *imp;
struct ImgsamplE *sample;
char *name;
int flag;
/*
 bit0= prepare for writing (else assume readonly)
 bit1= sample contains image data for possible pre-selection 
 bit2= name contains fileaddress for possible pre-selection
 bit3= for Imgmap_search_reset() : do not reset errmsg
 bit4= prepare for possible intermediate queries
*/
{
 int ret;
 char mode[8];

 if(flag&1)
   strcpy(mode,"r+b");
 else
   strcpy(mode,"rb");
 ret= Imgmap_assert_db(imp,mode,0);
 if(ret<=0)
   return(ret);
 Imgmap_search_reset(imp,!!(flag&8));
 ret= Imgdb_start_query(imp->imdb,sample,name,
                        ((imp->caching==3)<<3)|flag&(1|2|4|16));
 return(ret);
}


int Imgmap_end_query(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret= 2;

 if(imp->imdb!=NULL)
   ret= Imgdb_end_query(imp->imdb,0);
 return(ret);
}


int Imgmap_write_samp(imp,writesample,flag)
struct ImgmaP *imp;
struct ImgsamplE *writesample;
int flag;
/*
 bit1= this is certainly a new address

 bit13= no need to reopen and softclose database connection
 bit14= only write sample to cache, db aspects are already done
 bit15= mark written sample as completely invalid
*/
{
 int i,ret,head,sp,hflag;
 struct ImgsamplE *samp=NULL,*temp_samp= NULL;

 if(flag&(1<<15)) 
   writesample->invalid= -1;
 if(flag&2){
   ret= Imgmap_new_cachesamp(imp,writesample,&samp,0);
   if(ret<=0)
     goto ex;
 }else{
   if(Imgmap_find_in_cache(imp,writesample,&samp,0)<=0 && !(flag&(1<<15))) {
     ret= Imgmap_new_cachesamp(imp,writesample,&samp,0);
     if(ret<=0)
       goto ex;
   }
 }
 if(samp!=NULL) {                                         /* write to cache */
   if(flag&(1<<15)) {
     if(writesample==samp) {    /* save writesample from Imgmap_del_sample() */
       ret= Imgmap_temp_sample(imp,&temp_samp,0);
       if(ret<=0)
         goto ex;
       ret= Imgsample_copy(writesample,temp_samp,0);
       if(ret<=0)
         goto ex;
       writesample= temp_samp;
     }
     ret= Imgmap_del_sample(imp,&samp,0);
   } else {
     ret= Imgsample_copy(writesample,samp,0);
     if(ret<=0)
       goto ex;
   }
 }
 if(flag&(1<<14))
   {ret= 1; goto ex;}

 ret= Imgdb_address_samp(imp->imdb,writesample,0);
 if(ret<=0)
   goto ex;
 if(flag&(1<<15)) {
   if(flag&(1<<13))
     hflag= 3;
   else
     hflag= 0;
   ret= Imgdb_rm_sample(imp->imdb,writesample,hflag);
 } else {
   ret= Imgdb_write_samp(imp->imdb,writesample,flag&(1<<15));
 }
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(ret<=0)
   Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
 Imgsample_destroy(&temp_samp,0);
 return(ret);
}


int Imgmap_read_samp(imp,readsample,flag)
struct ImgmaP *imp;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= use address of readsample (else current db read position)

*/
{
 int i,sp,ret;
 struct ImgsamplE *samp=NULL;

 if((imp->caching&3)==3){
   if(flag&1) 
     Imgmap_find_in_cache(imp,readsample,&samp,0);
   else
     samp= imp->searchsamp;
   if(samp==NULL)
     {ret= 0; goto ex;}
 }
 if(samp!=NULL) {
   ret= Imgsample_copy(samp,readsample,1);
   if(ret<=0)
     goto ex;
   imp->searchsamp= samp->next;
   return(1);
 }
 ret= Imgdb_read_samp(imp->imdb,readsample,flag&1);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_record(imp,filename,writesample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *writesample;
int flag;
{
 int ret;

 Imgmap_reset_errmsg(imp,0);
 ret= Imgmap_assert_db(imp,"r+b",0);
 if(ret<=0)
   goto write_err;
 ret= Imgdb_record(imp->imdb,filename,writesample,0);
 if(ret<=0) {
   Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
   goto write_err;
 }
 /* add sample to cache */
 if(writesample->filename!=filename)
   if(Imgsample_set_filename(writesample,filename,0)<=0)
     return(-1);
 if(Imgmap_write_samp(imp,writesample,2|(1<<14))<=0) 
   goto write_err;
 return(1);

write_err:;
 return(0);
}


int Imgmap_del_back(imp,filename,flag)
struct ImgmaP *imp;
char *filename;
int flag;
{
 int ret;
 struct ImgsamplE *samp= NULL;
 char lastname[SfileadrL],name[SfileadrL];

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 ret= Imgmap_start_query(imp,samp,"",8);
 if(ret<=0) 
   goto ex;
 ret= Imgmap_search_adr(imp,lastname,samp,samp,1);
 Imgmap_end_query(imp,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"on deleting last entry: no entries found");
   goto ex;
 }
 if(strcmp(filename,lastname)!=0) {
   sprintf(imp->errmsg,"on deleting last entry: filenames differ");
   sprintf(imp->errmsg+strlen(imp->errmsg),"\ngiven '%s'",filename);
   sprintf(imp->errmsg+strlen(imp->errmsg),"\nfound '%s'",lastname);
   ret= 0; goto ex;
 }
 ret= Imgdb_del_back(imp->imdb,samp,0);
 if(ret<=0) {
   Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
   goto ex;
 }
 Imgmap_del_sample_back(imp,0);
 ret= 1;
ex:;
 Imgmap_end_query(imp,0);
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_search_reset(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= no Imgmap_reset_errmsg()
*/
{
 if(!(flag&1))
   Imgmap_reset_errmsg(imp,0);
 Imgdb_preserve_searchpos(imp->imdb,1<<15);
 if(imp->cache!=NULL)
   imp->searchsamp= imp->cache->start_samp;
 else
   imp->searchsamp= NULL;
 return(1);
}


int Imgmap__cmp_qsort_pt(coord1,coord2)
Imgsamp_coorD *coord1;
Imgsamp_coorD *coord2;
{
 if(*coord1<*coord2)
   return(-1);
 if(*coord1>*coord2)
   return(1);
 return(0);
}


int Imgmap_avg_limits(imp,samp,minmax,flag)
struct ImgmaP *imp;
struct ImgsamplE *samp;
double minmax[6];
int flag;
{
 double avg[3],low_sum[3],high_sum[3],missing,eps;
 Imgsamp_coorD *color;
 int pt[Imgsamp_max_sizE];
 int i,j,ret,spoints,search_method,max_misses;

 /* determine actual tolerance parameters */
 Imgmatch_choose_tolerances(imp->matcher,samp,0);
 Imgmatch_get_searchpar(imp->matcher,&search_method,&eps,&max_misses,0);
 if(search_method!=0) {
   for(i=0;i<3;i++) {
     minmax[2*i]= 0.0;
     minmax[2*i+1]= 256.0;
   }
   ret= 2; goto ex;
 }

 spoints= samp->spoints;
 for(i=0;i<3;i++) {
   Imgsample_get_colorvector(samp,i,&color,0);
   ret= Sregex_qsort_pt(color,spoints,sizeof(Imgsamp_coorD),
                        Imgmap__cmp_qsort_pt,pt,0);
   if(ret<=0)
     goto ex;
   /* low_sum = lowest + N*0 */
   low_sum[i]= 0.0;
   for(j= max_misses;j<spoints-max_misses;j++)
     low_sum[i]+= Imgsample_get_coord(samp,i,pt[j],0);
   /* high_sum = highest + N*Max */
   high_sum[i]= max_misses*256.0;
   for(j= max_misses;j<spoints;j++)
     high_sum[i]+= Imgsample_get_coord(samp,i,pt[j],0);
 }
 for(i=0;i<3;i++) {
   minmax[i*2]=   low_sum[i]/((double) spoints)-eps;
   minmax[i*2+1]= high_sum[i]/((double) spoints)+eps;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_is_exempted(imp,name1,name2,flag)
struct ImgmaP *imp;
char *name1,*name2;
int flag;
/*
 bit2= use exempt list despite imp->use_exemptlist==0
*/
{
 int ret;

 if(!(((flag&4)||imp->use_exemptlist) && name1!=NULL && name2!=NULL))
   {ret= 0; goto ex;}
 if(strcmp(name1,name2)==0)
   {ret= 0; goto ex;}
 ret= Imgexemptlist_is_exempted(imp->exemptlist,name1,name2,0);
 if(ret!=0)
   goto ex;
 /* i consider inequality to be symmetric */
 ret= Imgexemptlist_is_exempted(imp->exemptlist,name2,name1,0);
 if(ret!=0)
   goto ex;
 ret= 0;
ex:
 return(ret);
}
  

int Imgmap_search_adr(imp,filename,insample,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *insample;
struct ImgsamplE *foundsample;
int flag;
/*
 bit0= search last entry
 bit1= accept any valid entry
 bit2= use exempt list despite imp->use_exemptlist==0
*/
{
 int ret,i,j,c,sample_existing=0,invalid_mask= Imgsamp_valid_requireD;
 int max_misses,misses,colorful= 0,boring,nearby;
 char name[SfileadrL];
 struct ImgsamplE *samp= NULL;
 double eps,d,dd,cm[3];

 filename[0]= 0;
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 ret= Imgmatch_choose_tolerances(imp->matcher,insample,0);
 if(ret<=0)
   goto ex;

 ret= Imgmap_open(imp,"rb",0);
 if(ret<=0) {
   Imgcmp_stderr(imp->icmp,imp->errmsg,0);
   goto ex;
 }
 ret= Imgdb_activate_searchpos(imp->imdb,0);
 if(ret<=0)
   goto ex;

 while(1){
   if(Imgmap_read_samp(imp,samp,0)<=0)
 break;
   if(samp->invalid&Imgsamp_invalid_coorD)
 continue;
   sample_existing= 1;
   if(flag&1) {
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto ex;
 continue;
   }
   if((flag&2)){ 
     ret= 1;
   } else {
     ret= Imgmatch_test(imp->matcher,insample,samp,0);
   }
   if(ret>0) {
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto ex;
     Imgdb_preserve_searchpos(imp->imdb,0);
     if(foundsample->filename==NULL) {
       ret= Imgdb_read_nam(imp->imdb,foundsample,1);
       if(ret<=0) {
         Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
         goto ex;
       }
     }
     strcpy(filename,foundsample->filename);
     if(!(flag&2)) {
       ret= Imgmap_is_exempted(imp,insample->filename,filename,flag&4);
       if(ret<0)
         goto ex;
       if(ret>0)
 continue;
     }

     ret= 1; goto ex;
   }
 }
 if((flag&1) && sample_existing) {
   ret= Imgdb_read_nam(imp->imdb,foundsample,1);
   if(ret<=0) {
     Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
     goto ex;
   }
   strcpy(filename,foundsample->filename);
   ret= 1; goto ex;
 }
 ret= 0;
ex:;
 Imgmap_close(imp,1);
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_search(imp,filename,insample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *insample;
int flag;
/*
 flag : is also sent unchanged to Imgmap_search_adr()
 bit0= search last entry
 bit1= accept any valid entry
       copy found sample to insample
 bit2= use exempt list
*/
{
 int ret;
 struct ImgsamplE *samp= NULL;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 ret= Imgmap_search_adr(imp,filename,insample,samp,flag);
 if(ret<=0)
   goto ex;
 if(flag&2) {
   ret= Imgsample_copy(samp,insample,0);
   if(ret<=0)
     goto ex;
 }
   
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_compare(imp,name1,name2,result,flag)
struct ImgmaP *imp;
char *name1,*name2;
int *result;
int flag;
/*
 bit2= check exemptliste despite not globally enabled by imp
*/
{
 int ret;
 struct ImgsamplE *samp1= NULL,*samp2= NULL;

 *result= 0;
 ret= Imgmap_temp_sample(imp,&samp1,0);
 if(ret<=0)
   goto ex;
 ret= Imgmap_temp_sample(imp,&samp2,0);
 if(ret<=0)
   goto ex;
 ret= Imgmap_read_rgb(imp,name1,samp1,2);
 if(ret<=0)
   goto ex;
 ret= Imgmap_read_rgb(imp,name2,samp2,2);
 if(ret<=0)
   goto ex;
 ret= Imgmatch_choose_tolerances(imp->matcher,samp1,0);
 if(ret<=0)
   goto ex;
 ret= Imgmatch_test(imp->matcher,samp1,samp2,0);
 if(ret<0)
   goto ex;
 if(ret>0) {
   ret= Imgmap_is_exempted(imp,name1,name2,flag&4);
   if(ret<0)
     goto ex;
   if(ret==0)
     *result= 1;
 }
 ret= 1;
ex:;
 Imgsample_destroy(&samp2,0);
 Imgsample_destroy(&samp1,0);
 return(ret);
}



int Imgmap_offadr(imp,filename,foundnames,foundsamples,maxfnames,flag)
struct ImgmaP *imp;
char *filename;
char **foundnames;
struct ImgsamplE **foundsamples;
int maxfnames;
int flag;
/*
 bit0= report findings
 bit1= report findings as result
 bit2= return only one filename if possible
 bit3= require identical inode
 bit4= return if exact name is found
 bit5= return 0 if no caching 
*/
{
 int count= 0,ret,wpt= -1,i,passes=1,pass;
 char line[2*SfileadrL];
 struct ImgsamplE *samp= NULL,*foundsamp= NULL,*cache_sample;
 int search_method= -1,color_max_misses,bw_max_misses,cmm,bmm;
 double color_eps,bw_eps,ceps,beps;
 

 if(flag&32)
   if(Imgmap_find_in_cache(imp,0,&cache_sample,1)<=0)
     {ret= 0; goto ex;}

 /* avoid too many matches by very wide search parameters */
 Imgmatch_get_match_par(imp->matcher,&search_method,&color_eps,&bw_eps,
                                     &color_max_misses,&bw_max_misses,0);
 ceps= (color_eps>2*Imgmap_color_epS ? 2*Imgmap_color_epS : color_eps);
 beps= (bw_eps>2*Imgmap_bw_epS ? 2*Imgmap_bw_epS : bw_eps);
 cmm= (color_max_misses>2*Imgmap_color_max_misseS ? 2*Imgmap_color_max_misseS
                                                  : color_max_misses);
 bmm= (bw_max_misses>2*Imgmap_bw_max_misseS ? 2*Imgmap_bw_max_misseS
                                            : bw_max_misses);
 Imgmatch_set_match_par(imp->matcher,0,ceps,beps,cmm,bmm,0);

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 ret= Imgmap_temp_sample(imp,&foundsamp,0);
 if(ret<=0)
   goto ex;

 ret= Imgmap_read_rgb(imp,filename,samp,2);
 if(ret<=0) 
    goto ex;

 if(flag&(4|8|16))
   passes= 3;
 /* pass 3 : require identical name
    pass 2 : require identical inode
    pass 1 : require identical content
 */
 for(pass=passes;pass>!!(flag&8);pass--){
   ret= Imgmap_start_query(imp,samp,"",2);
   if(ret<=0) 
     goto ex;
   while(Imgmap_search_adr(imp,line,samp,foundsamp,0)>0){

     if(line[0]==0) {
       Imgcmp_stderr(imp->icmp,
               "--- program error : Imgmap_search_adr() returns empty line",0);
   continue;
     }

     if(pass==3) {
       if(strcmp(filename,line)!=0)
   continue;
     } 
     for(i=0;i<=wpt;i++) 
       if(strcmp(foundnames[i],line)==0) break;
     if(i<=wpt)
   continue;
     if(!Sfile_identical(filename,line,(pass>=2)))
   continue;
     if(wpt+1<maxfnames)
       wpt++;
     if(wpt>=0) {
       strcpy(foundnames[wpt],line);
       foundsamples[wpt]= NULL;
       if(Imgmap_find_in_cache(imp,foundsamp,&cache_sample,0)>0)
         foundsamples[wpt]= cache_sample;
       else
         /* >>> return a copy of local sample , somehow marked for deletion */;
     }
     if(imp->icmp!=NULL) {
       if(flag&1)
         Imgcmp_stderr(imp->icmp,line,0);
       if(flag&2)
         Imgcmp_result(imp->icmp,line,0);
     }
     count++;
     if(pass==3 && (flag&16))
       {ret= 1; goto ex;}
   }
   if(count>0)
 break;
 }
 ret= count;
ex:
 if(search_method>=0)
   Imgmatch_set_match_par(imp->matcher,search_method,color_eps,bw_eps,
                                       color_max_misses,bw_max_misses,0);
 Imgsample_destroy(&samp,0);
 Imgsample_destroy(&foundsamp,0);
 Imgmap_end_query(imp,0);
 return(ret);
}


int Imgmap_search_filename(imp,filename,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *foundsample;
int flag;
/*
 bit0= keep files open if search is successful
 bit1= open in mode r+
 bit2= search a whole directory (all data files, not subdirectories)
       Empty text finds all files without a directory
 bit3= search a whole directory tree (including subdirectories)
       Empty text finds all files
 bit4= also return invalid items
 bit5= write filename found into parameter filename
       (senseful with empty search text)
 bit6= files are already open
 <<< bit7= searching for exactly one file among whole database
       (permission to use Imgmap_offadr)
*/
{
 int ret,i,c,sample_count=0,searchnl,shortcut_successful=0;
 char *cpt;
 char name[SfileadrL],testname[SfileadrL],searchname[SfileadrL];
 char keptname[SfileadrL],*testname_pt;
 struct ImgsamplE *samp= NULL,*foundsample_pt;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 strcpy(searchname,filename);
 searchnl= strlen(searchname);
 if(flag&8) {
   if(searchnl)  
     if(searchname[searchnl-1]!='/')
       strcat(searchname,"/"); 
 } else if (flag&4) {
   if(searchnl>1)
     if(searchname[searchnl-1]=='/')
       searchname[searchnl-1]= 0;
 }
 searchnl= strlen(searchname);

 if(!(flag&64)) {
   if(flag&2)
     ret= Imgmap_open(imp,"r+b",0);
   else
     ret= Imgmap_open(imp,"rb",0);
   if(ret<=0)
     goto failed;
 }

 if((imp->caching&3)==3 && !(flag&(4|8|16|32))){
   /* simplyfied search over cache */
   ret= Imgmap_name_in_cache(imp,filename,&foundsample_pt,0);
   if(ret<0)
     goto failed;
   if(ret==0)
     goto done;
   sample_count= ret;
   ret= Imgsample_copy(foundsample_pt,foundsample,0);
   if(ret<=0)
     goto failed;
   ret= sample_count;
   goto success;
 }

 Imgdb_activate_searchpos(imp->imdb,0);
 while(1){
   ret= 0;
   if(Imgmap_read_samp(imp,samp,0)<=0)
     goto done;
   if(samp->invalid&Imgsamp_invalid_coorD)
 continue;

   sample_count++;
   Imgdb_preserve_searchpos(imp->imdb,0);

   if(samp->filename==NULL) {
     ret= Imgdb_read_nam(imp->imdb,samp,1);
     if(ret<=0) {
       Imgdb_get_errmsg(imp->imdb,imp->errmsg,0);
       goto failed;
     }
   }
   strcpy(testname,samp->filename);
   strcpy(keptname,testname);
   if(flag&8) { 
     testname[searchnl]= 0;
   } else if(flag&4) {
     cpt= strrchr(testname,'/');
     if(cpt==NULL)
       cpt= testname;
     *cpt= 0;
   }
   if(strcmp(searchname,testname)==0) {
     if(flag&32)
       strcpy(filename,keptname);
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto failed;
     ret= Imgsample_set_filename(foundsample,filename,0);
     if(ret<=0)
       goto failed;
     ret= sample_count;
     goto success;
   }
 }
done:;
failed:;
 Imgmap_close(imp,1);
 if(ret>=0)
   ret= 0;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
success:;
 if(ret<=0)
   ret= 1;
 if(!(flag&1))
   Imgmap_close(imp,1);
 goto ex;
}


int Imgmap_rm_sample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE *samp;
int flag;
/*
 bit13= no need to reopen and softclose database connection
*/
{
 int ret;
 
 if(Imgmap_write_samp(imp,samp,(1<<15)|(flag&(1<<13)))<=0)
   {ret= 0; goto ex;}
 ret= 1;
ex:
 return(ret);
}


int Imgmap_rename(imp,old_name,new_name,flag)
struct ImgmaP *imp;
char *old_name;
char *new_name;
int flag;
{
 int ret,counter= 0,hflag;
 struct ImgsamplE *samp= NULL,*rename_sample;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 hflag= 4;               /* allow pre-selection of filename (not guaranteed) */
 hflag|= 16;                      /* prepare for intermediate db transaction */
 ret= Imgmap_start_query(imp,samp,old_name,hflag);
 if(ret<=0)
   goto ex;
 while(Imgmap_search_filename(imp,old_name,samp,1|2)>0){
   ret= Imgmap_get_searchsample(imp,&rename_sample,1);
   if(ret<=0)
     goto ex;
   if(rename_sample==NULL)
     rename_sample= samp;
   ret= Imgdb_rename(imp->imdb,rename_sample,new_name,0);
   if(ret<=0)
     goto write_err;
   if(Imgsample_set_filename(rename_sample,new_name,0)<=0)
     {ret= -1; goto ex;}
   counter++;
 }
 ret= counter; goto ex;
write_err:;
 ret= -counter;
ex:;
 Imgmap_end_query(imp,0);
 Imgsample_destroy(&samp,0);
 return(ret);
}
 

int Imgmap_rm_filename(imp,filename,flag)
struct ImgmaP *imp;
char *filename;
int flag;
/*
 bit0= permission to apply empty text as filename
 bit2= remove whole directory ( rm dir/*  without subdirectories)
       Empty text finds all files without a directory
 bit3= remove a whole directory tree  ( rm -r dir )
       Empty text finds all files
*/
{
 static char name[SfileadrL];
 int counter=0,null=0,i,ret,hflag;
 double minus1= -1.0;
 struct ImgsamplE *samp= NULL;
 
 if(strlen(filename)==0 && !(flag&1))
   {ret= 0 ; goto ex;}

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 if(flag&(1|4|8))
   hflag= 0;
 else
   hflag= 4;             /* allow pre-selection of filename (not guaranteed) */
 hflag|= 16;                      /* prepare for intermediate db transaction */
 ret= Imgmap_start_query(imp,samp,filename,hflag);
 if(ret<=0)
   goto ex;
 
 while(Imgmap_search_filename(imp,filename,samp,1|2|(flag&(4|8)))>0){
   ret= Imgmap_rm_sample(imp,samp,0);
   if(ret<=0)
     goto write_err;
   counter++;
 }
 ret= counter; goto ex;
write_err:;
 ret= -counter;
ex:;
 Imgmap_end_query(imp,0);
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_refresh_filename(imp,filename,refreshsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *refreshsample;
int flag;
/*
 bit0= permission to apply empty text as filename
 bit2= refresh whole directory ( rm dir/*  without subdirectories)
       Empty text finds all files without a directory
 bit3= refresh a whole directory tree  ( rm -r dir )
       Empty text finds all files
*/
{
 int counter=0,null=0,i,ret,shortcut_successful= 0,hflag;
 struct ImgsamplE *samp= NULL;
 
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 if(flag&(1|4|8))
   hflag= 0;
 else
   hflag= 4;             /* allow pre-selection of filename (not guaranteed) */
 hflag|= 16;                      /* prepare for intermediate db transaction */
 ret= Imgmap_start_query(imp,samp,filename,hflag|1);
 if(ret<=0)
   goto ex;
 
 if(strlen(filename)==0 && !(flag&1)) 
   {ret= 0; goto ex;}
 
 hflag= 1|2|(flag&(4|8));
 if(!(flag&(1|4|8)))
   hflag|= 128;
 while((ret= Imgmap_search_filename(imp,filename,samp,hflag))>0){
   if(ret==2)
     shortcut_successful= 1;
   Imgsample_copy_adr(samp,refreshsample,0);
   if(Imgmap_write_samp(imp,refreshsample,0)<=0)
     goto write_err;
   counter++;
   if(shortcut_successful)
 break;
 }
 if(counter<=0)
   sprintf(imp->errmsg,"refresh failed. address not registered : '%s'",
                       filename);
 ret= counter; goto ex;
write_err:;
 strcpy(imp->errmsg,"cannot write sample data: ");
 Imgdb_get_errmsg(imp->imdb,imp->errmsg+strlen(imp->errmsg),0);
 ret= -counter;
ex:;
 Imgmap_end_query(imp,0);
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_lookup(imp,filename,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *foundsample;
int flag;
{
 int ret;

 ret= Imgmap_start_query(imp,foundsample,filename,4);
 if(ret<=0)
   goto ex;
 ret= Imgmap_search_filename(imp,filename,foundsample,0);
 if(ret>0)
   ret= 1;
ex:;
 Imgmap_end_query(imp,0);
 return(ret);
}


int Imgmap_copy_all(source,target,flag)
struct ImgmaP *source,*target;
int flag;
{
 int ret,i;
 char filename[SfileadrL],line[2*SfileadrL];
 struct ImgsamplE *samp= NULL;
 struct LstrinG *lookalikes= NULL;
 struct ImgexemptlisT *exempts;
 
 ret= Imgmap_temp_sample(source,&samp,0);
 if(ret<=0) 
   goto ex;
 
 Imgmap_reset_errmsg(source,0);
 Imgmap_reset_errmsg(target,0);
 ret= Imgmap_start_query(source,samp,"",0);
 if(ret<=0)
   goto ex;

 filename[0]= 0;
 while(Imgmap_search_filename(source,filename,samp,8|32)>0){
   ret= Imgmap_record(target,filename,samp,0);
   if(ret<=0)
     goto ex;
   filename[0]= 0;
 }
 if(Imgmap_get_errmsg(source,line,0)[0]!=0 || 
    Imgmap_get_errmsg(target,line,0)[0]!=0)
   {ret= 0; goto ex;}
 
 ret= Imgmap_activate_exempt(source,1);
 if(ret<=0)
   goto ex;
 /* write to target */
 exempts= source->exemptlist;
 for(i=0;i<exempts->count;i++) {
   ret= Lstring_copy_all(exempts->exempt[i]->lookalikes,&lookalikes,0);
   if(ret<=0)
     goto ex;
   ret= Imgmap_add_exempt(target,exempts->exempt[i]->filename,&lookalikes,0);
   if(ret<=0)
     goto ex;
   Lstring_destroy_all(&lookalikes,0); /* just in case they were not grabbed */
 }

 ret= 1;
ex:;
 Lstring_destroy_all(&lookalikes,0);
 Imgmap_end_query(source,0);
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_activate_exempt(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= read in any case
*/
{
 int ret;
 char listname[SfileadrL];

 ret= Imgmap_get_mapname(imp,listname,6);
 if(ret<=0)
   return(ret);
 if(imp->exemptlist==NULL) {
   ret= Imgexemptlist_new(&(imp->exemptlist),imp,listname,1);
   if(ret<=0)
     return(ret);
   flag|= 1;
 }
 /* if !bit0 : read only if file has changed */
 ret= Imgmap_read_exemptlist(imp,!(flag&1));
 return(ret);
}


int Imgmap_add_exempt(imp,filename,lookalikes,flag)
struct ImgmaP *imp;
char *filename;
struct LstrinG **lookalikes; /* these may get grabbed by Imgexempt_new()
                                *lookalikes is set to NULL in that case */
int flag;
{
 int ret;
 struct LstrinG *ls;
 char listname[SfileadrL];

 ret= Imgdb_add_exempt(imp->imdb,filename,lookalikes,0);
 if(ret<=0)
   goto ex;
 if(!Imgmap_get_use_exemptlist(imp,0))
   {ret= 1; goto ex;} 

 if(Imgexemptlist_can_still_add(imp->exemptlist,0)) {
   ret= Imgexemptlist_add(imp->exemptlist,filename,lookalikes,0);
 } else {
   ret= Imgmap_activate_exempt(imp,0);
 }
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_read_exemptlist(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= only read if listname or modification time have changed
*/
{
 int ret,count,i,wpt,state,followup;
 char listname[SfileadrL],filename[SfileadrL];
 unsigned long seal;
 struct LstrinG *lstring= NULL;

 ret= Imgmap_get_mapname(imp,listname,6);
 if(ret<=0)
   return(ret);
 if(flag&1) {
   /* is the list unchanged ? */
   ret= Imgdb_exemptlist_is_changed(imp->imdb,imp->exemptlist->list_seal,
                                    &seal,0);
   if(ret<0)
     return(ret);
   if(ret==0)
     if(strcmp(imp->exemptlist->listname,listname)==0)
       return(2);
 }
 Imgexemptlist_set_listname(imp->exemptlist,NULL,0);
 Imgexemptlist_set_seal(imp->exemptlist,seal,0);
 Imgexemptlist_destroy_exempts(imp->exemptlist,0);
 
 ret= Imgdb_count_exempts(imp->imdb,&count,0);
 if(ret<=0)
   return(ret);
 ret= Imgexemptlist_new_exempts(imp->exemptlist,count,0);
 if(ret<=0)
   return(ret);

 followup= 0;
 wpt= 0;
 while(1) {
   ret= Imgdb_read_exempt(imp->imdb,listname,filename,&lstring,followup);
   followup= 1;
   if(ret<0)
     goto ex;
   if(ret==0)
 break;
   ret= Imgexempt_new(&(imp->exemptlist->exempt[wpt]),filename,&lstring,0);
   if(ret<=0)
     goto ex;
   wpt++;
   if(wpt>=count)
 break;
 }
 Imgexemptlist_set_count(imp->exemptlist,wpt,0);

 ret= Imgexemptlist_set_listname(imp->exemptlist,listname,0);
 if(ret<=0)
   goto ex;
 Imgexemptlist_set_seal(imp->exemptlist,seal,0);
 ret= Imgexemptlist_sort(imp->exemptlist,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_exemptlist_is_changed(imp,list_seal,seal,flag)
struct ImgmaP *imp;
unsigned long list_seal;
unsigned long *seal;
int flag;
{
 int ret= 0;

 ret= Imgdb_exemptlist_is_changed(imp->imdb,list_seal,seal,0);
 return(ret);
}


int Imgmap_garbage_coll(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret;
 ret= Imgdb_garbage_coll(imp->imdb,0);
 return(ret);
}


int Imgmap_attach_cache(imp,cache,flag)
struct ImgmaP *imp;
struct ImgmapcachE *cache;
int flag;
{
 if(imp->cache!=NULL)
   Imgmap_delete_cache(imp,0);
 imp->cache= cache;
 if(cache!=NULL)
   cache->refcount++;
 return(1);
}


int Imgmap_make_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 struct ImgmapcachE *cache;

 if(Imgmapcache_new(&cache,imp->mapname,0)<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(*cache));
   return(-1);
 }
 Imgmap_attach_cache(imp,cache,0);
 return(1);
}


int Imgmap_new_cachesamp(imp,adr_sample,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE *adr_sample;
struct ImgsamplE **samp;
int flag;
{
 int ret;

 *samp= NULL;
 if((!(imp->caching&1))||(imp->cache==NULL))
   return(2);
 ret= Imgsample_new(samp,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(**samp));
   return(ret);
 }
 if(imp->cache->last_samp==NULL)
   imp->cache->start_samp= *samp;
 else
   imp->cache->last_samp->next= *samp;
 (*samp)->next= NULL;
 (*samp)->prev= imp->cache->last_samp;
 imp->cache->last_samp= *samp;
 Imgsample_copy_adr(adr_sample,*samp,0);
 return(1);
}


int Imgmap_del_sample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE **samp;
int flag;
/*
 bit0= set *samp to (*samp)->prev
*/
{
 struct ImgsamplE *next,*prev;

 if(*samp==NULL)
   return(0);
 next= (*samp)->next;
 prev= (*samp)->prev;
 if(*samp==imp->cache->start_samp)
   imp->cache->start_samp= next;
 if(*samp==imp->cache->last_samp)
   imp->cache->last_samp= prev;
 if(prev!=NULL) 
   prev->next= next;
 if(next!=NULL)  
   next->prev= prev;
 Imgsample_destroy(samp,0);
 if(flag&1)
   (*samp)= prev;
 else
   (*samp)= NULL;
 return(1);
}


int Imgmap_del_sample_back(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret;

 if((imp->caching&3)!=3)
   return(0);
 if(imp->cache==NULL)
   return(0);
 if(imp->cache->last_samp==NULL)
   return(0);
 ret= Imgmap_del_sample(imp,&(imp->cache->last_samp),1);
 if(imp->cache->last_samp==NULL)
   imp->cache->start_samp= NULL;
 imp->searchsamp= NULL;
 return(ret);
}


int Imgmap_db_exists(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret;

 ret= Imgmap_assert_db(imp,"r",2);
 if(ret==0)
   ret= -2;
 if(ret!=1)
   goto ex;
 ret= Imgdb_db_exists(imp->imdb,0);  
ex:
 return(ret);
}


int Imgmap_load_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret,is_open;
 struct ImgsamplE *samp=NULL,*newsamp=NULL;
 char filename[SfileadrL],line[2*SfileadrL];
 unsigned long counter=0;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;

 if(Imgmap_make_cache(imp,0)<=0) 
   {ret= 0; goto ex;}
 
 ret= Imgmap_db_exists(imp,0);
 if(ret==-1)
   goto ex;
 if(ret<0)
   {ret= 0; goto ex;}
 if(ret==0)
   {ret= 2; goto ex;}                    /* missing database is not an error */
 ret= Imgmap_start_query(imp,samp,"",0);
 if(ret<=0)
   goto ex;

 filename[0]= 0;
 while(Imgmap_search_filename(imp,filename,samp,1|8|32|is_open)>0){
   is_open= 64;
   if(Imgmap_new_cachesamp(imp,samp,&newsamp,0)<=0) 
     {ret= 0; goto ex;}
   ret= Imgsample_copy(samp,newsamp,0);
   if(ret<=0)
     goto ex;
   filename[0]= 0;
   counter++;
   if((counter%1000)==0)
     fprintf(stderr,"loading cache : %ld\r",counter);
 }
 imp->caching|= 2;
 fprintf(stderr,"entries loaded in cache : %ld\n",counter);
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 Imgmap_end_query(imp,0);
 return(ret);
}


int Imgmap_delete_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 struct ImgsamplE *samp,*alt=NULL;
 struct ImgmapcachE *cache;
 int ret,removed=0;

 ret= Imgmapcache_destroy(&(imp->cache),0);
 removed= (ret==2);
 imp->cache= NULL;
 imp->caching&= ~2;
 return(1+removed);
}


int Imgmap_find_in_cache(imp,searchsample,foundsample,flag)
struct ImgmaP *imp;
struct ImgsamplE *searchsample;
struct ImgsamplE **foundsample;
int flag;
/*
 bit0= do not search: return 1 if cache is valid, 0 if not
*/
{
 struct ImgsamplE *samp,*alt=NULL;

 *foundsample= NULL;
 if((!(imp->caching&1))||(imp->cache==NULL))
   return(0);
 if(flag&1)
   return(1);
 samp= imp->cache->start_samp;
 while(samp!=NULL){
   if(samp->invalid&1) {             /* is this sample solely id addressed ? */
     if(samp->id>0) {
       if(searchsample->id==samp->id)
         goto found;
     }
   } else {
     if(searchsample->samp_adr==samp->samp_adr) {
found:;
       *foundsample= samp;
       return(1);
     }
   }
   samp= samp->next;
 }
 return(0);
}


int Imgmap_name_in_cache(imp,name,foundsample,flag)
struct ImgmaP *imp;
char *name;
struct ImgsamplE **foundsample;
int flag;
/*
 bit0= do not search: return 1 if cache is valid, 0 if not
 bit1= reset search position before search
*/
{
 struct ImgsamplE *samp,*alt=NULL;
 int sample_count= 0;

 *foundsample= NULL;
 if((!(imp->caching&1))||(imp->cache==NULL))
   return(0);
 if(flag&1)
   return(1);
 if(flag&2)
   imp->searchsamp= samp= imp->cache->start_samp;
 else
   samp= imp->searchsamp;
 while(samp!=NULL){
   /* if(!(samp->invalid&2)) { */
   if(samp->filename!=NULL) {
     sample_count++;
     if(strcmp(samp->filename,name)==0) {
       *foundsample= samp;
       imp->searchsamp= samp->next;
       return(sample_count);
     }
   }
   samp= samp->next;
 }
 imp->searchsamp= NULL;
 return(0);
}


int Imgmap_open_rgb(imp,filename,rgb,flag)
struct ImgmaP *imp;
char *filename;
struct ImgrgB **rgb;
int flag;

{
 int ret;
 struct ImgstD *std= NULL;

 Imgmap_reset_errmsg(imp,0);
 Imgrgb_destroy(rgb,0);
 ret= Imgrgb_new(rgb,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
                       sizeof(*rgb));
   goto ex;
 }
 ret= Imgrgb_open(*rgb,filename,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0 && *rgb!=NULL) {
   if(imp->errmsg[0]==0)
     sprintf(imp->errmsg,"reading '%s' , got error : %s",
             Imgrgb_get_openname(*rgb,0),Imgrgb_get_errmsg(*rgb,0));
   if(strcmp(Imgrgb_get_openname(*rgb,0),filename)!=0)
     sprintf(imp->errmsg+strlen(imp->errmsg),
             "\n(original filename '%s')",filename);
   Imgrgb_destroy(rgb,0);
 }
 return(ret);
}


int Imgmap_read_rgb(imp,filename,readsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *readsample;
int flag;
/*
 bit0= do not check for serialized sample data in filename
 bit1= remove sample data from filename if it is a serialized sample
*/
{
 int ret,status;
 struct ImgrgB *rgb= NULL;
 struct ImgstD *std= NULL;

 Imgmap_reset_errmsg(imp,0);
 if(!(flag&1)) {
   ret= Imgsample_deserialize(readsample,filename,&status,0);
   if(ret>0) {      /* this is really a serialized sample and no fileaddress */
     if(flag&2) 
       strcpy(filename,readsample->filename);
     goto ex;
   }
 }
 ret= Imgmap_open_rgb(imp,filename,&rgb,0);
 if(ret<=0)
   goto ex;
 ret= Imgstd_new(&std,rgb,imp->spoints,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
                       sizeof(*std,0));
   goto ex;
 }
 ret= Imgstd_read_rgb(std,imp->errmsg,0);
 if(ret<=0)
   goto ex;
 ret= Imgsample_set_rgb(readsample,imp->spoints,
                        std->red,std->green,std->blue,0);
 if(ret<=0)
   goto ex;
 ret= Imgsample_set_filename(readsample,filename,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0 && rgb!=NULL) {
   if(imp->errmsg[0]==0)
     sprintf(imp->errmsg,"reading '%s' , got error : %s",
                          Imgrgb_get_openname(rgb,0),Imgrgb_get_errmsg(rgb,0));
   if(strcmp(Imgrgb_get_openname(rgb,0),filename)!=0)
     sprintf(imp->errmsg+strlen(imp->errmsg),
             "\n(original filename '%s')",filename);
 }
 Imgstd_destroy(&std,0);
 Imgrgb_destroy(&rgb,0);
 return(ret);
}


int Imgmap_list_exempts(imp,expr,flag)
struct ImgmaP *imp;
char *expr;
int flag;
{
 int idx,ret;
 char *result= NULL;
 struct SregeX *sregex= NULL;
 struct ImgexempT *exempt;

 if(imp->exemptlist==NULL) {
   /* >>> try to load exemptlist and mark for deactivation */;
   ret= 2; goto ex;
 }
 ret= Sregex_new(&sregex,expr,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"Cannot init regular expression : %s",expr);
   goto ex;
 }
 /* ist exempts with filenames matching the expression */
 idx= 0;
 while(1) {
   ret= Imgexemptlist_next_match(imp->exemptlist,sregex,&exempt,&idx,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 break;
   ret= Imgexempt_format(exempt,&result,0);
   if(ret<=0)
     goto ex;
   Imgcmp_result(imp->icmp,result,0);
 }
 ret= 1;
ex:;
 Sregex_string(&result,NULL,0);
 Sregex_destroy(&sregex,0);
 /* >>> eventually deactivate self activated exempt list */;
 return(ret);
}


int Imgmap_deformat_exempt(imp,text,filename,lookalikes,flag)
struct ImgmaP *imp;
char *text;
char **filename;       /* *filename must be NULL or point to freeable memory */
struct LstrinG **lookalikes;
int flag;
{
 int ret;
 struct ImgexempT *exempt= NULL;

 ret= Imgexempt_deformat(&exempt,text,0);
 if(ret<=0)
   goto ex;
 *filename= exempt->filename;
 exempt->filename= NULL;
 *lookalikes= exempt->lookalikes;
 exempt->lookalikes= NULL;
 ret= 1;
ex:;
 Imgexempt_destroy(&exempt,0);
 return(ret);
}


