
#ifndef Imgrgb_includeD
#define Imgrgb_includeD

#ifdef Imgrgb_itselF

#include "../s_tools/sfile.h"


/* Default name of directory for temporary files */
#define Imgrgb_tempfile_diR "/tmp"


struct Imgrgb_error_handlE {
  /* !!! Important : this must be the first field of the structure !!! */
  struct jpeg_error_mgr jerr;            /* inherited part of error handler */
  
  /* own fields of error handler */
  jmp_buf setjmp_buffer; 
  int reaction;                    /* 0=exit(1) , 1=longjump(setjmp_buffer) */
  struct ImgrgB *chef;                          /* Pointer to owning object */
};


struct ImgrgB {
  FILE *fp;
  struct jpeg_decompress_struct cinfo;

  struct Imgrgb_error_handlE error_handle;

  char originalname[SfileadrL];        /* original filename */
  char filename[SfileadrL];            /* Name of (possibly temporary) file */
  char openname[SfileadrL];            /* Name of currently open file 
                                         (possibly a single frame) */
  char tempdir[SfileadrL];
  int tempfile;                       /* 0=original file
                                         1=temporary copy
                                         2=temporary copy, already unlinked */
  int frameno;                        /* Number of frames
                                         (if multi-image or movie) */
  int current_frame;                  /* index of current frame (<frameno) */
  int started;                        /* 1= reading process is initiated */
  JSAMPARRAY buffer;                  /* library style linebuffer */
  int app_linebuffer_valid;
  double *red,*green,*blue;           /* application style linebuffer */
  char errmsg[2*SfileadrL];           /* possible error message */
};


/* Create buffer lines by builtin memory manager */
#define Imgrgb_line_buffer_liB

/* unlink temporary file immediately after opening */
#define Imgrgb_line_early_unlinK

#endif /* Imgrgb_itselF */


char *Imgrgb_get_openname();
char *Imgrgb_get_errmsg();

double *Imgrgb_get_red();
double *Imgrgb_get_green();
double *Imgrgb_get_blue();

#endif /* ! Imgrgb_includeD */


/*
 ImgrgB shields an application from the intestines of libjpeg and tries to
 make readable other file formats than JPEG.
 It provides size and RGB information about an image file. If the file contains
 more than one frame (image), in this version only the first frame is processed.

functions:
 Imgrgb...
  _new()            create new rgb object
  _open()           open image file
  _get_sizes()      Width, height, depth of an open image file
  _read_line(()     read next image line
  _get_pixel()      get a pixel from current line
  _sizeof()         get size of struct ImgrgB
  _get_openname()   get name of currently open file
  _get_errmsg()     get error message
  _get_red()        get red component of current line
  _get_green()      get green component of current line
  _get_blue()       get blue component of current line
  _destroy()        close and destroy rgb object

  _init()           set to empty default values (called by _new())
  _close()          close opened image file and remove temporary files
  _error_exit()     error handler including longjump
  _temp_unlink()    remove temporary files
  _frames()         counts or deletes all frames of an image file
  _del_frames()     frontend to _frames()
  _count_frames()   frontend to _frames()
  _may_be_convertible()  evaluate convertiblity of file format
  _convert()        make formats other than JPEG readable (by convert(1))
*/

