
/*
 cc -g -c imgstd.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imgrgb.h"

#include "imgstd.h"



/* ------------------------------- ImgstD ------------------------------ */


int Imgstd_new(std,rgb,spoints,flag)
struct ImgstD **std;
struct ImgrgB *rgb;
int spoints;
int flag;
{
 int i;
 struct ImgstD *s;

 *std= s= TSOB_FELD(struct ImgstD,1);
 if(s==NULL)
   return(-1);
 s->rgb= rgb;
 for(i=0;i<Imgsamp_max_sizE;i++) 
   s->red[i]= s->green[i]= s->blue[i]= 0.0;
 s->spoints= spoints;
 s->elength= 0;
 s->width= s->height= s->depth= 0;
 s->wstep= s->hstep= 0.0;
 s->lines_seen= 0;
 s->pixel_seen= 0;
 return(1);
}


int Imgstd_destroy(std,flag)
struct ImgstD **std;
int flag;
{
 if(*std==NULL)
   return(0);
 Smem_freE((char *) *std);
 *std= NULL;
 return(1);
}


int Imgstd_rgb_sizes(std,flag)
struct ImgstD *std;
int flag;
{
 double e;

 Imgrgb_get_sizes(std->rgb,&(std->height),&(std->width),&(std->depth),0);
 switch(std->spoints) {
 case 1:  e= 1; break;
 case 4:  e= 2; break;
 case 9:  e= 3; break;
 case 16: e= 4; break;
 case 25: e= 5; break;
 case 36: e= 6; break;
 case 49: e= 7; break;
 case 64: e= 8; break;
 default: e= 1; break; 
 }
 std->elength= e;
 std->wstep= ((double) std->width)/e;
 std->hstep= ((double) std->height)/e;
 return(1);
}


int Imgstd__idx(rgb_coord,std_coord,step,max,flag)
int rgb_coord;
double *std_coord;
double step;
int max;
int flag;
{
 int idx;

 *std_coord= ((double) rgb_coord)/step;
 if(*std_coord<0)
   *std_coord= 0;
 else if(*std_coord>max) 
   *std_coord= max;
 idx= *std_coord;
 if(idx>=max)
   idx--;
 return(idx);
}


int Imgstd_add_pixels(std,v_idx,h_weight,flag)
struct ImgstD *std;
int v_idx;
double h_weight;
int flag;
{
 int ret,idx,h_idx,og= 0,pixpos= 0,dummy;
 int lidx,hidx;
 double low,high,w_weight,weight,*red,*green,*blue;

 red= Imgrgb_get_red(std->rgb,0);
 green= Imgrgb_get_green(std->rgb,0);
 blue= Imgrgb_get_blue(std->rgb,0);

 for(h_idx=0;h_idx<std->elength;h_idx++) {
   idx= v_idx*std->elength+h_idx;
   og= (double)(h_idx+1)*std->wstep;
   for(;pixpos<og;pixpos++) {
     std->red[idx]+=   red[pixpos]*h_weight;
     std->green[idx]+= green[pixpos]*h_weight;
     std->blue[idx]+=  blue[pixpos]*h_weight;
   }
   if(pixpos>=std->width)
 break;
   lidx= Imgstd__idx(pixpos,&low,std->wstep,std->elength,0);
   hidx= Imgstd__idx(pixpos+1,&high,std->wstep,std->elength,0);
   w_weight= (high-hidx)/(high-low);
   weight= (1.0-w_weight)*h_weight;
   std->red[idx]+=    red[pixpos]*weight;
   std->green[idx]+= green[pixpos]*weight;
   std->blue[idx]+=  blue[pixpos]*weight;
   if(idx+1<std->elength) {
     weight= w_weight*h_weight;
     std->red[idx+1]+=   red[pixpos]*weight;
     std->green[idx+1]+= green[pixpos]*weight;
     std->blue[idx+1]+=  blue[pixpos]*weight;
   }
   pixpos++;
 }
 return(1);
}


int Imgstd_add_line(std,flag)
struct ImgstD *std;
int flag;
{
 double low,high,weight;
 int lidx,hidx;
 
 lidx= Imgstd__idx(std->lines_seen,&low,std->hstep,std->elength,0);
 hidx= Imgstd__idx(std->lines_seen+1,&high,std->hstep,std->elength,0);
 if(lidx==hidx) {
   Imgstd_add_pixels(std,lidx,1.0,0);
 } else {
   weight= (high-hidx)/(high-low);
   Imgstd_add_pixels(std,lidx,1.0-weight,0);
   Imgstd_add_pixels(std,hidx,weight,0);
 }
 return(1); 
}


int Imgstd_read_rgb(std,msg,flag)
struct ImgstD *std;
char *msg;
int flag;
/*
 bit0= verbous
*/
{
 int ret,i;
 struct ImgrgB *rgb;
 double weight;

 rgb= std->rgb;
 Imgstd_rgb_sizes(std,0);
 for(i=0;i<std->height;i++){
   if(Imgrgb_read_line(rgb,1)<=0) {
     if(flag&1) sprintf(msg+strlen(msg),"\n");
     sprintf(msg+strlen(msg),"reading line %ld of '%s' , got error : %s",
                   i+1,Imgrgb_get_openname(rgb,0),Imgrgb_get_errmsg(rgb,0));
     ret= 0; goto ex;
   }
   Imgstd_add_line(std,0);
   std->lines_seen++;
   std->pixel_seen+= std->width;
 } 
 weight= ((double) std->pixel_seen)/std->spoints;
 for(i=0;i<std->spoints;i++) {
   std->red[i]/= weight;
   std->green[i]/= weight;
   std->blue[i]/= weight;
 }
 ret= 1;
ex:
 return(ret);
}



