

/*
 cc -g -c imvguard.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imvmapper.h"
#include "imvguard.h"


/* ------------------------------- ImvlocatioN ---------------------------- */


int Imvlocation_new(location,fileadr,prev,flag)
struct ImvlocatioN **location;
char *fileadr;
struct ImvlocatioN *prev;
int flag;
{
 struct ImvlocatioN *l;
 int ret;
 
 *location= l= TSOB_FELD(struct ImvlocatioN,1);
 if(l==NULL)
   return(-1);
 l->fileadr= NULL;
 l->prev= NULL;
 l->next= NULL;
 ret= Sregex_string(&(l->fileadr),fileadr,0);
 if(ret<=0) 
   goto failed;
 l->prev= prev;
 if(prev!=NULL)
   prev->next= l;
 return(1);
failed:;
 Imvlocation_destroy(location,0);
 return(ret);
}


int Imvlocation_destroy(location,flag)
struct ImvlocatioN **location;
int flag;
{
 if(*location==NULL)
   return(0);
 if((*location)->fileadr!=NULL)
   Smem_freE((char *) (*location)->fileadr);
 if((*location)->prev!=NULL) 
   (*location)->prev->next= (*location)->next;
 if((*location)->next!=NULL) 
   (*location)->next->prev= (*location)->prev;
 Smem_freE((char *) (*location));
 *location= NULL;
 return(1);
}


/* -------------------------------- ImvguarD ------------------------------ */


int Imvguard_new(guard,agent,flag)
struct ImvguarD **guard;
struct AsagenT *agent;
int flag;
{
 struct ImvguarD *g;
 int ret;

 *guard= g= TSOB_FELD(struct ImvguarD,1);
 if(g==NULL)
   return(-1); 
 g->agent= agent;
 g->location= NULL;
 g->loc_count= 0;
 g->reg_adr= NULL;
 g->mapper= NULL;
 g->mapper_cmd= NULL;
 g->mapper_db= NULL;
 g->mapper_glue_char= Imvguard_mapper_glue_chaR;
 g->mapper_args= NULL;
 g->mapper_match_par= NULL;
 g->mapper_knows_rename= 1;
 ret= Sregex_string(&(g->mapper_cmd),Imvguard_mapper_cmD,0);
 if(ret<=0) 
   goto failed;
 ret= Imvguard_set_mapper_db(g,Imvguard_mapper_dB,0);
 if(ret<=0)
   goto failed;
 return(1);
failed:;
 Imvguard_destroy(guard,0);
 return(ret);
}


int Imvguard_destroy(guard,flag)
struct ImvguarD **guard;
int flag;
{
 struct ImvlocatioN *loc,*next_loc;

 if(*guard==NULL)
   return(0);
 loc= (*guard)->location;
 while(loc!=NULL) {
   next_loc= loc->next;
   Imvlocation_destroy(&loc,0);
   loc= next_loc;
 }
 if((*guard)->reg_adr!=NULL)
   Smem_freE((char *) (*guard)->reg_adr);
 if((*guard)->mapper!=NULL)
   Imvmapper_destroy(&((*guard)->mapper),0);
 Sregex_string(&((*guard)->mapper_cmd),NULL,0);
 Sregex_string(&((*guard)->mapper_db),NULL,0);
 Sregex_string(&((*guard)->mapper_args),NULL,0);
 Sregex_string(&((*guard)->mapper_match_par),NULL,0);
 Smem_freE((char *) *guard);
 *guard= NULL;
 return(1);
}


int Imvguard_set_mapper_db(guard,mapper_db,flag)
struct ImvguarD *guard;
char *mapper_db;
int flag;
{
 int ret;
 char line[SfileadrL];

 Imvmapper_destroy(&(guard->mapper),0);
 if(Sregex_string(&(guard->mapper_db),mapper_db,0)<=0)
   return(-1);
 sprintf(line,"%s%c%s",
              guard->mapper_cmd,guard->mapper_glue_char,guard->mapper_db);
 if(guard->mapper_args!=NULL)
   sprintf(line+strlen(line)," %s",guard->mapper_args);
 if(guard->mapper_match_par!=NULL)
   sprintf(line+strlen(line)," %s",guard->mapper_match_par);
 ret= Imvmapper_new(&(guard->mapper),guard,line,guard->mapper_glue_char,0);
 return(ret);
}


int Imvguard_set_mapper_par(guard,cmd,db,glue_char,flag)
struct ImvguarD *guard;
char *cmd,*db,glue_char;
int flag;
{
 int ret;

 if(cmd!=NULL)
   if(Sregex_string(&(guard->mapper_cmd),cmd,0)<=0)
     {ret= -1; goto ex;}
 if(glue_char!=0)
   guard->mapper_glue_char= glue_char;
 if(db!=NULL)
   if(Sregex_string(&(guard->mapper_db),db,0)<=0)
     {ret= -1; goto ex;}
 ret= Imvguard_set_mapper_db(guard,guard->mapper_db,0);
ex:;
 return(ret);
}


int Imvguard_set_mapper_args(guard,args,flag)
struct ImvguarD *guard;
char *args;
int flag;
/*
 bit0= set mapper_match_par rather than mapper_args
*/
{
 int ret;

 if(flag&1) {
   if(Sregex_string(&(guard->mapper_match_par),args,0)<=0)
     {ret= -1; goto ex;}
 } else {
   if(Sregex_string(&(guard->mapper_args),args,0)<=0)
     {ret= -1; goto ex;}
 }
 ret= Imvguard_set_mapper_db(guard,guard->mapper_db,0);
ex:;
 return(ret);
}


int Imvguard_get_mapper_args(guard,args,flag)
struct ImvguarD *guard;
char **args;                       /* must point to NULL or freeable memory */
int flag;
/*
 bit0= get mapper_match_par rather than mapper_args
*/
{
 int ret;

 if(flag&1) {
   if(Sregex_string(args,guard->mapper_match_par,0)<=0)
     {ret= -1; goto ex;}
 } else {
   if(Sregex_string(args,guard->mapper_args,0)<=0)
     {ret= -1; goto ex;}
 }
 ret= 1;
ex:;
 return(ret);
}


int Imvguard_is_guarded(guard,adr,flag)
struct ImvguarD *guard;
char *adr;
int flag;
{
 char dir[SfileadrL],*cpt;
 int ret,l,found= 0,i;
 struct ImvlocatioN *loc,*last_loc;
 struct stat stbuf;

 dir[0]= 0;
 if(adr[0]!='/') {
   getcwd(dir,sizeof(dir));
   strcat(dir,"/");
 }
 cpt= strrchr(adr,'/');
 l= strlen(dir);
 if(cpt==NULL){
   if(l>0)
     if(dir[l-1]=='/')
       dir[--l]= 0;
 }else{
   strncpy(dir+l,adr,cpt-adr);
   *(dir+l+(cpt-adr))= 0;
 }
 last_loc= loc= guard->location;
 for(i=0;i<guard->loc_count && loc!=NULL;i++){
   if(strcmp(loc->fileadr,dir)==0)
     found= i+1;
   last_loc= loc;
   loc= loc->next;
 }
 if(found)
   return(1);
 l= strlen(dir);
 strcat(dir,"/.imv_guarded_directory");
 if(lstat(dir,&stbuf)==-1)
   return(0);
 dir[l]= 0;
 ret= Imvlocation_new(&loc,dir,last_loc,0);
 if(ret<=0)
   return(ret);
 if(last_loc==NULL)
   guard->location= loc;
 else
   last_loc->next= loc;
 (guard->loc_count)++;
 return(1);
}


int Imvguard_open_adr(guard,adr,flag)
struct ImvguarD *guard;
char *adr;
int flag;
{
 char **filev= NULL;
 int filec= 0,ret;

 Sregex_set_dyn_string(&(guard->reg_adr),NULL,0);
 if(!Imvguard_is_guarded(guard,adr,0))
   return(2);
 /* get and memorize identifying information about the object adr */
 ret= Imvmapper_lookup_adr(guard->mapper,adr,&filec,&filev,0);
 if(ret<=0)
   goto ex;
 if(filec<=0) {
   ret= 0;
   goto ex;
 }
 guard->reg_adr= filev[0];
 filev[0]= NULL;
 ret= 1;
ex:;
 Sfile_destroy_argv(&filec,&filev,0);
 return(ret);
}


int Imvguard_valid_adress(guard,in_adr,adr,flag)
struct ImvguarD *guard;
char *in_adr;
char **adr;
int flag;
/*
 bit0= refresh opened reg_adr
*/
{
 if(flag&1) {
   *adr= guard->reg_adr;
   if(*adr==NULL)
     return(0);
 } else
   *adr= in_adr;
 if((*adr)[0]==0)
   return(0);
 if(!Imvguard_is_guarded(guard,*adr,0))
   return(2);
 return(1);
}


int Imvguard_refresh(guard,in_adr,flag)
struct ImvguarD *guard;
char *in_adr;
int flag;
/*
 bit0= refresh opened reg_adr
*/
{
 char *adr;
 int ret;

 ret= Imvguard_valid_adress(guard,in_adr,&adr,flag&1);
 if(ret==0)
   return(2);
 if(ret!=1)
   return(ret);
 ret= Imvmapper_refresh(guard->mapper,adr,0);
 return(ret);
}


int Imvguard_check_in(guard,in_adr,flag)
struct ImvguarD *guard;
char *in_adr;
int flag;
/*
 bit0= use opened reg_adr
 bit1= this is part of a move operation, 
       check_out may have been deferred until now
 bit2= this is an undo of a _check_out with bit1
*/
{
 char *adr;
 int ret,is_valid;

 is_valid= Imvguard_valid_adress(guard,in_adr,&adr,flag&1);
 if(guard->reg_adr!=NULL && (flag&2) && guard->mapper_knows_rename) { 
                                                /* resume a rename operation */
   if(flag&4) { /* abort announced rename */
     Sregex_string(&(guard->reg_adr),NULL,0);
     return(1);
   }
   if(is_valid==1) {
     ret= Imvmapper_rename(guard->mapper,guard->reg_adr,adr,0);
     if(ret<=0)
       return(ret);
   } else {
     ret= Imvmapper_del_file(guard->mapper,guard->reg_adr,1);/* del_file_adr */
     if(ret<=0)
       return(ret);
   }
   Sregex_string(&(guard->reg_adr),NULL,0);
   return(1);
 }
 if(is_valid==0)
   return(2);
 if(is_valid!=1)
   return(is_valid);
 ret= Imvmapper_append(guard->mapper,adr,0);
 return(ret);
}


int Imvguard_check_out(guard,in_adr,flag)
struct ImvguarD *guard;
char *in_adr;
int flag;
/*
 bit0= use opened reg_adr
 bit1= this is part of a move operation and may be deferred until _check_in()
*/
{
 char *adr;
 int ret;

 if(!(flag&1))
   Sregex_string(&(guard->reg_adr),NULL,0);
 ret= Imvguard_valid_adress(guard,in_adr,&adr,flag&1);
 if(ret==0)
   return(2);
 if(ret!=1)
   return(ret);
 if((flag&2) && guard->mapper_knows_rename) {
   /* record official filename for later use with _check_in() */;
   ret= Imvguard_open_adr(guard,adr,0);
   if(ret<=0)
     return(ret);
 } else {
   ret= Imvmapper_del_file(guard->mapper,adr,0);
 }
 return(ret);
}


int Imvguard_seems_known(guard,in_adr,filec,filev,flag)
struct ImvguarD *guard;
char *in_adr;
int *filec;
char ***filev;
int flag;
/*
 bit0= use opened reg_adr
 bit1= search even if in guarded location
*/
{
 char *adr;
 int ret;

 ret= Imvguard_valid_adress(guard,in_adr,&adr,flag&1);
 if(ret<=0)
   return(ret);
 if(ret==1 && !(flag&2))
   return(2);
 ret= Imvmapper_search(guard->mapper,adr,filec,filev,0);
 if(ret<0)
   return(ret);
 return(ret);
}


