

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>


#include "../s_tools/smem.h"

#include "../misc/asxv.h"


int Asxv_new(asxv,flag)
struct AsxV **asxv;
int flag;
{
 struct AsxV *m;
 char *initcmd;
 int ret;
 int Asxv_interpreter();

 *asxv= m= TSOB_FELD(struct AsxV,1);
 if(m==NULL)
   return(-1);
 m->frontend= NULL;
 m->got_any_reply= 0;
 m->reply_timeout= 300;
 m->double_files= 0;
 m->servername= NULL;
 m->portnumber= -1;
 m->username= NULL;
 m->current_idx= -1;
 m->lazy_refresh= 1;
 m->need_refresh= 0;
 m->hide_on_empty= 1;
 m->raise_window= 1;
 m->lower_window= 0;
 m->auto_jpgb= 1;
 m->last_jpgb= NULL;
 m->standalone= 0;
 return(1);
failed:;
 Asxv_destroy(asxv,0);
 return(-1);
}


int Asxv_destroy(asxv,flag)
struct AsxV **asxv;
int flag;
{
 struct AsxV *m;
 
 m= *asxv;
 if(m==NULL)
   return(0);
 Asfrontend_destroy(&(m->frontend),0);
 Sregex_string(&(m->servername),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->last_jpgb),NULL,0);
 Smem_freE((char *) m);
 *asxv= NULL;
 return(1);
}


int Asxv_get_frontend(asxv,frontend,flag)
struct AsxV *asxv;
struct AsfrontenD **frontend;
int flag;
{
 *frontend= asxv->frontend;
 return(1);
}


int Asxv_get_fd(asxv,flag)
struct AsxV *asxv;
int flag;
{
 return(Asfrontend_get_fd(asxv->frontend,0));
}


int Asxv_is_connected(asxv,flag)
struct AsxV *asxv;
int flag;
{
 if(asxv==NULL)
   return(0);
 return(asxv->frontend!=NULL);
}


int Asxv_is_standalone(asxv,flag)
struct AsxV *asxv;
int flag;
{
 if(asxv==NULL)
   return(0);
 return(!!asxv->standalone);
}


int Asxv_has_data(asxv,flag)
struct AsxV *asxv;
int flag;
/*
 bit0= only reply yes if lazy_refresh
*/
{
 if(!Asxv_is_connected(asxv,0))
   return(0);
 if((flag&1) && !asxv->lazy_refresh)
   return(0);
 return(Sfile_fd_is_ready(Asxv_get_fd(asxv,0),0,1)>0); 
}


int Asxv_define_service(asxv,argv,flag)
struct AsxV *asxv;
char *argv[4];
int flag;
{
 char *servername= NULL, *username= NULL;
 int portnumber= -1,ret;

 asxv->double_files= 0;
 if(strcmp(argv[0],"double_files")==0)
   asxv->double_files= 1;
 if(Sregex_string(&(asxv->servername),argv[1],0)<=0)
   {ret= -1; goto ex;}
 sscanf(argv[2],"%ld",&(asxv->portnumber));
 if(Sregex_string(&(asxv->username),argv[3],0)<=0)
   {ret= -1; goto ex;}
 ret= 1;
ex:;
 return(ret);
}


int Asxv_modify_settings(asxv,line,flag)
struct AsxV *asxv;
char *line;
int flag;
{
 int ret;
 char *rpt;
 
 for(rpt= line; rpt!=NULL; rpt= strchr(rpt,',')) {
   if(rpt!=line)
     rpt++;
   if(strncmp(rpt,"lazy_refresh=on",15)==0)
     asxv->lazy_refresh= 1;
   else if(strncmp(rpt,"lazy_refresh=off",16)==0)
     asxv->lazy_refresh= 0;
   else if(strncmp(rpt,"hide_on_empty=on",16)==0)
     asxv->hide_on_empty= 1;
   else if(strncmp(rpt,"hide_on_empty=off",17)==0)
     asxv->hide_on_empty= 0;
   else if(strncmp(rpt,"raise_window=on",15)==0)
     asxv->raise_window= 1;
   else if(strncmp(rpt,"raise_window=off",16)==0)
     asxv->raise_window= 0;
   else if(strncmp(rpt,"lower_window=on",15)==0)
     asxv->lower_window= 1;
   else if(strncmp(rpt,"lower_window=off",16)==0)
     asxv->lower_window= 0;
   else if(strncmp(rpt,"auto_jpgb=on",12)==0)
     asxv->auto_jpgb= 1;
   else if(strncmp(rpt,"auto_jpgb=off",13)==0)
     asxv->auto_jpgb= 0;
   else if(strncmp(rpt,"standalone=on",13)==0)
     asxv->standalone= 1;
   else if(strncmp(rpt,"standalone=off",14)==0)
     asxv->standalone= 0;
 }
 ret= 1;
ex:;
 return(ret);
}


int Asxv_connect(asxv,flag)
struct AsxV *asxv;
int flag;
{
 char *initcmd;
 int ret;
 int Asxv_interpreter();

 if(asxv==NULL)
   return(2);
 Asfrontend_destroy(&(asxv->frontend),0);
 if(asxv->servername==NULL || asxv->portnumber<=0 || asxv->username==NULL)
   return(2);

 ret= Asfrontend_new(&(asxv->frontend),asxv,Asxv_interpreter,
                     asxv->servername,asxv->portnumber,asxv->username,0);
 if(ret<=0)
   goto failed;
 if(asxv->double_files) 
   initcmd= "-add_view:-:double_files:address:open_and_show";
 else 
   initcmd= "-add_view:-:current_file:address:open_and_show";
 ret= Asfrontend_send(asxv->frontend,initcmd,1);
 if(ret<=0)
   goto failed;
 Asxv_hide(asxv,0);
 return(1);
 ret= 1;
failed:;
 Asfrontend_destroy(&(asxv->frontend),0);
 if(ret>0)
   ret= 0;
 return(ret);
}


int Asxv_mini_corner(asxv,flag)
struct AsxV *asxv;
int flag;
{
 XWindowAttributes xwa;

 GetWindowPos(&xwa);
 xwa.width= 4;
 xwa.height= 4;
 if(asxv->double_files)
   xwa.x = dispWIDE - xwa.width - 2*xwa.border_width;
 else
   xwa.x= 0;
 xwa.y= 0;
 SetWindowPos(&xwa);
 return(1);
}


int Asxv_hide(asxv,flag)
struct AsxV *asxv;
int flag;
{
 Asxv_mini_corner(asxv,0);
 if(asxv->raise_window)                /* if it will be raised when shown */
   XLowerWindow(theDisp,mainW);
 return(1);
}


int Asxv_show(asxv,flag)
/* show first file in list */
struct AsxV *asxv;
int flag;
/*
 bit0= no lazy_refresh
*/
{
 int no_repair= 0,ret;
 char *first_name= NULL,line[160];
 XWindowAttributes xwa;
 char *tv;                                            /* should be (void *) */

 Window win;
 int vis;

 if(asxv->lazy_refresh && Sfile_fd_is_ready(Asxv_get_fd(asxv,0),0,1) 
    && !(flag&1)) {
   asxv->need_refresh= 1;
   ret= 2; goto ex;
 }

 if(numnames<=0) { 
   if(asxv->hide_on_empty)
     Asxv_hide(asxv,0);
   if(asxv->lower_window && !asxv->hide_on_empty)
     XLowerWindow(theDisp,mainW);
   ret= 0; goto ex;
 }

 /* hide text windows */
 AsXv_LowerTextWindows(0);

 if(asxv->double_files) {
   /* position window to upper right corner */;
   Asxv_mini_corner(asxv,0);
 } else {
   /* position window to upper left corner */;
   GetWindowPos(&xwa);
   xwa.x = xwa.y= 0;
   SetWindowPos(&xwa);
 }

 if(Sregex_string(&first_name,namelist[0],0)<=0)
   {ret= -1; goto ex;}
 openPic(0);
 if(numnames<=0 && !(Asxv_has_data(asxv,0) && asxv->lazy_refresh)) {
   /* Not a single file was viewable. Show like a canceled file */
   Asxv_show(asxv,1);
   if(asxv->last_jpgb!=NULL)
     if(strcmp(first_name,asxv->last_jpgb)==0)
       no_repair= 1;
   if(asxv->auto_jpgb && !asxv->double_files && !no_repair) {
     /* Make a repair attempt */
     if(asxv->current_idx>=0)
       sprintf(line,"-sidestep:-for:%ld:%ld:-jpgb:interruptible",
                    asxv->current_idx,asxv->current_idx);
     else
       sprintf(line,"-jpgb");
     ret= Asfrontend_send(asxv->frontend,line,1);
     if(ret<=0)
       goto ex;
     if(Sregex_string(&(asxv->last_jpgb),first_name,0)<=0)
       {ret= -1; goto ex;}
   }
   ret= 0; goto ex;
 }
 Sregex_string(&(asxv->last_jpgb),NULL,0);
 asxv->need_refresh= 0;

 if(asxv->raise_window)
   XRaiseWindow(theDisp,mainW);
 ret= 1;
ex:;
 Sregex_string(&first_name,NULL,0);
 return(1);
}


int Asxv_interpreter(asxv,frontend,flag)
struct AsxV *asxv;
struct AsfrontenD *frontend;
int flag;
/*
 possible return values:
  -1=failure, abort receiving
   0=failure, continue receiving
   1=all went well
   2=all went well, more data expected (currently no difference to 1)
   3=all went well, end receiving
*/
{
 int ret,fieldcount,bytecount,i,start_field;
 char *msgtype,*fieldname,*bytept;           /* only references, not copies */
 char *representation,*spt;                  /* only references, not copies */
 char nametext[80];
 struct AsviewmsG *msg;                     /* only reference, not a copy */

 if(Asfrontend_has_ended(frontend,0))
   return(3);
 if(Asfrontend_get_vmsg(frontend,&msg,0)<=0)
   return(0);
 if(Asviewmsg_is_complete(msg,0)<=0)
   return(0); /* we were called prematurely */ 

 Asviewmsg_fetch_type(msg,&msgtype,0);
 fieldcount= Asviewmsg_get_fieldcount(msg,0);
 if(asxv->double_files) {
   if(strcmp(msgtype,"double_files")!=0)
     {ret= 0; goto ex;}
   start_field= 0;
 } else {
   if(strcmp(msgtype,"current_file")!=0)
     {ret= 0; goto ex;}
   start_field= 1;
   if(fieldcount>0) {
     Asviewmsg_fetch_field_idx(msg,0,&fieldname,&bytept,&bytecount,0);
     sscanf(bytept,"%ld",&(asxv->current_idx));
   } else {
     asxv->current_idx= -1;
   }
 }
 
 /* remove all files from xv's list */;
 for(i=numnames-1;i>=0;i--) {
   deleteFromList(i);
 }

 /* add file(s) to xv's list */
 for(i=start_field; i<fieldcount; i++) {
   Asviewmsg_fetch_frep_idx(msg,i,&representation,0);
   if(strcmp(representation,"address")!=0)
 continue;
   Asviewmsg_fetch_field_idx(msg,i,&fieldname,&bytept,&bytecount,0);
   if(Sfile_type(bytept,1)<0)
 continue;
   spt= strrchr(bytept,'/');
   /* beware : this may temporarily change the original message field */
   if(spt==NULL) {
     spt= bytept;
     bytept= "";
   } else {
     *spt= 0;
     spt++;
   }
   AddFNameToCtrlList(bytept,spt);
   /* restore possibly change message field */
   if(bytept[0]!=0)
     *(spt-1)= '/';
 }
 ChangedCtrlList();
 ret= Asxv_show(asxv,0);
 if(ret<0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Asxv_act(asxv,waiting,flag)
/* check for events at frontend and handle them */
struct AsxV *asxv;
int *waiting;                  /* gets set to 0 if connection is established */
int flag;
{
 int ret,msec;
 struct AsfrontenD *frontend;

 if(!Asxv_is_connected(asxv,0)) 
   return(0);
 Asxv_get_frontend(asxv,&frontend,0);
 *waiting= 0;

 msec= 10000;
 if(XPending(theDisp))
   msec= 0;
 if(!Sfile_fd_is_ready(Asxv_get_fd(asxv,0),msec,1)) 
   {ret= 2; goto ex;}
 ret= Asfrontend_callback(frontend);
 if(ret==3) /* request to end service */
   ret= -3;
 if(ret==0)
   if(!Asfrontend_is_valid(frontend,0))
     ret= -2; /* connection broken */
ex:;
 if(ret==1)
   asxv->got_any_reply= 1;
 if(asxv->got_any_reply==0)
   if(Asfrontend_uptime(frontend,0)>asxv->reply_timeout)
     ret= -2;
 return(ret);
}


