
/*
 Compile :
   cc -g -o syenc_decode syenc_decode.c
*/

#include <stdio.h>
#include <string.h>


/* for debugging purposes */
static FILE *fpd= NULL;
static int debug_tee= 0;
static char debug_tee_name[]= {"/u/x.tee"};

#define Syenc_versioN "syenc_decode version 0.2"


int Yenc_getc(fpi)
FILE *fpi;
{
 int c;

 c= getc(fpi);
 if(debug_tee && c!=EOF && fpd!=NULL)
   putc(c,fpd);
 return(c);
}


int Yenc_cut_between_marks(fpi,start_mark,end_mark,result,max_size,flag)
FILE *fpi;
unsigned char *start_mark;
unsigned char *end_mark;
char *result;                   /* ready to take at least  max_size  bytes */
int max_size;                   /* max_size-1 input bytes and a trailing 0 */
int flag;
{
 int c;
 char *rpt,*wpt;

 if(max_size<=0)
   goto overflow;
 rpt= start_mark;
 while(*rpt!=0) {
   c= Yenc_getc(fpi);
   if(c==EOF) 
     goto premature_end;
   if(c==*rpt)
     rpt++;
   else {
     rpt= start_mark;
     if(c=='\n') 
       goto not_in_line;
   }
 }
 wpt= result;
 *wpt= 0;
 rpt= end_mark;
 while(*rpt!=0) {
   c= Yenc_getc(fpi);
   if(c==EOF) 
     goto premature_end;
   if(c==*rpt)
     rpt++;
   else {
     rpt= end_mark;
     if(c=='\n') 
       goto not_in_line;
     if(wpt-result>=max_size-1)
       goto overflow;
     *wpt= c;
     wpt++;
     *wpt= 0;
   }
 }
 return(1);

not_in_line:;
 return(0);
premature_end:;
 return(-2);
overflow:;
 return(-1);
}


main(argc,argv)
int argc;
char **argv;
{
 int i,c,ret= 1,begin_decoding=0,linestart;
 int raw_stream_data= 0,dot_state= 0,dot_warned= 0,parts_decoded= 0;
 long startbyte,endbyte,maxbytes= 0x7FFFFFFF-1,bytecount; 
 long bytelimit= 650*1024*1024;   
 FILE *fpi,*fpo;
 char numtext[80];

 fpi= stdin;
 fpo= stdout;

 for(i= 1; i<argc; i++) {
   if(strcmp(argv[i],"-o")==0 && i+1<argc) {
     i++;
     fpo= fopen(argv[i],"r+b");
     if(fpo==NULL) 
       fpo= fopen(argv[i],"w+b");
     if(fpo==NULL) {
       fprintf(stderr,"%s : failed to open output file  %s\n",argv[0],argv[i]);
       perror("error message ");
       fprintf(stderr,"%s : aborted\n",argv[0]);
       exit(5);
     }
   } else if(strcmp(argv[i],"-l")==0 && i+1<argc) {
     i++;
     sscanf(argv[i],"%ld",&bytelimit);
   } else if(strcmp(argv[i],"-r")==0) {
     raw_stream_data= 1;
   } else if(strcmp(argv[1],"-version")==0 || strcmp(argv[1],"-v")==0) {
     printf("%s\n",Syenc_versioN);
     exit(0);
   } else if(strcmp(argv[1],"-help")==0 || strcmp(argv[1],"-h")==0) {
     ret= 0;
     goto usage;
   } else {
     ret= 1;
     fprintf(stderr,"%s : Unknown option : %s\n",argv[0],argv[i]);
usage:;
     fprintf(stderr,"\nusage : %s [-r] [-o outfile] [-l sizelimit] [-v]\n\n",
                    argv[0]);
     fprintf(stderr,"Decodes binary data given in yEnc format.\n");
     fprintf(stderr,"yEnc sometimes is used in the newsgroups instead\n");
     fprintf(stderr,"of uuencode or MIME/base64.\n");
     fprintf(stderr,"Feed data including \"=ybegin\"-line and \"=yend\".\n");
     fprintf(stderr,"Use option -r if the feeder does not remove leading\n");
     fprintf(stderr,"periods from the raw NNTP stream.\n");
     fprintf(stderr,"For information about yEnc see:  http://www.yenc.org\n\n");
     fprintf(stderr,"-o allows to process multiple message parts in any\n");
     fprintf(stderr,"order. Each one may create or update the binary\n");
     fprintf(stderr,"output file. Without -o the result is sent to stdout.\n");
     fprintf(stderr,"There is a size limit to avoid malicious multiparts.\n");
     fprintf(stderr,"Default is -l 681574400 (650 MB).\n");
     fprintf(stderr,"-v prints the version and exits with return value 0.\n");
     if(ret!=0)
       fprintf(stderr,"\n%s : aborted\n",argv[0]);
     exit(ret);
   }
 }

 if(debug_tee)
   fpd= fopen(debug_tee_name,"w");

look_for_begin:;
 bytecount= 0;
 /* read until the =ybegin line is consumed */
 while(!begin_decoding) {
   linestart= 1;
   while(1) {
     if(linestart) {
       ret= Yenc_cut_between_marks(fpi,"=ybegin "," ",
                                   numtext,sizeof(numtext),0);
       if(ret==1)
         begin_decoding= 1;
       if(ret==-2) 
         goto ex;
       if(ret==0)
   break;
     }

     /* >>> interpretation of headline */;

     linestart= 0;
     c= Yenc_getc(fpi);
     if(c==EOF) 
       goto ex;
     if(c=='\n')
   break;
   }
 }

 while(1) {
   /* process next character */
   c= Yenc_getc(fpi);
   /* ignore CR and LF */
   if(c=='\n') {
     if(raw_stream_data) 
       dot_state= 1;
 continue;
   } else if (dot_state==1) {
     if(c=='.')
       dot_state= 2;
     else
       dot_state= 0;
   } else if (dot_state==2) {
     dot_state= 0;
     if(c=='.')
 continue;
     if(!dot_warned)
       fprintf(stderr,"%s -r : Warning: single dot found at start of line\n",
                      argv[0]);
     dot_warned= 1;
   }    

   if(c=='\r') 
 continue;
   if(c==EOF) 
     goto ex;
   if(c=='=') {
     /* handle escacpes */
     do {
       c= Yenc_getc(fpi);
       if(c==EOF) 
         goto ex;
     } while (c=='\r' || c=='\n');
     if(c=='y') {
       c= Yenc_getc(fpi);
       if(c==EOF)
         goto ex;
       if(c=='e') {
         /* =yend (end regularly) */;

         /* >>> interpretation of footline */

 break;
       } else if(c=='p') {
         /* =ypart */;
         if(fpo==stdout) {
           fprintf(stderr,"%s : Warning : multipart gets decoded to stdout.\n",
                          argv[0]);
         }
         
         /* read begin and end addresses */
         ret= Yenc_cut_between_marks(fpi,"begin="," ",
                                     numtext,sizeof(numtext),0);
         if(ret<=0) {
           fprintf(stderr,"%s : detected code \"=yp\" without \"begin=\".\n",
                          argv[0]);
           fprintf(stderr,"%s : aborted\n",argv[0]);
           exit(4);
         }
         sscanf(numtext,"%ld",&startbyte);
         ret= Yenc_cut_between_marks(fpi,"end=","\n",
                                     numtext,sizeof(numtext),0);
         if(ret<=0) {
           fprintf(stderr,"%s : detected code \"=yp\" without \"end=\".\n",
                          argv[0]);
           fprintf(stderr,"%s : aborted\n",argv[0]);
           exit(4);
         }
         sscanf(numtext,"%ld",&endbyte);
         if(startbyte>bytelimit || startbyte<=0) {
           fprintf(stderr,"%s : Multipart begin=%ld out of range (1 to %ld)\n",
                          argv[0],startbyte,bytelimit);
           fprintf(stderr,"%s : aborted\n",argv[0]);
           exit(4);
         }
         if(endbyte>bytelimit || endbyte<=0) {
           fprintf(stderr,"%s : Multipart end=%ld out of range (1 to %ld)\n",
                          argv[0],endbyte,bytelimit-1);
           fprintf(stderr,"%s : aborted\n",argv[0]);
           exit(4);
         }
         maxbytes= endbyte+1-startbyte;
         bytecount= 0;
         if(fpo!=stdout)
           fseek(fpo,startbyte-1,0);
 continue;
       } else {
         fprintf(stderr,"%s : detected unknown code beginning with \"=y\"\n",
                        argv[0]);
         fprintf(stderr,"%s : aborted\n",argv[0]);
         exit(3);
       }
     } else {
       c-= 64;
       if(c<0) 
         c+= 256;
     }
   }
   c-= 42;
   if(c<0)
     c+= 256;
   bytecount++;
   if(bytecount>maxbytes) {
      if(bytecount==maxbytes+1) 
        fprintf(stderr,
                "%s : bytecount overflow (announced %ld). Ignoring surplus.\n",
                argv[0],maxbytes);
 continue;
   }
   putc(c,fpo);
 }
 if(bytecount>maxbytes) 
   fprintf(stderr,"%s : Ignored %ld bytes.\n",argv[0],bytecount-maxbytes);

 parts_decoded++;
 goto look_for_begin; 

ex:;
 if(parts_decoded<1) {
   fprintf(stderr,"%s : premature EOF of input data\n",argv[0]);
   if(ferror(stdin))
     perror("error message ");
   fprintf(stderr,"%s : aborted\n",argv[0]);
     exit(2);
 }
 if(fpo!=stdout)
   fclose(fpo);
 if(fpd!=NULL)
   fclose(fpd);
 exit(0);
}

