
/*
  cc -g -c nntphead.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#include "../s_tools/sregex.h"
#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntphead.h"


/* ------------------------ NntpheaD ------------------------------ */


int Nntphead_new(head,conn,msgid,flag)
struct NntpheaD **head;
struct NntpconN *conn;
char *msgid;
int flag;
/*
 bit0= do not aquire HEAD from NNTP server
 bit1= do not see invalid article as error
*/
{
 struct NntpheaD *h;
 int i,ret;

 *head= h= TSOB_FELD(struct NntpheaD,1);
 if(h==NULL){
   conn->errn= errno;
   sprintf(conn->errmsg,
           "Error %d on creating internal header structure:\n%s",
           errno,strerror(errno));
   return(-1);
 }
 for(i=0;i<Nntphead_max_boundarY;i++)
   h->boundary_stack[i]= NULL;
 h->boundary_stack_pt= 0;
 h->lines= NULL;
 Nntphead_reset(h,1);
 h->conn= conn;
 h->msgno= 0;
 strcpy(h->msgid,msgid);
 if(!(flag&1)) {
   if((ret= Nntphead_read(*head,flag&2))<=0){
     Nntphead_destroy(head,0);
     return(ret);
   }
 }
 
 return(1);
}


int Nntphead_destroy_boundary_stack(head,flag)
struct NntpheaD *head;
int flag;
{
 int i;

 for(i=0;i<head->boundary_stack_pt;i++)
   if(head->boundary_stack[i]!=NULL) {
     free((char *) head->boundary_stack[i]);
     head->boundary_stack[i]= NULL;
   }
 return(1);
}


int Nntphead_destroy(head,flag)
struct NntpheaD **head;
int flag;
{
 int i;

 if((*head)==NULL)
   return(0);
 Nntphead_reset(*head,0);
 free((char *) *head);
 *head= NULL;
 return(1);
}


int Nntphead_reset(h,flag)
struct NntpheaD *h;
int flag;
/*
 bit0= reset msgid too
*/
{
 h->status= 0;
 if(flag&1){
   h->msgno= -1;
   h->msgid[0]= 0;
 }
 h->from[0]= 0;
 h->date[0]= 0;
 h->subject[0]= 0;
 h->reference[0]= 0;
 h->all_references[0]= 0;
 h->boundary[0]= 0;
 Nntphead_destroy_boundary_stack(h,0);
 h->is_mime_multipart= 0;
 h->is_mime= 0;
 h->is_uuencode= 0;
 h->byte_count= -1;
 h->line_count= -1;
 Lstring_destroy_all(&(h->lines),0);
 return(1);
}


int Nntphead_is_validmsg(head,flag)
struct NntpheaD *head;
int flag;
{
 return(head->msgno>=0 && head->status>0);
}


int Nntphead_set_msgadr(head,msgno,msgid,flag)
struct NntpheaD *head;
int msgno;
char *msgid;
int flag;
/*
 bit0= do not reset head
*/
{
 head->msgno= msgno;
 if(msgid!=NULL)
   strcpy(head->msgid,msgid);
 if(!(flag&1))
   Nntphead_reset(head,0);
 return(1);
}


int Nntphead_boundary_push(head,boundary,flag)
struct NntpheaD *head;
char *boundary;
int flag;
{
 if(head->boundary_stack_pt>=Nntphead_max_boundarY) 
   return(0);
 
 head->boundary_stack[head->boundary_stack_pt]= 
   TSOB_FELD(char,strlen(head->boundary)+1);
 if(head->boundary_stack[head->boundary_stack_pt]==NULL)
   return(-1);
 strcpy(head->boundary_stack[head->boundary_stack_pt],head->boundary);
 head->boundary_stack_pt++;
 strcpy(head->boundary,boundary);
 return(1);
}


int Nntphead_boundary_pop(head,flag)
struct NntpheaD *head;
int flag;
{
 if(head->boundary_stack_pt<=0)
   return(0);
 head->boundary_stack_pt--;
 strcpy(head->boundary,head->boundary_stack[head->boundary_stack_pt]);
 free((char *) head->boundary_stack[head->boundary_stack_pt]);
 head->boundary_stack[head->boundary_stack_pt]= NULL;
 return(1);
}


int Nntphead_parse_line(line,key,foundkey,data,flag)
char *line;
char *key;
char *foundkey;
char *data;
int flag;
{
 int l;
 char *cpt;

 cpt= line;
 if(strcmp(foundkey,key)!=0) {
   if(!Sregex_complines(key,line,1|2)) {
     /* foundkey[0]= 0; */
     data[0]= 0;
     return(0);
   }
   cpt= line+strlen(key);
 }
 if(strstr(line,";\r\n")!=NULL || strstr(line,";\n")!=NULL ||
    line[strlen(line)-1]==';')
   strcpy(foundkey,key);
 else
   foundkey[0]= 0;
 strcpy(data,cpt);
 Sregex_trimline(data,2); /* preserving trailing blanks just for tradition */

 return(1);
}


int Nntphead_read(head,flag)
struct NntpheaD *head;
int flag;
/*
 bit0= ARTICLE Format: do not send HEAD, end at empty line
 bit1= do not see invalid article as error
 bit3= write to conn->outlet
*/
{
 char line[Nntphead_strL],foundkey[80],data[Nntphead_strL],*cpt;
 char groupname[4096];
 struct NntpconN *conn;
 int ret;

 conn= head->conn;

 Nntphead_reset(head,0);

 if(!(flag&1)){
   /* request header without reading data lines */
   if(head->msgno>0){
     sprintf(data,"HEAD %ld",head->msgno);
   } else if(head->msgid[0]!=0){
     sprintf(data,"HEAD %s",head->msgid);
   } else {
     sprintf(data,"HEAD");
   }
   ret= Nntpconn_transaction(conn,data,conn->buffer,
                             sizeof(conn->buffer),"",2|1024);
   if(ret<=0){
     head->status= -1;
     return(ret);
   }
   if(conn->status_code!=221) {
     head->status= -1;
     if(flag&2)
       return(1);
     conn->errn= -27;
     sprintf(conn->errmsg,"request refused : %s\nstatus reply    : %s",
                          data,conn->status_line);
     return(0);
   }
 }

 /* read and interpret reply */
 head->status= 1;
 while(Nntpconn_getline(conn,line,sizeof(line),0)>0){
   if(flag&8)
     Nntpconn_result(conn,line,"",0,"",8);
   if(Sregex_complines(".",line,0))
 break;
   if(flag&1)
     if(Sregex_complines("",line,0))
 break;

   ret= Lstring_new(&(head->lines),line,head->lines,1);
   if(ret<=0){
     head->status= -1;
     return(ret);
   }
   /* scan reply for useful informations */
   if(Nntphead_parse_line(line,"Date:",foundkey,data,0)>0){
     if(strlen(data)<=Nntphead_strL)
       strcpy(head->date,data);
   } else if(Nntphead_parse_line(line,"From:",foundkey,data,0)>0){
     if(strlen(data)<=Nntphead_strL)
       strcpy(head->from,data);
   } else if(Nntphead_parse_line(line,"Subject:",foundkey,data,0)>0){
     if(strlen(data)<=Nntphead_strL)
       strcpy(head->subject,data);
   } else if(Nntphead_parse_line(line,"Content-Type:",foundkey,data,0)>0){
     if(strncmp(line,"Content-Type: text/plain",24)==0 ||
        strncmp(line,"Content-Type: Text/Plain",24)==0) {
       head->is_mime= 0;
     } else {
       Sregex_read_text(data,"boundary=\"","\"",head->boundary,
                        sizeof(head->boundary),1); /* do not clear on failure */
       if(head->boundary[0]!=0)
         head->is_mime_multipart= 1;
     }
   } else if(Nntphead_parse_line(line,"Message-ID:",foundkey,data,0)>0){
     strcpy(head->msgid,data);
     if(head->msgno<0)
       head->msgno= 0;
   } else if(Nntphead_parse_line(line,"References:",foundkey,data,0)>0){
     Sregex_trimline(data,0);
     /* only take the last reference id from the list */
     cpt= strrchr(data,' ');
     if(cpt==NULL) cpt= data;
     else          cpt++;
     if(strlen(cpt)<=Nntphead_strL)
       strcpy(head->reference,cpt);
     if(strlen(head->all_references)+strlen(data)<Nntphead_refL) {
       if(head->all_references[0])
         strcat(head->all_references," ");
       strcat(head->all_references,data);
     }
   } else if(Nntphead_parse_line(line,"Xref:",foundkey,data,0)>0){
     if(head->msgno<=0){
       Nntpconn_get_groupname(head->conn,groupname,0);
       if(groupname[0]!=0){
         cpt= strstr(data,groupname);
         if(cpt!=NULL){
           cpt= strchr(cpt,':');
           if(cpt!=NULL)
             sscanf(cpt+1,"%ld",&(head->msgno));
         }
       }
     }
   } else if(Nntphead_parse_line(line,"Mime-Version:",foundkey,data,0)>0){
     head->is_mime= 1;
   } else if(Nntphead_parse_line(line,"MIME-Version:",foundkey,data,0)>0){
     head->is_mime= 1;
   } else if(Nntphead_parse_line(line,"MIME-VERSION:",foundkey,data,0)>0){
     head->is_mime= 1;
   } else if(Nntphead_parse_line(line,"Lines:",foundkey,data,0)>0){
     sscanf(data,"%ld",&(head->line_count));
   }
 }
 Lstring_revert(&(head->lines),0);
 head->status= 2;
 return(1);
}


int Nntphead_check_for_mime_boundary(head,line,flag)
struct NntpheaD *head;
char *line;
int flag;
{
 char foundkey[80],data[Nntphead_strL];

 if(Nntphead_parse_line(line,"Content-Type:",foundkey,data,0)>0){
try_boundary:; 
   Sregex_read_text(data,"boundary=\"","\"",head->boundary,
                    sizeof(head->boundary),1); /* do not clear on failure */
   if(head->boundary[0]!=0)
     head->is_mime_multipart= 1;
   else if(strstr(line,";\r\n")!=NULL || strstr(line,";\n")!=NULL ||
           line[strlen(line)-1]==';')
     head->is_mime= 3;                            /* wait for continued line */
   else if(head->is_mime=3)
     head->is_mime= 1;                         /* line will not be continued */
   return(1);
 } else if(head->is_mime==3) {
   strcpy(data,line);
   goto try_boundary; 
 }
 return(0);
}


int Nntphead_guess_filename(line,name,namesize,flag)
char *line;
char *name;
int namesize;
int flag;
{
 char *cpt,*npt;
 int i,l;
 static char suffix[][8]= {
       "JPG","jpg","Jpg","JPEG","jpeg","Jpeg","GIF","gif","Gif",
       "" };

 cpt= line;
 while(1){
   npt= strchr(cpt+1,'.');
   if(npt==NULL)
 break;
   cpt= npt;
   for(i=0;suffix[i][0]!=0;i++)
     if(strncmp(cpt+1,suffix[i],strlen(suffix[i]))==0)
   break;
   if(suffix[i][0]==0)
 continue;
   npt= cpt+strlen(suffix[i])+1;
   for(;cpt!=line;cpt--) 
     if(isspace(*cpt) || *cpt=='/' || *cpt=='\\' || *cpt=='"')
   break;
   if(isspace(*cpt) || *cpt=='/' || *cpt=='\\' || *cpt=='"')
     cpt++;
   l= npt-cpt;
   if(l>=namesize)
     return(-1);
   for(i=0;i<l;i++)
     name[i]= cpt[i];
   name[i]= 0;
   while(name[0]=='.' || name[0]=='-') {
     l= strlen(name);
     for(i=1;i<=l;i++)
       name[i-1]= name[i];
   }
   return(1);
 }
 return(0);
}


int Nntphead_get_property(head,code,prop_text,prop_size,flag)
struct NntpheaD *head;
char *code;
char *prop_text;
int prop_size;
int flag;
{
 int l;

 if(head==NULL)
   return(0);

 prop_text[0]= 0;
 if(strcmp(code,"from:")==0){
   if(head->from[0]!=0)
     strncpy(prop_text,head->from,prop_size-1);
 } else if(strcmp(code,"date:")==0){
   if(head->date[0]!=0)
     strncpy(prop_text,head->date,prop_size-1);
 } else if(strcmp(code,"subject:")==0){
   if(head->subject[0]!=0)
     strncpy(prop_text,head->subject,prop_size-1);
 } else if(strcmp(code,"subjectfile:")==0){
    char name[SfileadrL];
    if(head->subject[0]!=0)
      if(Nntphead_guess_filename(head->subject,name,sizeof(name),0)>0)
        strncpy(prop_text,name,prop_size-1);

 } else if(strcmp(code,"references:")==0){
   strncpy(prop_text,head->all_references,prop_size-1);
 } else if(strcmp(code,"msgid:")==0){
   strncpy(prop_text,head->msgid,prop_size-1);
 } else if(strcmp(code,"msgno:")==0){
   char numbertext[80];
   sprintf(numbertext,"%ld",head->msgno);
   strncpy(prop_text,numbertext,prop_size-1);
 } else if(strcmp(code,"lines:")==0){
   if(head->line_count>=0)
     sprintf(prop_text,"%ld lines",head->line_count);
   else 
     sprintf(prop_text,"? lines");
 } else if(strcmp(code,"bytes:")==0){
   if(head->byte_count>=0)
     sprintf(prop_text,"%ld bytes",head->byte_count);
   else 
     sprintf(prop_text,"? lines");
 } else if(strcmp(code,"size:")==0 || 
           strcmp(code,"larger:")==0 ||
           strcmp(code,"smaller:")==0 ){
   if(head->byte_count>=0)
     sprintf(prop_text,"%ld",head->byte_count);
   else if(head->line_count>=0)
     sprintf(prop_text,"%ld",head->line_count*80);
   else
     sprintf(prop_text,"");
 } else
   return(0);
 prop_text[prop_size-1]= 0; /* just in case that strncpy truncated */
 l= strlen(prop_text);
 if(l>0)
   if(prop_text[l-1]=='\n')
     prop_text[--l]= 0;
 if(l>0)
   if(prop_text[l-1]=='\r')
     prop_text[--l]= 0;
 return(1);
}


int Nntphead_set_property(head,code,prop_text,flag)
struct NntpheaD *head;
char *code;
char *prop_text;
int flag;
/*
 bit0= estimate bytes from lines if necessary
*/
{
 int l;
 double bytes;
 char *cpt;

 if(strcmp(code,"from:")==0){
   strncpy(head->from,prop_text,Nntphead_strL-1);
   head->from[Nntphead_strL-1]= 0;
 } else if(strcmp(code,"date:")==0){
   strncpy(head->date,prop_text,Nntphead_strL-1);
   head->date[Nntphead_strL-1]= 0;
 } else if(strcmp(code,"subject:")==0){
   strncpy(head->subject,prop_text,Nntphead_strL-1);
   head->subject[Nntphead_strL-1]= 0;
 } else if(strcmp(code,"references:")==0){
   strncpy(head->all_references,prop_text,Nntphead_refL-1);
   head->all_references[Nntphead_refL-1]= 0;
   /* only take the last reference id from the list */
   cpt= strrchr(prop_text,' ');
   if(cpt==NULL) cpt= prop_text;
   else          cpt++;
   if(strlen(cpt)<=Nntphead_strL)
     strcpy(head->reference,cpt);
 } else if(strcmp(code,"msgid:")==0){
   strncpy(head->msgid,prop_text,Nntphead_strL-1);
   head->msgid[Nntphead_strL-1]= 0;
 } else if(strcmp(code,"msgno:")==0){
   sscanf(prop_text,"%ld",&(head->msgno));
 } else if(strcmp(code,"bytes:")==0){
   head->byte_count= 0;
   sscanf(prop_text,"%ld",&(head->byte_count));
   if(flag&1)
     if(head->byte_count<=0 && head->line_count>0)
       head->byte_count= head->line_count*63; /* estimating uuencoded lines */
 } else if(strcmp(code,"lines:")==0){
   sscanf(prop_text,"%ld",&(head->line_count));
   if(flag&1)
     if(head->byte_count<=0 && head->line_count>0)
       head->byte_count= head->line_count*63; /* estimating uuencoded lines */
 } else
   return(0);
 return(1);
}


/* ------------------------ NntpsubheaH ------------------------ */


int Nntpsubhead_new(head,boss,flag)
struct NntpsubheaD **head;
struct NntpheaD *boss;
int flag;
/*
 bit0= do not read data from NNTP server
 bit8= for Nntpsubhead_read(): write to conn->outlet
*/
{
 struct NntpconN *conn;
 struct NntpsubheaD *h;
 int ret;

 *head= h= TSOB_FELD(struct NntpsubheaD,1);
 if(h==NULL){
   conn->errn= errno;
   sprintf(conn->errmsg,
           "Error %d on creating internal header structure:\n%s",
           errno,strerror(errno));
   return(-1);
 }

 h->lines= NULL;
 Nntpsubhead_reset(h,0);
 h->boss= boss;
 if(!(flag&1)) {
   if((ret= Nntpsubhead_read(*head,flag&8))<=0){
     Nntpsubhead_destroy(head,0);
     return(ret);
   }
 }
 return(1);
}


int Nntpsubhead_reset(h,flag)
struct NntpsubheaD *h;
int flag;
{
 h->status= 0;
 h->is_loadable= 0;
 h->encoding= 0;
 h->filename[0]= 0;
 Lstring_destroy_all(&(h->lines),0);
 return(1);
}


int Nntpsubhead_destroy(head,flag)
struct NntpsubheaD **head;
int flag;
{
 if((*head)==NULL)
   return(0);
 Lstring_destroy_all(&((*head)->lines),0);
 free((char *) *head);
 *head= NULL;
 return(1);
}


int Nntpsubhead_read(head,flag)
struct NntpsubheaD *head;
int flag;
/*
 bit3= write to conn->outlet
*/
{
 char line[Nntphead_strL],foundkey[80],data[Nntphead_strL];
 char boundary[Nntphead_strL];
 static char buffertext[4096];
 struct NntpconN *conn;
 int ret,l;

 conn= head->boss->conn;
 Nntpsubhead_reset(head,0);

 /* read and interpret reply */
 head->status= 1;
 while(Nntpconn_getline(conn,line,sizeof(line),0)>0){
   if(flag&8)
     Nntpconn_result(conn,line,"",0,"",8);
   if(Sregex_complines("",line,0))
 break;

   ret= Lstring_new(&(head->lines),line,head->lines,1);
   if(ret<=0){
     head->status= -1;
     return(ret);
   }
   /* scan reply for useful informations */
   if(Nntphead_parse_line(line,"Content-Transfer-Encoding:",foundkey,data,0)>0){
     l= strlen("Content-Transfer-Encoding:");
     sscanf(line+l,"%s",buffertext);
     /* determine type */;
     if(strcmp(buffertext,"base64")==0) {
       head->encoding= 1;
     } else if(strcmp(buffertext,"x-uuencode")==0) {
       head->encoding= 2;
     } else {
       head->encoding= 0;
     }
   } else if(Nntphead_parse_line(line,"Content-Type:",foundkey,data,0)>0){
     l= strlen("Content-Type:");
     if(Sregex_complines(" picture",line+l,1))
       head->is_loadable= 1;
     if(Sregex_complines(" image",line+l,1))
       head->is_loadable= 1;
     if(Sregex_complines(" application",line+l,1))
       head->is_loadable= 1;
     Sregex_read_text(data,"name=\"","\"",head->filename,
                      sizeof(head->filename),1); /* do not clear on failure */
     Sregex_read_text(data,"boundary=\"","\"",boundary,
                      sizeof(boundary),0);
     if(boundary[0]!=0) {
        ret= Nntphead_boundary_push(head->boss,boundary,0);
        if(ret<=0)
          /* >>> what to do if boundary stack overflows ? */;
     }
   } else if(Nntphead_parse_line(line,"Content-Disposition:",
                                 foundkey,data,0)>0){
     l= strlen("Content-Disposition:");
     Sregex_read_text(data,"filename=\"","\"",head->filename,
                      sizeof(head->filename),1); /* do not clear on failure */
   }
 }
 Lstring_revert(&(head->lines),0);
 head->status= 2;
 return(1);
}
