/*
  cc -g -c nntpload.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <utime.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


#include "../s_tools/sregex.h"
#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntpfiler.h"
#include "nntphead.h"
#include "nntpfilter.h"
#include "nntpload.h"



/* --------------------- Nntpload -------------------- */


int Nntpload_new(loadjob,conn,head,dir,filename,flag)
struct NntploaD **loadjob;
struct NntpconN *conn;
struct NntpheaD *head;
char *dir,*filename;
int flag;
/*
 bit0= disable filename_auto
 bit1= set short_header=1
 bit2+3= set multi_binary
 bit4+5= allow yenc encoding (needs command yenc_decode in path)
         0= deny
         1= decode first part of a multi_part 
         2= decode all parts
 bit6+7= set background_1st_part
         0= try to use normal file
         1= request pseudo collision
         2= omit decoding
*/
{
 struct NntploaD *lj;
 int ret;

 *loadjob= lj= TSOB_FELD(struct NntploaD, 1);
 if(lj==NULL)
   return(-1);
 lj->boss= NULL;
 lj->conn= conn;
 lj->head= head;
 lj->head_extern= 1;
 lj->subhead= NULL;
 lj->subhead_extern= 1;
 lj->header_seen= 0;
 lj->announced_size= -1.0;
 lj->article_state= 0;
 lj->allow_yenc= (flag>>4)&3;
 lj->encoding= 0;
 lj->bin_tee= 0;
 lj->show_tee_in_body= 0;
 lj->tee_with_cr= 0;
 lj->ync_tee= 0;
 lj->looks_like_yenc= -1;
 lj->html= 0;
 lj->short_header= !!(flag&2);
 lj->filter= NULL;
 strcpy(lj->dir,dir);
 lj->filename_auto= !(flag&1);
 strcpy(lj->filename,filename);
 lj->bin_chmod= umask(Nntpload_access_rightS);   /* just to ask */
 umask(lj->bin_chmod);                           /* restore old value */
 lj->bin_chmod= 0666 & ~lj->bin_chmod;
 lj->bin_count= 0;
 lj->bin_name[0]= 0;
 lj->tee_name[0]= 0;
 lj->tee_fp= NULL;
 lj->multi_binary= (flag>>2)&3;
 lj->bin_parts= 1;
 lj->current_bin_part= 1;
 lj->background_1st_part= (flag>>6)&3;

 if(lj->head==NULL){
   /* make header object but do not request and read header data from server */
   ret= Nntphead_new(&(lj->head),lj->conn,"",1);
   if(ret<=0)
     goto cancel;
   lj->head_extern= 0;
 }

 return(1);

cancel:
 Nntpload_destroy(loadjob,0);
 return(ret);
}


int Nntpload_destroy(loadjob,flag)
struct NntploaD **loadjob;
int flag;
{
 if((*loadjob)==NULL)
   return(0);
 if((*loadjob)->head_extern==0 && (*loadjob)->head!=NULL)
   Nntphead_destroy(&((*loadjob)->head),0);
 if((*loadjob)->subhead_extern==0 && (*loadjob)->subhead!=NULL)
   Nntpsubhead_destroy(&((*loadjob)->subhead),0);
 if((*loadjob)->tee_fp!=NULL)
   Sfile_fclose((*loadjob)->tee_fp);
 free((char *) *loadjob);
 *loadjob= NULL;
 return(1);
}


int Nntpload_set_boss(loadjob,boss,flag)
struct NntploaD *loadjob;
struct NntpclienT *boss;
int flag;
{
 loadjob->boss= boss;
 return(1);
}


int Nntpload_set_filename(loadjob,filename,flag)
struct NntploaD *loadjob;
char *filename;
int flag;
{
 strcpy(loadjob->filename,filename);
 return(1);
}


int Nntpload_set_filter(loadjob,filter,flag)
struct NntploaD *loadjob;
struct NntpfilteR *filter;
int flag;
{
 loadjob->filter= filter;
 return(1);
}


int Nntpload_set_html(loadjob,html,flag)
struct NntploaD *loadjob;
int html;
int flag;
{
 loadjob->html= html;
 return(1);
}


int Nntpload_result(loadjob,line,flag)
struct NntploaD *loadjob;
char *line;
int flag;
/*
 bit0= do not escape HTML-brackets, omit if not loadjob->html
*/
{
 int ret;
 char *buffertext,buffertext_data[4*Nntpconn_strL];

 if(loadjob->html && !(flag&1)) {
   ret= Nntpfiler_escape_html(buffertext_data,sizeof(buffertext_data),
                              line,0);
   if(ret<=0)
     return(ret);
   buffertext= buffertext_data;
 } else {
   if((flag&1)&&!loadjob->html)
     return(2);
   buffertext= line;
 }
 ret= Nntpconn_result(loadjob->conn,buffertext,"",0,"",8);
 return(ret); 
}


int Nntpload_tee(loadjob,line,flag)
struct NntploaD *loadjob;
char *line;
int flag;
/*
 bit0= do not suppress empty lines and single dot with CR or LF
 bit1= do not write to file, only prepare filename 
 bit2= do not write to file, rename to final name
 bit3= name is .msg.tee rather than .tee
 bit4= use ->tee_fp 
 bit5= return final .tee name in parameter 'line'
 bit6= return .tee name for processed files in 'line'
*/
{
 int ret,nl=0;
 char filename[SfileadrL],*cpt,*wpt,*rpt;
 unsigned long crc;

 if(flag&(32|64)) {
   cpt= strstr(loadjob->tee_name,"bin/N_");
   if(cpt==NULL)
     strcpy(line,loadjob->tee_name);
   else {
     wpt= line;
     for(rpt=loadjob->tee_name;rpt!=cpt+4;rpt++)
       *(wpt++)= *rpt;
     if(flag&64)
       *(wpt++)= 'X';
     for(rpt=cpt+5;*rpt!=0;rpt++)
       *(wpt++)= *rpt;
     *(wpt++)= 0;
   }
   return(1);
 }

 if(loadjob->bin_tee<=0)
   return(2);
 if(flag&4) {
   if(loadjob->tee_name[0]==0 || loadjob->tee_name[1]==0)
     return(2);
   Nntpload_tee(loadjob,filename,32);
   rename(loadjob->tee_name,filename);
   return(1);
 }

 if(loadjob->tee_name[0]==0){
   if(loadjob->tee_fp!=NULL) {
     Sfile_fclose(loadjob->tee_fp);
     loadjob->tee_fp= NULL;
   }
   wpt= filename;
   *(wpt++)= 'N';                                     /* mark as unfinished */
   if(loadjob->head->subject[0]!=' ')
     *(wpt++)= '_';
   for(cpt= loadjob->head->subject;(*cpt)!=0;cpt++) {
     if(strchr("[]()/",*cpt)!=NULL)
       *(wpt++)= '+';
     *(wpt++)= *cpt;
   }
   *wpt= 0;
                      /* crc distinguishies tee files from different posters */
   Sregex_crc32(loadjob->head->from,strlen(loadjob->head->from),&crc,1);
   sprintf(wpt,"_%8.8X",crc);
   Sfile_esc_filename(filename,sizeof(filename)-8,4|8);
   if(flag&8) 
     sprintf(loadjob->tee_name,"bin/%s.msg.tee",filename);
   else if(loadjob->ync_tee)
     sprintf(loadjob->tee_name,"bin/%s.ync.tee",filename);
   else
     sprintf(loadjob->tee_name,"bin/%s.tee",filename);
   unlink(loadjob->tee_name);
 }  
 if(flag&2)
   return(1);
 if(line[0]==0)
   return(1);
 if(!(flag&1)){
   if(line[0]=='.' && (line[1]=='\n' || line[1]=='\r'))
     return(3);
   if(line[0]=='\n' || line[0]=='\r')
     return(3);
 }
 /* append to tee file */
 if((flag&16) || loadjob->tee_fp!=NULL) {
   if(loadjob->tee_fp==NULL) {
     loadjob->tee_fp= Sfile_fopen(loadjob->tee_name,"ab",0);
     if(loadjob->tee_fp==NULL)
       return(-1);
   }
   if(strlen(line)>0)
     ret= fwrite(line,strlen(line),1,loadjob->tee_fp);
 } else {
   ret= Sfile_string_to_file(loadjob->tee_name,line,2);
 }
 return(ret);
}


int Nntpload_write_tee_link(loadjob,flag)
struct NntploaD *loadjob;
int flag;
{
 char final_tee_name[SfileadrL],line[4*SfileadrL],rn[4];

 if(loadjob->bin_tee<=0)
   return(2);
 Nntpload_tee(loadjob,final_tee_name,32);
 if(final_tee_name[0]==0)
   return(2);
 strcpy(rn,"\r\n");
 if(loadjob->show_tee_in_body==1) {
   strcpy(rn,"");
   sprintf(line,"%s%s\r\n",rn,
           "[ message content shown and also downloaded to tee file ]");
 } else {
   sprintf(line,"%s%s\r\n",rn,"[ message content downloaded to tee file ]");
 }
 Nntpload_result(loadjob,line,0);
 if(loadjob->html) {
   sprintf(line,"%s <A HREF=\"%s\">%s</A>\r\n",
                rn,final_tee_name,final_tee_name);
   Nntpload_result(loadjob,line,1);
   Nntpload_tee(loadjob,final_tee_name,64);
   sprintf(line,"%s after -merge_tee : <A HREF=\"%s\">%s</A>\r\n\r\n\r\n",
                rn,final_tee_name,final_tee_name);
   Nntpload_result(loadjob,line,1);
 } else {
   sprintf(line,"%s   %s\r\n\r\n\r\n",rn,final_tee_name);
   Nntpload_result(loadjob,line,0);
 }
 return(1);
}


int Nntpload_getline(loadjob,line,linsiz,flag)
struct NntploaD *loadjob;
char *line;
int linsiz;
int flag;
{
 int ret,i,l;
 ret= Nntpconn_getline(loadjob->conn,line,linsiz,0);
 if(ret<=0)
   return(ret);
 l= strlen(line);
 if(l<2)
   return(1);
 if(line[0]=='.' && line[1]=='.')
   for(i=1;i<=l;i++)
     line[i-1]= line[i];
 return(1);
}


int Nntpload_is_boundary(loadjob,line,flag)
struct NntploaD *loadjob;
char *line;
int flag;
{
 int end_marker= 0;
 char *cpt;

 /* look for boundary string */
 if(line[0]!='-' || line[1]!='-')
   return(0);
 cpt= line+2;
 if(!Sregex_complines(loadjob->head->boundary,cpt,1)) 
   return(0);
 cpt+= strlen(loadjob->head->boundary);
 if(*(cpt)=='-' && *(cpt+1)=='-'){
   cpt+= 2;
   end_marker= 1;
 }
 if(*cpt==0 || *cpt=='\r' || *cpt=='\n')
   return(1+end_marker);
 return(0);
}


int Nntpload_msg_target(loadjob,targetname,targetadr,msgno,flag)
struct NntploaD *loadjob;
char *targetname,*targetadr;
int msgno;
int flag;
/*
 bit0= do not check for existing message
 bit1= do not copy result to targetname
 bit2= do not copy result to targetadr
*/
{
 int ret;
 if(loadjob->filename_auto) {
   ret= Nntpfiler_msg_target(targetname,targetadr,loadjob->dir,msgno,flag&7);
   return(ret);
 }else{
   sprintf(targetadr,"%s/%s",loadjob->dir,targetname);
 }
 return(1);
}


int Nntpload_chdir(loadjob,flag)
struct NntploaD *loadjob;
int flag;
{ 
 int ret;
 ret= Nntpfiler_chdir_primitive(loadjob->dir,&(loadjob->conn->errn),
                                loadjob->conn->errmsg,0);
 return(ret);
}


int Nntpload_test_sfn_loaded(loadjob,flag)
struct NntploaD *loadjob;
int flag;
{
 char *spt;

 if(loadjob->filter==NULL) 
   return(0);
 spt= strrchr(loadjob->bin_name,'/');
 if(spt==NULL) spt= loadjob->bin_name;
 else          spt++;
 if(loadjob->filter->sfrec_name!=NULL) 
   if(strcmp(spt,loadjob->filter->sfrec_name)==0) {
     Nntpfilter_set_sfn_loaded(loadjob->filter,0);
     return(1);
   }
 return(0);
}  


int Nntpload_err_premature(loadjob,flag)
struct NntploaD *loadjob;
int flag;
/*
 bit1= abort on incomplete binary data
*/
{
 char buffertext[4096];

 if(loadjob->conn->errn==0)
   loadjob->conn->errn= -19;
 sprintf(loadjob->conn->errmsg,"Premature end of binary data");
 sprintf(buffertext,"[ ERROR %ld : %s : %s ]\r\n",loadjob->conn->errn,
                    loadjob->bin_name,loadjob->conn->errmsg);
 Nntpload_result(loadjob,buffertext,0);
 sprintf(buffertext,
         "</PRE><CENTER><IMG SRC=\"%s\"></CENTER><PRE>\r\n",
         loadjob->bin_name);
 Nntpload_result(loadjob,buffertext,1);
 if(flag&2){
   return(-1);
 } else {
   Nntpconn_errmsg_reset(loadjob->conn,0);
 }
 return(2);
}


int Nntpload_is_multibin_subject(loadjob,flag)
struct NntploaD *loadjob;
int flag;
/*
 return:
  0= no multibinary
  1= first part of a multibinary
  2= further part of a multibinary
*/
{
 static struct SregeX *re_yenc,*re_multi,*re_multi_first,*re_multi_single;
 static struct SregeX *re_multi_2dg,*re_multi_first_2dg;
 static int initialized= 0;

 char *cpt;
 int ret;

 if(!initialized){
   if(Sregex_new(&re_yenc," [Yy][Ee][Nn][Cc] ",0)<=0){
err_no_regex:;
     loadjob->conn->errn= -3;
     sprintf(loadjob->conn->errmsg,
             "cannot initialize regular expressions for multibinary detection");
     return(-1);
   }
   if(Sregex_new(&re_multi,".*[[(][1-9]/[2-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_first,".*[[(]1/[2-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_single,".*[[(]1/1[])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_2dg,".*[[(][0-9][0-9]/[1-9][0-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_first_2dg,".*[[(]0[01]/[1-9][0-9][])].*",0)<=0)
     goto err_no_regex;
   initialized= 1;
 }

 /*  is this a multiple binary ? */
 if(! loadjob->header_seen )
   return(0);
 if(loadjob->multi_binary<=0)
   return(0);
 ret= 0;
 if( (!Sregex_is_match(re_multi_single,loadjob->head->subject,0)) &&
     ( Sregex_is_match(re_multi,loadjob->head->subject,0) ||
       ( loadjob->multi_binary>=2 && 
         Sregex_is_match(re_multi_2dg,loadjob->head->subject,0) )) ) {
   ret= 2;
   if( Sregex_is_match(re_multi_first,loadjob->head->subject,0) ||
       ( loadjob->multi_binary>=2 &&
         Sregex_is_match(re_multi_first_2dg,loadjob->head->subject,0) ) ) 
     ret= 1;
 }
 /* yEnc multibinaries can be decoded independendly */
 if( Sregex_is_match(re_yenc,loadjob->head->subject,0) 
     && loadjob->allow_yenc==2)
   ret= 0;
 return(ret);
}


int Nntpload_try_mime_multipart(loadjob,line,linsiz,flag)
struct NntploaD *loadjob;
char *line;
int linsiz;
int flag;
/*
 bit1= abort on incomplete binary data
*/
{
 char *rpt,*wpt;
 int l,ret,lin_count,pseudo_collision,omit_decoding;
 FILE *fp= NULL;
 char buffertext[4096]={""}; /* was static until A20427 */
 struct LstrinG *ls;

 Nntpload_tee(loadjob,line,0);
 if(line[0]==0)
   return(0);

test_for_boundary:;
 /* look for boundary string */
 ret= Nntpload_is_boundary(loadjob,line,0);
 if(ret<=0)
   return(ret);
 if(ret==2) {/* end marker found */
   if(loadjob->short_header)
     line[0]= 0;
   Nntphead_boundary_pop(loadjob->head,0);
   return(1);
 }

 if(!loadjob->short_header)
   Nntpload_result(loadjob,line,0);
 line[0]= 0;

 /* read Subheader , write to message file (if not short header format) */
 loadjob->bin_name[0]= 0;
 if(loadjob->subhead_extern==0 && loadjob->subhead!=NULL)
   Nntpsubhead_destroy(&(loadjob->subhead),0);
 if(Nntpsubhead_new(&(loadjob->subhead),loadjob->head,
                    8*!loadjob->short_header)<=0)
   return(-1);

 for(ls= loadjob->subhead->lines; ls!=NULL; ls= ls->next) 
   Nntpload_tee(loadjob,ls->text,0);
 Nntpload_tee(loadjob,"\r\n",1);

 loadjob->subhead_extern= 0;
 if((!loadjob->subhead->is_loadable)||loadjob->encoding==0) 
   return(0);

 /* decode binary data */  
 ret= Nntpload_chdir(loadjob,0);
 if(ret<=0)
   return(ret);

 loadjob->bin_count++;
 sprintf(loadjob->bin_name,"%s",loadjob->subhead->filename);
 if(loadjob->bin_name[0]==0){
   sprintf(loadjob->bin_name,"%s_%ld.bin",loadjob->filename,
                                          loadjob->bin_count);
 }

 if(!Nntpfilter_filename_match(loadjob->filter,loadjob->bin_name,line,0)) {
   while(1){
     ret= Nntpload_getline(loadjob,line,linsiz,0);
     if(ret<=0) 
       goto errn_19;
     lin_count++;
/*
     if(!loadjob->conn->show_io_stat)
       fprintf(stderr,"\r  %ld MIME-base64 lines (skip)\r",lin_count);
*/
     if(Nntpload_is_boundary(loadjob,line,0)) {
       Nntpload_tee(loadjob,line,0);
   break;
     }
   }
   fprintf(stderr,"  %ld MIME-base64 lines skipped\n",lin_count);
   sprintf(buffertext,"[skipped : MIME-base64 file : %s ]\r\n",
                      loadjob->bin_name);
   Nntpload_result(loadjob,buffertext,0);
   return(1);
 }

 if(loadjob->bin_tee>0 && loadjob->background_1st_part==1)
   pseudo_collision= 1;
 else
   pseudo_collision= 0;
 if(loadjob->bin_tee>0 && loadjob->background_1st_part==2)
   omit_decoding= 1;
 else
   omit_decoding= 0;
 Nntpfiler_esc_filename(loadjob->bin_name,sizeof(loadjob->bin_name),1);
 sprintf(buffertext,"%s/%s",Nntpload_bindiR,loadjob->bin_name);
 strcpy(loadjob->bin_name,buffertext);
 Sfile_unique_adr(loadjob->bin_name,!!pseudo_collision);

 if(loadjob->subhead->encoding==2) {
   if(Nntpload_getline(loadjob,line,linsiz,0)>0){
     ret= Nntpload_try_uuencode(loadjob,line,linsiz,1);
     return(ret);
   }
 } else {
   /* open pipe to decoding program redirecting output to filename */
   sprintf(buffertext,Nntpload_mime_decodeR,loadjob->bin_name);

   Nntpconn_show_stat(loadjob->conn,1);
   if(!omit_decoding) {
     fprintf(stderr,"  %s\n",buffertext);
     fp= popen(buffertext,"w");
     if(fp==NULL){
       loadjob->conn->errn= errno;
       sprintf(loadjob->conn->errmsg,"Error %ld on opening pipe to : %s\n%s",
                                     errno,buffertext,strerror(errno));
       return(-1);
     }
   }
   /* write to pipe until boundary string is found at start of line */
   lin_count= 0;
   while(1){
     ret= Nntpload_getline(loadjob,line,linsiz,0);
     if(ret<=0) {
       if(fp!=NULL)
         pclose(fp);
       chmod(loadjob->bin_name,loadjob->bin_chmod);
errn_19:;
       if(loadjob->conn->errn!=-6 && loadjob->conn->errn!=-29 && 
          loadjob->conn->errn!=-38)
         ret= Nntpload_err_premature(loadjob,flag&2);
       return(ret);
     }
     Nntpload_tee(loadjob,line,0);
     if(Nntpload_is_boundary(loadjob,line,0))
   break;

     lin_count++;
     if(!omit_decoding) {
       if(fwrite(line,strlen(line),1,fp)<=0){
         loadjob->conn->errn= errno;
         sprintf(loadjob->conn->errmsg,"Error %ld on writing to pipe : %s\n%s",
                                       errno,buffertext,strerror(errno));
         pclose(fp);
         chmod(loadjob->bin_name,loadjob->bin_chmod);
         return(-1);
       }
     }
   }
   Nntpconn_show_stat(loadjob->conn,1);
   fprintf(stderr,"  %ld MIME-base64 lines read\n",lin_count);
   if(fp!=NULL)
     pclose(fp);
   chmod(loadjob->bin_name,loadjob->bin_chmod);

   /* only enable subjectfilerec if this could not be a multi binary 
      (which is not tee'd with MIME up to now) */
   if(Nntpload_is_multibin_subject(loadjob,0)==0)
     Nntpload_test_sfn_loaded(loadjob,0);

   /* reset state */
 }
 sprintf(buffertext,"[extracted : MIME-base64 file : %s ]\r\n",
                    loadjob->bin_name);
 Nntpload_result(loadjob,buffertext,0);
 sprintf(buffertext,"</PRE><CENTER><IMG SRC=\"%s\"></CENTER><PRE>\r\n",
                    loadjob->bin_name);
 Nntpload_result(loadjob,buffertext,1);
 goto test_for_boundary;


/*
221 10006 <...>
Message-ID: <...>
Date: ...
From: ...
Organization: ...
X-Mailer: Mozilla 2.02 (Win16; I)
MIME-Version: 1.0
Subject: ...
Content-Type: multipart/mixed; boundary="------------6C27281832D7"
Newsgroups: ...
Lines: ...
Path: ...
Xref: ...
NNTP-Posting-Host: ...
.
222 10006 <...>

This is a multi-part message in MIME format.

--------------6C27281832D7
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit

...

--------------6C27281832D7
Content-Type: image/jpeg
Content-Transfer-Encoding: base64
Content-Disposition: inline; filename="..."

/9j/4AAQSkZJRgABAQEASABIAAD/2wBDAA0JCgsKCA0LCgsODg0PEyAVExISEyccHhcgLikx
...
Em2x+0VXRWw9KVKg0VzdfrTs/dpUqX1SO6BzNhTo2NKlTJLH9qC4+7SpU1pNKlSoS//Z
--------------6C27281832D7--

.
*/
 
}


int Nntpload_try_mime_raw(loadjob,line,linsiz,flag)
struct NntploaD *loadjob;
char *line;
int linsiz;
int flag;
/*
*/
{
 char *rpt,*wpt,*cpt;
 int l,ret,lin_count;
 FILE *fp;

 char buffertext[4096];

 if(strchr(line,' ') || strchr(line,'\t'))
   return(0);
 cpt= strchr(line,'\r');
 if(cpt==NULL)
   cpt= line+strlen(line);
 if(cpt-line!=60)
   return(0);

 /* guess filename from subject */
 loadjob->bin_name[0]= 0;
 Nntphead_guess_filename(loadjob->head->subject,loadjob->bin_name,
                         sizeof(loadjob->bin_name),0);

 /* decode binary data */  
 ret= Nntpload_chdir(loadjob,0);
 if(ret<=0)
   return(ret);

 loadjob->bin_count++;
 if(loadjob->bin_name[0]==0){
   sprintf(loadjob->bin_name,"%s_%ld.bin",loadjob->filename,
                                          loadjob->bin_count);
 }

 if(!Nntpfilter_filename_match(loadjob->filter,loadjob->bin_name,line,0)) {
   while(1){
     ret= Nntpload_getline(loadjob,line,linsiz,0);
     if(ret<=0) 
   break;
     lin_count++;
/*
     if(!loadjob->conn->show_io_stat)
       fprintf(stderr,"\r  %ld MIME-base64 lines (skip)\r",lin_count);
*/
   }
   fprintf(stderr,"  %ld MIME-base64 lines skipped\n",lin_count);
   sprintf(buffertext,"[skipped : MIME-base64 file : %s ]\r\n",
                      loadjob->bin_name);
   Nntpload_result(loadjob,buffertext,0);
   if(ret<0)
     return(-1);
   return(1);
 }

 Nntpfiler_esc_filename(loadjob->bin_name,sizeof(loadjob->bin_name),1);
 sprintf(buffertext,"%s/%s",Nntpload_bindiR,loadjob->bin_name);
 strcpy(loadjob->bin_name,buffertext);
 Sfile_unique_adr(loadjob->bin_name,0);

 /* open pipe to decoding program redirecting output to filename */
 sprintf(buffertext,Nntpload_mime_decodeR,loadjob->bin_name);

 if(loadjob->conn->show_io_stat)
   Nntpconn_show_stat(loadjob->conn,1);
 fprintf(stderr,"  %s\n",buffertext);
 fp= popen(buffertext,"w");
 if(fp==NULL){
   loadjob->conn->errn= errno;
   sprintf(loadjob->conn->errmsg,"Error %ld on opening pipe to : %s\n%s",
                                 errno,buffertext,strerror(errno));
   return(-1);
 }

 /* write to pipe until boundary string is found at start of line */
 lin_count= 0;
 while(1){
   if(lin_count>0) /* first line has already been read */
     ret= Nntpload_getline(loadjob,line,linsiz,0);
   else
     ret= 1;
   if(ret<=0)
 break;
   lin_count++;
/*
   if(!loadjob->conn->show_io_stat)
     fprintf(stderr,"\r  %ld MIME-base64 lines\r",lin_count);
*/
   if(fwrite(line,strlen(line),1,fp)<=0){
     loadjob->conn->errn= errno;
     sprintf(loadjob->conn->errmsg,"Error %ld on writing to pipe : %s\n%s",
                                   errno,buffertext,strerror(errno));
     pclose(fp);
     chmod(loadjob->bin_name,loadjob->bin_chmod);
     return(-1);
   }
 }
 fprintf(stderr,"  %ld MIME-base64 lines read\n",lin_count);
 pclose(fp);
 chmod(loadjob->bin_name,loadjob->bin_chmod);

 /* only enable subjectfilerec if this could not be a multibinary 
    (which is not tee'd with MIME up to now) */
 if(Nntpload_is_multibin_subject(loadjob,0)==0)
   Nntpload_test_sfn_loaded(loadjob,0);

 sprintf(buffertext,"[extracted : MIME-base64 file : %s ]\r\n",
                    loadjob->bin_name);
 Nntpload_result(loadjob,buffertext,0);
 sprintf(buffertext,"</PRE><CENTER><IMG SRC=\"%s\"></CENTER><PRE>\r\n",
                    loadjob->bin_name);
 Nntpload_result(loadjob,buffertext,1);
 if(ret<0)
   return(-1);
 return(1);

/*
article 117593
220 117593 <...@...> article                 
Path: ...
From: ...
Newsgroups: ...
Subject: ...
Date: ...
Organization: ...
Lines: ...
Message-ID: ...
NNTP-Posting-Host: ...
Mime-Version: 1.0
Content-Type: image/jpeg; name="..."
Content-Transfer-Encoding: base64
User-Agent: Xnews/03.03.12
Cache-Post-Path: ...
X-Cache: ...
Xref: ...

/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAIBAQEBAQIBAQECAgICAgQDAgIC

.
*/
 
}


/*
     not implemented yet

...
Mime-Version: 1.0
...

Message-ID: <xBoROfuur=qIUlAaj9gBFB5hYiZh@4ax.com>
Content-Type: image/jpeg; name="CQ - Cloning Dance.jpg"
Content-Transfer-Encoding: base64
Content-Disposition: attachment; filename="CQ - Cloning Dance.jpg"

/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAQDAwQDAwQEAwQFBAQFBgoHBgYGBg0JCggKDw0QEA8N

*/


int Nntpload_try_uuencode(loadjob,line,linsiz,flag)
struct NntploaD *loadjob;
char *line;
int linsiz;
int flag;
/*
 bit0= just ignore loadjob->encoding (uuencoded MIME multipart)
 bit1= abort on incomplete binary data
*/
{
 static int initialized= 0,begin_len;
 static struct SregeX *re_filename1,*re_filename2;
 static struct SregeX *re_yenc,*re_yenc_subject;
 static int filename_start_char= 'b',yenc_start_char='=';

#ifdef Nntpload_uudecode_olD
 static struct SregeX *re_multi,*re_multi_first;
 static struct SregeX *re_multi_2dg,*re_multi_first_2dg;
#endif /* Nntpload_uudecode_olD */

 char *cpt,*npt;
 char encodingtext[80];
 char buffertext[4096];
 int lin_count,l,ret,end_found,is_yenc= 0,skip_l,decode_yenc_multipart;
 int ync_tee,pseudo_collision,omit_decoding,first_line;
 FILE *fp;

 /* reject/accept state */
 if(loadjob->encoding==3 && !(flag&1))
   return(0);

 if(loadjob->bin_tee==2){
   if(!loadjob->tee_with_cr) {
     Sregex_trimline(line,2|4);
     if(strlen(line)+1<linsiz)
       strcat(line,"\n");
   }
   Nntpload_tee(loadjob,line,0);
   if(loadjob->show_tee_in_body==1)
     Nntpload_result(loadjob,line,0);
   line[0]= 0;
   return(0);
 }

 if(!initialized){
   if(Sregex_new(&re_filename1,"^begin [0-9][0-9][0-9] .*",0)<=0){
err_no_regex:;
     loadjob->conn->errn= -3;
     sprintf(loadjob->conn->errmsg,
             "cannot initialize regular expressions for uudecode");
     return(-1);
   }
   if(Sregex_new(&re_filename2,"^begin 0[0-9][0-9][0-9] .*",0)<=0)
     goto err_no_regex;
   begin_len= strlen("begin ??? ");
   filename_start_char= 'b';

#ifdef Nntpload_uudecode_olD
   if(Sregex_new(&re_multi,".*[[(][1-9]/[2-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_first,".*[[(]1/[2-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_2dg,".*[[(][0-9][0-9]/[1-9][0-9][])].*",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_multi_first_2dg,".*[[(]0[01]/[1-9][0-9][])].*",0)<=0)
     goto err_no_regex;
#endif /* Nntpload_uudecode_olD */

   if(Sregex_new(&re_yenc_subject," [Yy][Ee][Nn][Cc] ",0)<=0)
     goto err_no_regex;
   if(Sregex_new(&re_yenc,"^=ybegin.* name=.*",0)<=0)
     goto err_no_regex;
   yenc_start_char= '=';
   initialized= 1;
 }

 /*  is this a multiple binary ? */
 if(loadjob->looks_like_yenc<0)
   loadjob->looks_like_yenc= Sregex_is_match(re_yenc_subject,
                                             loadjob->head->subject,0);
 decode_yenc_multipart= loadjob->looks_like_yenc>0 && loadjob->allow_yenc==2;
 loadjob->ync_tee= loadjob->looks_like_yenc>0 && loadjob->allow_yenc==1;

 if(loadjob->header_seen && loadjob->multi_binary && loadjob->bin_tee==0 &&
    !decode_yenc_multipart ) {
   loadjob->bin_tee= -1;                    /* now it is decision time */
   
#ifdef Nntpload_uudecode_olD
   if(Sregex_is_match(re_multi,loadjob->head->subject,0) ||
      ( loadjob->multi_binary>=2 && 
        Sregex_is_match(re_multi_2dg,loadjob->head->subject,0) ) ) {

     /* >>> automatically extract encoded text part. */
     /* >>> are all parts available on disc ? */
        /* >>> concatenate and decode parts */

     /* >>> preliminary : tee binary data as plain text */
     loadjob->bin_tee= 1;
     if(!( Sregex_is_match(re_multi_first,loadjob->head->subject,0) ||
           ( loadjob->multi_binary>=2 &&
             Sregex_is_match(re_multi_first_2dg,loadjob->head->subject,0) ))) {

#else /* Nntpload_uudecode_olD */

   ret= Nntpload_is_multibin_subject(loadjob,0);
   if(ret<0)
     return(ret);
   if(ret>0) {
     loadjob->bin_tee= 1;
     if(ret==2) {

#endif /* ! Nntpload_uudecode_olD */

       loadjob->bin_tee= 2;
       if(loadjob->head->byte_count>0) {
         if(loadjob->head->byte_count<Nntpload_show_tee_byte_limiT)
           loadjob->show_tee_in_body= 1;
       } else {
         if(loadjob->head->line_count>0 &&
            loadjob->head->line_count<Nntpload_show_tee_line_limiT)
           loadjob->show_tee_in_body= 1;
       }
       if(!loadjob->tee_with_cr) {
         Sregex_trimline(line,2|4);
         strcat(line,"\n");
       }
       Nntpload_tee(loadjob,line,0);
       Nntpload_write_tee_link(loadjob,0);
       if(loadjob->show_tee_in_body==1)
         Nntpload_result(loadjob,line,0);
       line[0]= 0;
       return(2);
     }
   }
 }

 /* check for  begin ??? filename */
 skip_l= begin_len;
 strcpy(encodingtext,"uuencoded");
 if(line[0]!=filename_start_char ||
    !(Sregex_is_match(re_filename1,line,0) ||
      Sregex_is_match(re_filename2,line,0))) {
   if(line[0]==yenc_start_char && Sregex_is_match(re_yenc,line,0)) {
     /* yEnc (nearly) 8-Bit encoding */;
     is_yenc= 1;
     strcpy(encodingtext,"yencoded");
     npt= strstr(line," name=");
     if(npt!=NULL)
       npt+= 6;
     else
       npt= line;
     skip_l= npt-line;
     if(!loadjob->allow_yenc)
       return(0);
   } else
     return(0);
 }

 ret= Nntpload_chdir(loadjob,0);
 if(ret<=0)
   return(ret);

 /* get filename */
 if(strlen(line)-skip_l>=sizeof(loadjob->bin_name)/2) {
   loadjob->conn->errn= -4;
   sprintf(loadjob->conn->errmsg,
           "file name too long at %s attachment\n%s",encodingtext,line);
   return(-1);
 }

 cpt= strrchr(line+skip_l,'/');
 if(cpt==NULL || cpt<line+skip_l) 
   cpt= line+skip_l;
 else
   cpt++;
 strcpy(loadjob->bin_name,cpt);
 Sregex_trimline(loadjob->bin_name,8);

 if(loadjob->bin_name[0]==0)
   sprintf(loadjob->bin_name,"%s.bin",loadjob->filename);

 if(!Nntpfilter_filename_match(loadjob->filter,loadjob->bin_name,line,0)) {
   while(1){
     ret= Nntpload_getline(loadjob,line,linsiz,0);
     if(ret<=0)
       goto errn_19;
     lin_count++;
/*
     if(!loadjob->conn->show_io_stat)
       fprintf(stderr,"\r  %ld %s lines (skip)\r",lin_count,encodingtext);
*/
     if(Sregex_complines("end",line,0)) {
       Nntpload_tee(loadjob,line,0);
   break;
     }
   }
   fprintf(stderr,"  %ld %s lines skipped\n",lin_count,encodingtext);
   sprintf(buffertext,"[skipped : %s file : %s ]\r\n",
                      encodingtext,loadjob->bin_name);
   Nntpload_result(loadjob,buffertext,0);
   return(1);
 }
 if(loadjob->bin_tee>0 && loadjob->background_1st_part==1)
   pseudo_collision= 1;
 else
   pseudo_collision= 0;
 if(loadjob->bin_tee>0 && loadjob->background_1st_part==2)
   omit_decoding= 1;
 else
   omit_decoding= 0;
 Nntpfiler_esc_filename(loadjob->bin_name,sizeof(loadjob->bin_name),1);
 sprintf(buffertext,"%s/%s",Nntpload_bindiR,loadjob->bin_name);
 strcpy(loadjob->bin_name,buffertext);
 if(!decode_yenc_multipart)
   Sfile_unique_adr(loadjob->bin_name,!!pseudo_collision);

 /* open pipe to decoding program redirecting output to filename */

 if(is_yenc)
   sprintf(buffertext,Nntpload_yenc_decodeR,loadjob->bin_name);
 else
   sprintf(buffertext,Nntpload_uuencode_decodeR,loadjob->bin_name);

 Nntpconn_show_stat(loadjob->conn,1);
 if(!omit_decoding) {
   fprintf(stderr,"  %s\n",buffertext);
   fp= popen(buffertext,"w");
   if(fp==NULL){
     loadjob->conn->errn= errno;
     sprintf(loadjob->conn->errmsg,"Error %ld on opening pipe to : %s\n%s",
                                    errno,buffertext,strerror(errno));
     if(is_yenc) {
check_for_syenc:;
       if(strncmp(Nntpload_yenc_decodeR,"syenc_decode ",13)==0)
         if(system("syenc_decode -v")!=0) {
           sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
           sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "\nThere may be a problem with reaching program  syenc_decode\n");
           sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "Either disable yEnc decoding by :  -allow_yenc off \n");
           sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "or make sure that shell command  syenc_decode -v  works.\n");
           sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n");
         }
     }
     return(-1);
   }
 }

 if(loadjob->tee_with_cr)
   Nntpload_tee(loadjob,line,0);
 l= strlen(line);
 if(l>1)                /* remove CarriageReturn from end of line */
   if(line[l-2]=='\r'){
     line[l-2]='\n';
     line[--l]= 0;
   }
 first_line= 1;
 if(!omit_decoding) 
   if(fwrite(line,strlen(line),1,fp)<=0){
write_err:;
     loadjob->conn->errn= errno;
     sprintf(loadjob->conn->errmsg,
             "Error %ld on opening or writing pipe to : %s\n%s",
             errno,buffertext,strerror(errno));
     pclose(fp);
     if(is_yenc) 
       goto check_for_syenc;
     chmod(loadjob->bin_name,loadjob->bin_chmod);
     return(-1);
   }
 if(!loadjob->tee_with_cr)
   Nntpload_tee(loadjob,line,0);
 first_line= 0;

 /* write to pipe up to 'end' line */
 lin_count= 0;
 end_found= 0;
 while(1){
   ret= Nntpload_getline(loadjob,line,linsiz,0);
   if(ret<=0) {
     if(!end_found) {
       fprintf(stderr,"+++ incomplete %s file. Adding end mark\n",
                      encodingtext,lin_count);
       strcpy(line,"`\nend\n");
       if(!omit_decoding) 
         if(fwrite(line,strlen(line),1,fp)<=0)
           goto write_err;
     }
     if(!omit_decoding)
       pclose(fp);
     chmod(loadjob->bin_name,loadjob->bin_chmod);
errn_19:;
     if(loadjob->conn->errn!=-6 && loadjob->conn->errn!=-29 && 
        loadjob->conn->errn!=-38)
       ret= Nntpload_err_premature(loadjob,flag&2);
     return(ret);
   }
   lin_count++;
/*
   if(!loadjob->conn->show_io_stat)
     fprintf(stderr,"\r  %ld %s lines\r",lin_count,encodingtext);
*/

   if(loadjob->tee_with_cr)
     Nntpload_tee(loadjob,line,0);
   l= strlen(line);
   if(l>1)                /* remove CarriageReturn from end of line */
     if(line[l-2]=='\r'){
       line[l-2]='\n';
       line[--l]= 0;
     }
   if(!omit_decoding) 
     if(fwrite(line,strlen(line),1,fp)<=0)
       goto write_err;
   if(!loadjob->tee_with_cr)
     Nntpload_tee(loadjob,line,0);
   if(is_yenc) {
     if(Sregex_complines("=yend ",line,1)) {
       end_found= 1;
 break;
     }
   } else {
     if(Sregex_complines("end",line,0)) {
       end_found= 1;
 break;
     }
   }
 }
 Nntpconn_show_stat(loadjob->conn,1);
 fprintf(stderr,"  %ld %s lines read\n",lin_count,encodingtext);
 if(!end_found){
   fprintf(stderr,"+++ incomplete %s file. Adding end mark\n",
                  encodingtext,lin_count);
   strcpy(line,"`\nend\n");
   if(!omit_decoding) 
     if(fwrite(line,strlen(line),1,fp)<=0)
       goto write_err;
 }
 if(!omit_decoding)
   pclose(fp);
 chmod(loadjob->bin_name,loadjob->bin_chmod);
 Nntpload_test_sfn_loaded(loadjob,0);

 sprintf(line,"[extracted : %s file : %s ]\r\n",encodingtext,loadjob->bin_name);
 if(Nntpload_result(loadjob,line,0)<=0)
   return(-1);
 sprintf(line,"</PRE><CENTER><IMG SRC=\"%s\"></CENTER><PRE>\r\n",
              loadjob->bin_name);
 if(Nntpload_result(loadjob,line,1)<=0)
   return(-1);
 line[0]= 0;
 return(1);

/*
221 10009 <...>
From: ...
Subject: ...
Post-Count: ...
Message-ID: ...
Newsgroups: ...
Date: ...
Lines: ...
Path: ...
Xref: ...
NNTP-Posting-Host: ...
.
222 10009 <...>



BEGIN --- CUT HERE --- Cut Here --- cut here --- ....jpg

begin 644 ....jpg
M_]C_X``02D9)1@`!`0$`2`!(``#__@`F1FEL92!W<FET=&5N(&)Y($%D;V)E
...
E!&DB,WF]FB-QFHYLK6KR3+<B.45@W*QU'._$"WXH'37/\Q__V5C!
`
end
END --- CUT HERE --- Cut Here --- cut here --- ....jpg

.


*/
 return(0);
}


int Nntpload_try_bin(loadjob,line,linsiz,flag)
struct NntploaD *loadjob;
char *line;
int linsiz;
int flag;
/*
 bit0= abort on incomplete binary data
*/
{
 int ret;
 char buffertext[80];

 if(loadjob->head->is_mime && loadjob->bin_tee==0 && loadjob->multi_binary>0) {
   if(Nntpload_is_multibin_subject(loadjob,0)) {
     /* write msgno and all message headers */
     loadjob->bin_tee= 2;
     loadjob->tee_with_cr= 1;
     sprintf(buffertext,"X-Snntpbatch-Dl-Msgno: %ld\r\n",loadjob->head->msgno);
     Nntpload_tee(loadjob,buffertext,8|16);
     Nntpload_header_result(loadjob,1|2);
     Nntpload_tee(loadjob,"\r\n",1);
   } else {
     loadjob->bin_tee= -1;
   }
 }
 if(loadjob->head->is_mime_multipart) {
   loadjob->encoding= 1;
   ret= Nntpload_try_mime_multipart(loadjob,line,linsiz,(flag&1)<<1);
   if(ret<0)
     return(ret);
   if(loadjob->short_header){
     if(loadjob->subhead==NULL)
       line[0]= 0;
     else if(loadjob->subhead->status!=2)
       line[0]= 0;
   }
 }else{
   if(loadjob->head->is_mime) {
     /* look for boundary definition header within message body */;
     ret= Nntphead_check_for_mime_boundary(loadjob->head,line,0);
     if(ret>0) {
       Nntpload_tee(loadjob,line,0);
       return(1);
     }
   }
   ret= Nntpload_try_uuencode(loadjob,line,linsiz,(flag&1)<<1);
   if(ret<0) 
     return(ret);
   if(ret==0 && loadjob->head->is_mime) {
     if(loadjob->head->is_mime==2 && !loadjob->head->is_mime_multipart) {
       ret= Nntpload_try_mime_raw(loadjob,line,linsiz,0);
       loadjob->head->is_mime= 0; /* only one try, when enabled */
     }
     if(strcmp(line,"\r\n")==0)
       loadjob->head->is_mime= 2; /* enabled */
     if(ret<0) 
       return(ret);
   }
   if(ret==2)
     return(2);                                 /* HTML-Message generated */
 }
fertig:;
 return(1);
}


int Nntpload_header_result(loadjob,flag)
struct NntploaD *loadjob;
int flag;
/*
 bit0= output complete header
 bit1= (with bit0) output via Nntpload_tee()
*/
{
 char text[Nntphead_strL],ref[Nntphead_strL];
 char msgadr[SfileadrL],msgname[SfileadrL];
 int i,ret;
 struct LstrinG *ls;

 if(flag&1)
   goto show_full_header;

 sprintf(text,"msg-id : %s  (#%ld)\r\n",
              loadjob->head->msgid,loadjob->head->msgno);
 if((ret= Nntpload_result(loadjob,text,0))<=0)
   return(ret);
 sprintf(text,"from   : %s\r\n",loadjob->head->from);
 if((ret= Nntpload_result(loadjob,text,0))<=0)
   return(ret);
 sprintf(text,"date   : %s\r\n",loadjob->head->date);
 if((ret= Nntpload_result(loadjob,text,0))<=0)
   return(ret);
 if((ret= Nntpload_result(loadjob,"subject: ",0))<=0)
   return(ret);
 if((ret= Nntpload_result(loadjob,"<B>",1))<=0)
   return(ret);
 sprintf(text,"%s\r\n",loadjob->head->subject);
 if((ret= Nntpload_result(loadjob,text,0))<=0)
   return(ret);
 if((ret= Nntpload_result(loadjob,"</B>",1))<=0)
   return(ret);
 if(loadjob->head->reference[0]!=0){
   if((ret= Nntpload_result(loadjob,"refers : ",0))<=0)
     return(ret);
   if(loadjob->html){
     if((ret= Nntpfiler_escape_html(ref,sizeof(ref),
                                    loadjob->head->reference,0))<=0)
       return(ret);
     /* get msgname */
     ret= Nntpload_id_to_msgadr(loadjob,msgadr,msgname,0);
     if(ret<=0)
       return(ret);
     if(ret==1)
       sprintf(text,"<A HREF=%s>%s</A>\r\n",msgname,ref);
     else
       sprintf(text," %s\r\n",ref);
     if((ret= Nntpload_result(loadjob,text,1))<=0)
       return(ret);
   } else {
     sprintf(text,"%s\r\n",loadjob->head->reference);
     if((ret= Nntpload_result(loadjob,text,0))<=0)
       return(ret);
   }
 }
 return(1);
show_full_header:;
 if((flag&2) || !loadjob->short_header) {
   for(ls= loadjob->head->lines; ls!=NULL; ls= ls->next) {
     if(flag&2) {
       Nntpload_tee(loadjob,ls->text,0);
     } else {
       if((ret= Nntpload_result(loadjob,ls->text,0))<=0)
         return(ret);
     }
   }
 }
 return(1);
}


int Nntpload_article(loadjob,flag)
struct NntploaD *loadjob;
int flag;
/*
 bit0= abort on incomplete binary data
 bit1= do not see invalid article as error
 bit2= this is a merge run (no _transaction() , no previous message)
 bit3= conn: do not reset transaction count
*/
{
 struct NntpconN *conn;
 char *dir,*filename,*buffer,*cpt;
 char fileadr[4096],line[4096],cmd[80],links[321],prevname[321],nextname[321];
 char listname[321],count_text[81];
 int ret,buff_siz,artstat,html_closed=1,bytes_shown,byte_count_merk,prevno;
 FILE *fp= NULL;
 struct stat stbuf;

 /* remove pending popen()-zombies */
 wait3(NULL,WNOHANG,NULL);

 conn= loadjob->conn;
 dir= loadjob->dir;
 filename= loadjob->filename;
 buffer= loadjob->conn->buffer;
 buff_siz= sizeof(loadjob->conn->buffer);

 if(loadjob->head->status==0){
   Nntphead_set_msgadr(loadjob->head,-1,"",0);
   ret= Nntphead_read(loadjob->head,0);
   if(ret<=0)
     return(ret);
 }

 /* create message result file */
 ret= Nntpload_msg_target(loadjob,filename,fileadr,loadjob->head->msgno,1);
 if(ret<=0)
   return(0);
 fp= fopen(fileadr,"w");
 if(fp==NULL){
   conn->errn= errno;
   sprintf(conn->errmsg,"Error %ld on opening output file : %s\n%s",
                        errno,fileadr,strerror(errno));
   return(0);
 }
 Nntpconn_set_outlet(conn,fp,fileadr,0); /* for subordinate readers */

 if(!(flag&4)) {
   /* Adjust links in previously downloaded article */
   if(loadjob->html && loadjob->filename_auto &&
      loadjob->filter->msgno_prev[0]>0){
     if(loadjob->filter->msgno_prev[1]>0){
       ret= Nntpload_msg_target(loadjob,prevname,line,
                                loadjob->filter->msgno_prev[1],1|4);
     }else{
       ret= Nntpload_msg_target(loadjob,prevname,line,
                                loadjob->filter->msgno_prev[0]-1,1|4);
     }
     if(ret<=0)
       {ret= 0; goto ex;}

     Nntpfiler_det_indexfile(loadjob->conn->filer,
                             loadjob->filter->msgno_prev[0],listname,0);
     Nntpfiler_adjust_mlinks(loadjob->conn->filer,
                             loadjob->filter->msgno_prev[0],listname,prevname,
                             filename,0);
   }

   /* require complete article (do not read reply data yet) */
   loadjob->article_state= 0;
   if(loadjob->head->msgno>0)
     sprintf(cmd,"ARTICLE %ld",loadjob->head->msgno);
   else
     sprintf(cmd,"ARTICLE %s",loadjob->head->msgid);
   artstat= Nntpconn_transaction(conn,cmd,buffer,buff_siz,"",2|(flag&8)|1024);
   if(artstat<=0)
     {ret= artstat; goto ex;}
   if((!(flag&2)) && conn->status_code!=220) {
     conn->errn= -27;
     sprintf(conn->errmsg,"request refused : %s\nstatus reply    : %s",
                          cmd,conn->status_line);
     {ret= 0; goto ex;}
   }
 }

 byte_count_merk= loadjob->head->byte_count;
 ret= Nntphead_read(loadjob->head,1|(flag&2));
 if(ret<=0)
   goto ex;
 loadjob->header_seen= 1;
 loadjob->head->byte_count= byte_count_merk;

 if(loadjob->filter==NULL)
   prevno= loadjob->head->msgno-1;
 else
   prevno= loadjob->filter->msgno_prev[0];
 ret= Nntpload_msg_target(loadjob,prevname,line,prevno,1|4);
 if(ret<=0)
   {ret= 0; goto ex;}
 ret= Nntpload_msg_target(loadjob,nextname,line,loadjob->head->msgno+1,1|4);
 if(ret<=0)
   {ret= 0; goto ex;}
 Nntpfiler_det_indexfile(loadjob->conn->filer,loadjob->head->msgno,listname,0);
 ret= Nntpfiler_format_linkline(loadjob->conn->filer,loadjob->head->msgno,
                                links,listname,filename,prevname,nextname,0);
 if(ret<=0)
   {ret= 0; goto ex;}

 if(loadjob->conn->show_io_stat)
   Nntpconn_show_stat(loadjob->conn,1);
 if(!(flag&4)) 
   fprintf(stderr,"Article size : ");
 bytes_shown= 0;
 loadjob->announced_size= -1.0;
 if(Nntphead_get_property(loadjob->head,"bytes:",
                          count_text,sizeof(count_text),0)>0)
   if(count_text[0]!=0 && count_text[0]!='?') {
     fprintf(stderr,"%s in ",count_text);
     bytes_shown= 1;
     sscanf(count_text,"%lf",&(loadjob->announced_size));
   }
 if(Nntphead_get_property(loadjob->head,"lines:",
                          count_text,sizeof(count_text),0)>0) {
   fprintf(stderr,"%s\n",count_text);
 }
 /* process lines of article */ 
 sprintf(line,"<HTML><TITLE>%s</TITLE><BODY><PRE>\r\n",filename);
 if(Nntpload_result(loadjob,line,1)<=0)
   {ret= 0; goto ex;}
 html_closed= 0;
 if(Nntpload_result(loadjob,links,1)<=0)
   {ret= 0; goto ex;}

 if(loadjob->header_seen) {
   if(Nntpload_header_result(loadjob,0)<=0)
     {ret= 0; goto ex;}
   if(Nntpload_result(loadjob,"<HR>",1)<=0)
     {ret= 0; goto ex;}
 }
 
 while(Nntpload_getline(loadjob,line,sizeof(line),0)>0) {
   /* look for binary data */
   ret= Nntpload_try_bin(loadjob,line,sizeof(line),flag&1);
   if(ret<=0)
     goto ex;
   /* put out textual line (not necessarily the one read above) */
   if(loadjob->short_header&&!loadjob->header_seen)
 continue;
   if(strlen(line)) {
     if(Nntpload_result(loadjob,line,(ret==2))<=0)
       {ret= 0; goto ex;}
   }
 }
 if(loadjob->head->is_mime)
   Nntpload_write_tee_link(loadjob,0);
 if(loadjob->short_header) {
   if(!loadjob->header_seen)
     if(Nntpload_header_result(loadjob,0)<=0)
       {ret= 0; goto ex;}
 } else {
   if(Nntpload_result(loadjob,"<HR>",1)<=0)
     {ret= 0; goto ex;}
   if(Nntpload_header_result(loadjob,1)<=0)
     {ret= 0; goto ex;}
   if(Nntpload_result(loadjob,"<HR>\r\n",1)<=0)
     {ret= 0; goto ex;}
 }
 if(Nntpload_result(loadjob,links,1)<=0)
   {ret= 0; goto ex;}
 if(Nntpload_result(loadjob,"<HR>\r\nReply list :<BR>\r\n",1)<=0)
   {ret= 0; goto ex;}
 if(Nntpload_result(loadjob,"</PRE></BODY></HTML>\r\n",1)<=0)
   {ret= 0; goto ex;}
 html_closed= 1;

 ret= 1;
ex:;
 Nntpload_tee(loadjob,"",4);       /* close tee file resp. remove open mark */
 if(fp!=NULL) {
   if(!html_closed)
     Nntpload_result(loadjob,"</PRE></BODY></HTML>\r\n",1);
   fclose(fp);
   strcpy(line,fileadr);
   cpt= strrchr(line,'/');
   if(cpt==NULL) cpt= line;
   else          cpt++;
   strcpy(cpt,"msg_new.html");
   if(stat(line,&stbuf)==-1)
     Sfile_copy_file(fileadr,line,
                     &(loadjob->conn->errn),loadjob->conn->errmsg,0);
 }
 Nntpconn_set_outlet(conn,NULL,"",0); /* for subordinate readers */

 if (ret>0 && loadjob->announced_size>1024) {
   if(loadjob->announced_size>loadjob->conn->t_bytecount) {
     loadjob->conn->errn= -37;
     sprintf(loadjob->conn->errmsg,
             "Server returns much less article bytes than announced :\n");
     sprintf(loadjob->conn->errmsg+strlen(loadjob->conn->errmsg),
             "  expected: more than %.lf  , received: %.lf\n",
             loadjob->announced_size,loadjob->conn->t_bytecount);
     ret= 0;
   }
 }

 return(ret);
}


int Nntpload_id_to_msgadr(loadjob,msgadr,msgname,flag)
struct NntploaD *loadjob;
char *msgadr,*msgname;
int flag;
{
 int ret;

 if(loadjob->conn->filer==NULL)
   return(4);
 ret= Nntpfiler_id_to_msgadr(loadjob->conn->filer,loadjob->head->reference,
                             msgadr,msgname,0);
 return(ret);
}


