/*
  cc -g -c nntpnov.c
*/


#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#include "nntpconn.h"
#include "nntpgroup.h"
#include "nntphead.h"
#include "nntpnov.h"


/* -------------------------------- NntpnoV ------------------------------ */


int Nntpnov_new(nov,conn,start,end,flag)
struct NntpnoV **nov;
struct NntpconN *conn;
int start,end;
int flag;
{
 struct NntpnoV *n;
 int i;

 *nov= n= TSOB_FELD(struct NntpnoV,1);
 if(n==NULL)
   return(-1);
 n->conn= conn;
 n->try_xpat= 1;
 n->range_start= start;
 n->range_end= end;
 n->line_max= end-start;
 if(n->line_max<0)
   return(0);
 n->lines= TSOB_FELD(char *,n->line_max+1);
 if(n->lines==NULL) {
   Nntpnov_destroy(nov,0);
   return(-1);
 }
 for(i=0;i<=n->line_max;i++)
   n->lines[i]= NULL;
 return(1);
}


int Nntpnov_destroy(nov,flag)
struct NntpnoV **nov;
int flag;
{
 int i;
 char **lines;

 if(nov==NULL)
   return(0);
 lines= (*nov)->lines;
 if(lines!=NULL) 
   for(i=0;i<=(*nov)->line_max;i++)
     if(lines[i]!=NULL)
       free(lines[i]);
 free((char *) *nov);
 *nov= NULL;
 return(0);
}


int Nntpnov_set_line(nov,line,flag)
struct NntpnoV *nov;
char *line;
int flag;
{
 int wpt;

 wpt= -1;
 sscanf(line,"%ld",&wpt);
 wpt-= nov->range_start;
 if(wpt<0 || wpt>nov->line_max)
   return(0);
 if(nov->lines[wpt]!=NULL)
   free(nov->lines[wpt]);
 nov->lines[wpt]= TSOB_FELD(char,strlen(line)+1);
 if(nov->lines[wpt]==NULL) {
   nov->conn->errn= -17;
   sprintf(nov->conn->errmsg,
           "Not enough memory to allocate News OverView structure");
   return(-1);
 }
 strcpy(nov->lines[wpt],line);
 return(1);
}


int Nntpnov_read(nov,flag)
struct NntpnoV *nov;
int flag;
/*
 bit0= only Date, Bytes and Lines are needed
*/
{
 static int bd_trig= 100;
 int count= 0,ret,nov_stat,wpt;
 char cmd[321],line[Nntpconn_strL];

 if(nov==NULL)
   return(0);

 if((flag&1) && nov->try_xpat && nov->range_end-nov->range_start>bd_trig) {
   ret= Nntpnov_read_bytes_date(nov,flag);
   if(ret<=0)
     nov->try_xpat= 0; /* do it with XOVER */
   else
     return(ret);
 }

 sprintf(cmd,"XOVER %ld-%ld",nov->range_start,nov->range_end);
 nov_stat= Nntpconn_transaction(nov->conn,cmd,nov->conn->buffer,
                                sizeof(nov->conn->buffer),"",2|1024);
 if(nov_stat<=0)
   {ret= nov_stat; goto ex;}

 while(1) {
   ret= Nntpconn_getline(nov->conn,line,sizeof(line),0);
   if(ret<0)
     goto ex;
   if(ret==0) {
     if(Nntpconn_transaction_is_completed(nov->conn,0))
 break;
     goto ex;
   }
   ret= Nntpnov_set_line(nov,line,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;
   count++;
   if(!nov->conn->show_io_stat)
     fprintf(stderr,"\rXOVER : %ld lines read ",count);
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpnov_read_bytes_date(nov,flag)
struct NntpnoV *nov;
int flag;
{
 int count= 0,ret,nov_stat,wpt,pass,i,tcount;
 char cmd[321],line[Nntpconn_strL],header[81],*cpt,buffer[2*Nntpconn_strL+100];

 if(nov==NULL)
   return(0);

 for(pass=0;pass<3;pass++){
   if(pass==0)       strcpy(header,"Date");
   else if(pass==1)  strcpy(header,"Bytes");
   else              strcpy(header,"Lines");
   sprintf(cmd,"XPAT %s %ld-%ld *",header,nov->range_start,nov->range_end);
   nov_stat= Nntpconn_transaction(nov->conn,cmd,nov->conn->buffer,
                                  sizeof(nov->conn->buffer),"",
                                  2|1024|((pass>0)<<3));
   if(nov_stat<=0)
     {ret= nov_stat; goto ex;}
   if(nov->conn->status_code!=221) {
     nov->conn->errn= -32;
     sprintf(nov->conn->errmsg,"= XPAT command not supported by server");
     ret= 0; goto ex;
   }

   while(Nntpconn_getline(nov->conn,line,sizeof(line),0)>0) { 
     wpt= -1;
     sscanf(line,"%ld",&wpt);
     wpt-= nov->range_start;
     if(wpt<0 || wpt>nov->line_max)
   continue;
     if(pass==0) {
       if(nov->lines[wpt]!=NULL){
         free(nov->lines[wpt]);
         nov->lines[wpt]= NULL;
       }
       cpt= strchr(line,' ');
       if(cpt==NULL)
   continue;

       *cpt= 0;
       sprintf(buffer,"%s\t-\tx@y\t%s\t<#%s@y>\t",line,cpt+1,line);
       if(Sregex_set_dyn_string(&(nov->lines[wpt]),buffer,0)<=0){
no_dyn_string:;
         nov->conn->errn= -17;
         sprintf(nov->conn->errmsg,
               "Not enough memory to allocate News OverView structure");
         ret= -1;
         goto ex;
       }
       count++;
       if(!nov->conn->show_io_stat)
         fprintf(stderr,"\rXPAT : %ld lines read ",count);
     } else {
       if(strlen(line)>=Nntpconn_strL/2){
          fprintf(stderr,"\noversized %s header\n",header);
          fprintf(stderr,"'%s'\n",line);
       } 
       if(nov->lines[wpt]!=NULL){
         cpt= strchr(line,' ');
         if(cpt!=NULL){
           sprintf(buffer,"\t%s",cpt+1);
           if(Sregex_set_dyn_string(&(nov->lines[wpt]),buffer,1)<=0)
             goto no_dyn_string;
         }
       }
     }
   }
 }
 /* check for incomplete lines */
 for(i=0;i<=nov->range_end-nov->range_start;i++){
   tcount= 0;
   if(nov->lines[i]==NULL)
 continue;
   for(cpt=nov->lines[i];(*cpt)!=0;cpt++)
     if((*cpt)=='\t') tcount++;
   if(tcount<7) {
     free(nov->lines[i]);
     nov->lines[i]= NULL;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpnov_read_file(nov,fp,flag)
struct NntpnoV *nov;
FILE *fp;
int flag;
{
 int ret,count= 0;
 char line[Nntpconn_strL];

 if(nov==NULL)
   return(0);

 while(fgets(line,sizeof(line),fp)!=NULL){
   ret= Nntpnov_set_line(nov,line,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;
   count++;
   if(!nov->conn->show_io_stat)
     fprintf(stderr,"\rXOVER : %ld lines read ",count);
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpnov_write_file(nov,fp,flag)
struct NntpnoV *nov;
FILE *fp;
int flag;
{
 int i,ret;
 char *wpt;

 if(nov==NULL)
   return(0);

 for(i=nov->range_start;i<=nov->range_end;i++) {
   wpt= nov->lines[i-nov->range_start];
   if(wpt!=NULL)
     fprintf(fp,"%s",wpt);
 }  
 return(1);
}


int Nntpnov_fill_head(nov,head,msgno,flag)
struct NntpnoV *nov;
struct NntpheaD *head;
int flag;
{
 static char fields[][81]= {"msgno:","subject:","from:","date:",
                            "msgid:","references:","bytes:","lines:",
                            "@END@"};
 int fpt,ept_mem;
 char *spt,*ept;

 Nntphead_reset(head,1);
 if(msgno<nov->range_start||msgno>nov->range_end)
   return(0);
 spt= nov->lines[msgno-nov->range_start];
 if(spt==NULL)
   return(0);

 for(fpt= 0;strcmp(fields[fpt],"@END@")!=0;fpt++){
   ept= strchr(spt,'\t');
   if(ept==NULL) 
     ept= spt+strlen(spt);
   ept_mem= *ept;
   *ept= 0;
   Nntphead_set_property(head,fields[fpt],spt,0);
   *ept= ept_mem;
   if((*ept)!=0)
     spt= ept+1;
 }
 return(1);
}


