
/*
 cc -g -c sdb_odbc.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "sdb_odbc_stq.h"

#include "sdb_odbc.h"


static char trace_linE[1024];


/* --------------------- SDB wrapper for ODBC ------------------ */


int Sdb_odbc_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_odbC **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 struct Sdb_odbC *m;

 m= *sdb= TSOB_FELD(struct Sdb_odbC,1);
 if(m==NULL)
   return(-1);

 m->hostname= NULL;
 m->username= NULL;
 m->password= NULL;
 m->dbname= NULL;
 m->oret= SQL_SUCCESS;
 m->initialized= 0;
 m->opened= 0;
 m->auto_commit= 1;
 m->fetch_mode= 0;
 m->stq= NULL;
 m->message= message_pt;
 m->debug_mode= 0;

 if(Sregex_string(&(m->hostname),hostname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->username),username,0)<=0)
   goto failed;
 if(Sregex_string(&(m->password),password,0)<=0)
   goto failed;
 if(Sregex_string(&(m->dbname),dbname,0)<=0)
   goto failed;
 return(1);

failed:;
 Sdb_odbc_destroy(sdb,0);
 return(-1);
}


int Sdb_odbc_destroy(sdb,flag)
struct Sdb_odbC **sdb;
int flag;
{
 struct Sdb_odbC *m;

 m= *sdb;
 if(m==NULL)
   return(0);
 Sdb_odbc_close(m,0);
 Sregex_string(&(m->dbname),NULL,0);
 Sregex_string(&(m->password),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->hostname),NULL,0);
 Smem_freE((char *) m);
 *sdb= NULL;
 return(1);
}


int Sdb_odbc_trace(sdb,msg,flag)
struct Sdb_odbC *sdb;
char *msg;
int flag;
{
 if(sdb->debug_mode<1)
   return(2);
 if(flag&1)
   if(sdb->debug_mode<2)
     return(2);
 fprintf(stderr,"%s\n",msg);
 fflush(stderr);
 return(1);
}


int Sdb_odbc_checkerr(sdb,in_oret,errmsg,flag)
struct Sdb_odbC *sdb;
SQLRETURN in_oret;
char errmsg[Sdb_odbc_min_msg_sizE];
int flag;
{
 int ret= 1,i;
 SQLRETURN oret;
 SQLCHAR sql_state[6],sql_msg[SQL_MAX_MESSAGE_LENGTH];
 SQLINTEGER native_error; 
 SQLSMALLINT sql_msg_len;

 switch (in_oret) {
 case SQL_SUCCESS:
   ret= 2;                                             /* indicates no error */
   break;
 case SQL_SUCCESS_WITH_INFO:
   ret= 2;                                             /* indicates no error */
   strcpy(errmsg,"ErrorSQL_SUCCESS_WITH_INFO");
   break;
 case SQL_NEED_DATA:
   strcpy(errmsg,"ErrorSQL_NEED_DATA");
   break;
 case SQL_NO_DATA:
   strcpy(errmsg,"ErrorSQL_NO_DATA");
   break;
 case SQL_ERROR:
   errmsg[0]= 0;
   if(sdb->stq==NULL || sdb->stq->stmt==NULL) {
     for(i=1;;i++) {
       Sdb_odbc_trace(sdb,"o1e vor : SQLGetDiagRec()",1);
       oret= SQLGetDiagRec(SQL_HANDLE_ENV,sdb->db,(SQLSMALLINT) i,sql_state,
                           &native_error,sql_msg,(SQLSMALLINT) sizeof(sql_msg),
                           &sql_msg_len);
       Sdb_odbc_trace(sdb,"o1e nach: SQLGetDiagRec()",1);
       if(oret==SQL_NO_DATA)
     break;
       if(sql_msg_len+strlen(errmsg)>=Sdb_odbc_min_msg_sizE)
     break;

       Sdb_odbc_trace(sdb,"SQLGetDiagRec() : ",0);
       Sdb_odbc_trace(sdb,sql_msg,0);
       
       strcpy(errmsg+strlen(errmsg),sql_msg);
     }
   } else {
     for(i=1;;i++) {
       Sdb_odbc_trace(sdb,"o1s vor : SQLGetDiagRec()",0);
       oret= SQLGetDiagRec(SQL_HANDLE_STMT,sdb->stq->stmt,(SQLSMALLINT) i,
                           sql_state,&native_error,sql_msg,
                           (SQLSMALLINT) sizeof(sql_msg),&sql_msg_len);
       Sdb_odbc_trace(sdb,"o1s nach: SQLGetDiagRec()",0);
       if(oret==SQL_NO_DATA)
     break;
       if(sql_msg_len+strlen(errmsg)>=Sdb_odbc_min_msg_sizE)
     break;

       Sdb_odbc_trace(sdb,"SQLGetDiagRec() : ",0);
       Sdb_odbc_trace(sdb,sql_msg,0);
       
       strcpy(errmsg+strlen(errmsg),sql_msg);
     }
   }
   break;
 case SQL_INVALID_HANDLE:
   strcpy(errmsg,"ErrorSQL_INVALID_HANDLE\n");
   break;
 case SQL_STILL_EXECUTING:
   strcpy(errmsg,"ErrorSQL_STILL_EXECUTE\n");
   break;
 default:
   sprintf(errmsg,"Unknown Error #%lu\n",in_oret);
   ret= 0;                                       /* indicates unknown error */
   break;
 }
 return(ret);
}



int Sdb_odbc_get_query_result(sdb,value,flag)
struct Sdb_odbC *sdb;
struct Sdb_odbc_stQ **value;
int flag;
{
 *value= sdb->stq;
 return(1);
}


int Sdb_odbc_set_query_result(sdb,value,flag)
struct Sdb_odbC *sdb;
struct Sdb_odbc_stQ *value;
int flag;
{
 sdb->stq= value;
 return(1);
}



int Sdb_odbc_init(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int version_value,ret;

 sdb->initialized= 1;


#ifdef Sdb_odbc_odbc2

 Sdb_odbc_trace(sdb,"o2 vor : SQLAllocEnv()",0);
 sdb->oret= SQLAllocEnv(&(sdb->db));
 Sdb_odbc_trace(sdb,"o2 nach: SQLAllocEnv()",0);

#else /* Sdb_odbc_odbc2 */

 Sdb_odbc_trace(sdb,"o3 vor : SQLAllocHandle()",1);
 sdb->oret= SQLAllocHandle(SQL_HANDLE_ENV,SQL_NULL_HANDLE,&(sdb->db));
 Sdb_odbc_trace(sdb,"o3 nach: SQLAllocHandle()",1);

#endif /* Sdb_odbc_odbc2 */

 if(sdb->oret != SQL_SUCCESS) {
   ret= Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);
   if(ret<=1)
     return(0);
 } 
 version_value= SQL_OV_ODBC3;
 Sdb_odbc_trace(sdb,"o4 vor : SQLSetEnvAttr()",1);
 sdb->oret= SQLSetEnvAttr(sdb->db,SQL_ATTR_ODBC_VERSION,
                     (SQLPOINTER) version_value,0);
 Sdb_odbc_trace(sdb,"o4 nach: SQLSetEnvAttr()",1);
 if(sdb->oret != SQL_SUCCESS) {
   ret= Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);
   if(ret<=1)
     return(0);
 } 

#ifdef Sdb_odbc_odbc2

 Sdb_odbc_trace(sdb,"o5 vor : SQLAllocConnect()",0);
 sdb->oret= SQLAllocConnect(sdb->db,&(sdb->svchp));
 Sdb_odbc_trace(sdb,"o5 nach: SQLAllocConnect()",0);

#else /* Sdb_odbc_odbc2 */

 Sdb_odbc_trace(sdb,"o6 vor : SQLAllocHandle()",1);
 sdb->oret= SQLAllocHandle(SQL_HANDLE_DBC,sdb->db,&(sdb->svchp));
 Sdb_odbc_trace(sdb,"o6 nach: SQLAllocHandle()",1);

#endif /* Sdb_odbc_odbc2 */

 if(sdb->oret != SQL_SUCCESS) {
   ret= Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);
   if(ret<=1)
     return(0);
 } 
 Sdb_odbc_trace(sdb,"Sdb_odbc_init : return(1)",1);
 return(1);
}


int Sdb_odbc_open(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit0= no Sdb_odbc_handle_error() in case of error
*/
{
 int retry_count= 0,ret;
 char errmsg[Sdb_odbc_min_msg_sizE],*datasource= NULL;

 /* with ODBC there seems to be no separate host and database names */
 /* this implementation ignores the hostname and only uses dbname */
 if(Sregex_string(&datasource,sdb->dbname,0)<=0)
   {ret= -1; goto ex;}
retry:;

 Sdb_odbc_trace(sdb,"o7 vor : SQLConnect()",0);
 sdb->oret= SQLConnect(sdb->svchp,datasource,(SQLSMALLINT) strlen(datasource),
                  sdb->username,(SQLSMALLINT) strlen(sdb->username),
                  sdb->password,(SQLSMALLINT) strlen(sdb->password));
 Sdb_odbc_trace(sdb,"o7 nach: SQLConnect()",0);
 if(sdb->oret != SQL_SUCCESS && sdb->oret != SQL_SUCCESS_WITH_INFO) {
   Sdb_odbc_trace(sdb,"o7 failed : SQLConnect()",0);
   if(!flag&1) {
     ret= Sdb_odbc_handle_error(sdb,0);
     if(ret==1) {
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY) {
         Sdb_odbc_trace(sdb,"o7 retry : SQLConnect()",0);
         goto retry;
       }     
     }     
   }     
   ret= 0; goto ex;
 }
 sdb->opened= 1;
 ret= 1;
 Sdb_odbc_trace(sdb,"Sdb_odbc_open : return(1)",0);
ex:;
 Sregex_string(&datasource,NULL,0);
 return(ret);
}


int Sdb_odbc_close(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{

 Sdb_odbc_free_result(sdb,0);

 if(!sdb->opened)
   return(2);

 Sdb_odbc_trace(sdb,"o8 vor : SQLDisconnect()",0);
 sdb->oret= SQLDisconnect(sdb->svchp);
 Sdb_odbc_trace(sdb,"o8 nach: SQLDisconnect()",0);
 if(sdb->oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);

 Sdb_odbc_trace(sdb,"o9 vor : SQLFreeHandle()",0);
 sdb->oret= SQLFreeHandle(SQL_HANDLE_DBC,sdb->svchp);
 Sdb_odbc_trace(sdb,"o9 nach: SQLFreeHandle()",0);
 if(sdb->oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);

 Sdb_odbc_trace(sdb,"o10 vor : SQLFreeHandle()",0);
 sdb->oret= SQLFreeHandle(SQL_HANDLE_ENV,sdb->db);
 Sdb_odbc_trace(sdb,"o10 nach: SQLFreeHandle()",0);
 if(sdb->oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,sdb->oret,sdb->message,0);
 sdb->opened= 0;
 Sdb_odbc_trace(sdb,"o10 : return(1)",0);
 return(1);
}


int Sdb_odbc_handle_error(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 ??? bit2= copy o c i _error() to sdb->message
*/
{
 int ret,siz;
 char errmsg[Sdb_odbc_min_msg_sizE];
 unsigned int merrno= 0;

 Sdb_odbc_checkerr(sdb,sdb->oret,errmsg,0);
 if(flag&2) 
   strcpy(sdb->message,errmsg);

 if(merrno==123456789) {                                      /* out of duty */

   fprintf(stderr,"%s\n... trying to reconnect ...\n",errmsg);

   /* shut down */;
   Sdb_odbc_free_result(sdb,0);
   Sdb_odbc_close(sdb,0);
   /* bring up again */;
   ret= Sdb_odbc_init(sdb,0);
   if(ret<=0)
     return(ret);
   /* try to reconnect */;
   ret= Sdb_odbc_open(sdb,1);
   if(ret<=0)
     return(ret);
   return(1);
 }

 return(0);
}


int Sdb_odbc_query(sdb,query_text,query_length,flag)
struct Sdb_odbC *sdb;
char *query_text;
int query_length;
int flag;
/*
 bit0= do send COMMIT WORK by temporary statement
       (usually this is controlled by sdb->auto_commit)
 bit1= do not send query_text (but only COMMIT WORK if bit0 is set)
*/
{
 char *dummy;
 int retry_count= 0,ret,send_commit= 0;
 struct Sdb_odbc_stQ *commit_stq= NULL;


 Sdb_odbc_trace(sdb,"Sdb_odbc_query 1",0);

 if(sdb->opened==0) {
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 2",0);
*/
   ret= Sdb_odbc_open(sdb,0);
   if(ret<=0)
     goto ex;
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 3",0);
*/
 }

 if(!(flag&2)) {
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 4",0);
*/
   Sdb_odbc_stq_destroy(&(sdb->stq),0);
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 5",0);
*/
   ret= Sdb_odbc_stq_new(&(sdb->stq),sdb,query_text,query_length,0);
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 6",0);
*/
   if(ret<=0) 
     goto ex;
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 7",0);
*/
   ret= Sdb_odbc_stq_prepare(sdb->stq,0);
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 8",0);
*/
   if(ret<=0) 
     goto ex;
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 9",0);
*/
   if(sdb->auto_commit) {
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 10",0);
*/
     ret= Sdb_odbc_stq_is_commitable(sdb->stq,0);
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 11",0);
*/
     if(ret<0)
       goto ex;
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 12",0);
*/
     if(ret>0)
       send_commit= 1;
   }
retry:;
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 13",0);
*/
   ret= Sdb_odbc_stq_execute(sdb->stq,0);

/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 14",0);
*/
   if(ret<=0) {
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 15",0);
*/
     ret= Sdb_odbc_handle_error(sdb,2);
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 16",0);
*/
     if(ret==1) {
/*
       Sdb_odbc_trace(sdb,"Sdb_odbc_query 17",0);
*/
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY)
         goto retry;
/*
       Sdb_odbc_trace(sdb,"Sdb_odbc_query 18",0);
*/
     }     
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 19",0);
*/
     ret= 0; goto ex;
   }
 }
/*
 Sdb_odbc_trace(sdb,"Sdb_odbc_query 20",0);
*/
 if(flag&1)
   send_commit= 1;
/*
 Sdb_odbc_trace(sdb,"Sdb_odbc_query 21",0);
*/ 
 if(send_commit) {
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 22",0);
*/ 
   ret= Sdb_odbc_stq_new(&commit_stq,"COMMIT WORK",11,0);
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 23",0);
*/ 
   if(ret<=0) 
     goto ex;
retry_commit:;
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 24",0);
*/ 
   ret= Sdb_odbc_stq_execute(commit_stq,0);
/*
   Sdb_odbc_trace(sdb,"Sdb_odbc_query 25",0);
*/ 
   if(ret<=0) {
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 26",0);
*/ 
     ret= Sdb_odbc_handle_error(sdb,2);
     if(ret==1) {
/*
       Sdb_odbc_trace(sdb,"Sdb_odbc_query 27",0);
*/ 
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY)
         goto retry_commit;
     }
/*
     Sdb_odbc_trace(sdb,"Sdb_odbc_query 28",0);
*/ 
     ret= 0; goto ex;
   }
 }
/*
 Sdb_odbc_trace(sdb,"Sdb_odbc_query 29",0);
*/ 
 ret= 1;
ex:;
/*
 Sdb_odbc_trace(sdb,"Sdb_odbc_query 30",0);
*/ 
 Sdb_odbc_stq_destroy(&(commit_stq),0);
/*
 Sdb_odbc_trace(sdb,"Sdb_odbc_query 31",0);
*/ 
 return(ret);
}


int Sdb_odbc_free_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 sprintf(trace_linE,"Sdb_odbc_free_result : sdb->stq= %X",sdb->stq);
 Sdb_odbc_trace(sdb,trace_linE,0);
 
 ret= Sdb_odbc_stq_destroy(&(sdb->stq),0);

 Sdb_odbc_trace(sdb,"Sdb_odbc_free_result : nach Sdb_odbc_stq_destroy",0);
 
 return(ret);
}


int Sdb_odbc_request_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret;

 ret= Sdb_odbc_stq_rowform(sdb->stq,0);
 if(ret<=0)
   return(ret);
 ret= Sdb_odbc_stq_resultbuf(sdb->stq,(flag&16));
 if(ret<=0)
   return(ret);
 return(1);
}


int Sdb_odbc_get_fieldcount(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int fieldcount;

 fieldcount= Sdb_odbc_stq_column_count(sdb->stq,0);
 return(fieldcount);
}


int Sdb_odbc_get_row(sdb,fields,lengths,flag)
struct Sdb_odbC *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret;

retry:;
 ret= Sdb_odbc_stq_next_row(sdb->stq,fields,lengths,0);
 return(ret);
}


int Sdb_odbc_end_query(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 if(sdb->stq==NULL)
   return(2);
 ret= Sdb_odbc_stq_end_query(sdb->stq,0);
 return(ret);
}



int Sdb_odbc_update_worked(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 ret= Sdb_odbc_stq_update_worked(sdb->stq,0);
 return(ret);
}


int Sdb_odbc_esc_string(sdb,target,source,source_length,flag)
struct Sdb_odbC *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 char inflator= '\'',*rpt,*wpt;

 wpt= target;
 for(rpt= source; *rpt!= 0; rpt++) {
   if((*rpt)==inflator)
     *(wpt++)= inflator;
   *(wpt++)= *rpt;
 } 
 *wpt= 0;
 return(1);
}


/* >>> BAUSTELLE */


int Sdb_odbc_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_odbC *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 /* >>> */;
 *seed= 0;
 return(1);
}


int Sdb_odbc_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_odbC *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 /* >>> this only works up to 32 bit, the MySQl type is 64 bit as it seems */
 /* >>> how to use 64 bit in a portable way ? */


 /* >>> *id=  SELECT @@IDENTITY FROM table  */

 return(1);
}


#ifdef Sdb_odbc_gen_apI

/* ------------------ general SDB API (if desired) ---------------- */


int Sdb_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_generiC **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 int ret;
 ret= Sdb_odbc_new(sdb,hostname,username,password,dbname,message_pt,flag);
 return(ret);
}


int Sdb_destroy(sdb,flag)
struct Sdb_generiC **sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_destroy(sdb,flag);
 return(ret);
}


int Sdb_get_query_result(sdb,value,flag)
struct Sdb_generiC *sdb;
Sdb_res_generiC **value;
int flag;
{
 int ret;
 ret= Sdb_odbc_get_query_result(sdb,value,flag);
 return(ret);
}


int Sdb_set_query_result(sdb,value,flag)
struct Sdb_generiC *sdb;
Sdb_res_generiC *value;
int flag;
{
 int ret;
 ret= Sdb_odbc_set_query_result(sdb,value,flag);
 return(ret);
}


int Sdb_init(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret; 
 ret= Sdb_odbc_init(sdb,flag);
 return(ret);
}


int Sdb_handle_error(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit2= copy odbc_error() to sdb->message
*/
{
 int ret;
 ret= Sdb_odbc_handle_error(sdb,flag);
 return(ret);
}


int Sdb_open(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit0= no Sdb_odbc_handle_error() in case of error
*/
{
 int ret;
 ret= Sdb_odbc_open(sdb,flag);
 return(ret);
}


int Sdb_close(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_close(sdb,flag);
 return(ret);
}


int Sdb_query(sdb,query_text,query_length,flag)
struct Sdb_generiC *sdb;
char *query_text;
int query_length;
int flag;
{
 int ret;
 ret= Sdb_odbc_query(sdb,query_text,query_length,flag);
 return(ret);
}


int Sdb_request_result(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret;
 ret= Sdb_odbc_request_result(sdb,flag);
 return(ret);
}


int Sdb_esc_string(sdb,target,source,source_length,flag)
struct Sdb_generiC *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 int ret;
 ret= Sdb_odbc_esc_string(sdb,target,source,source_length,flag);
 return(ret);
}


int Sdb_end_query(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_end_query(sdb,flag);
 return(ret); 
} 


int Sdb_free_result(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_free_result(sdb,flag); 
 return(ret);  
}  


int Sdb_unique_seed(sdb,fieldname,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_generiC *sdb;
char *fieldname;
unsigned long *seed;
int flag;
{
 int ret; 
 ret= Sdb_odbc_unique_seed(sdb,fieldname,seed,flag);  
 return(ret);   
}   


int Sdb_get_unique_id(sdb,fieldname,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_generiC *sdb;
char *fieldname;
unsigned long *id;
int flag;
{
 int ret;  
 ret= Sdb_odbc_get_unique_id(sdb,fieldname,id,flag);
 return(ret);    
}    


int Sdb_get_fieldcount(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;   
 ret= Sdb_odbc_get_fieldcount(sdb,flag); 
 return(ret);     
}     
 

int Sdb_get_row(sdb,fields,lengths,flag)
struct Sdb_generiC *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int ret;    
 ret= Sdb_odbc_get_row(sdb,fields,lengths,flag);  
 return(ret);      
}      


int Sdb_update_worked(sdb,flag)
struct Sdb_generiC *sdb;
int flag;
{
 int ret;

 ret= Sdb_odbc_update_worked(sdb,flag);
 return(ret);
}


char *Sdb__get_typelist_pt(flag)
int flag;
{
 return(Sdb_odbc_typelisT);
}


#endif /* Sdb_odbc_gen_apI */
