
/*
 cc -g -c sdb_odbc.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "sdb_odbc_stq.h"

#include "sdb_odbc.h"


/* --------------------- SDB wrapper for ODBC ------------------ */


int Sdb_odbc_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_odbC **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 struct Sdb_odbC *m;

 m= *sdb= TSOB_FELD(struct Sdb_odbC,1);
 if(m==NULL)
   return(-1);

 m->hostname= NULL;
 m->username= NULL;
 m->password= NULL;
 m->dbname= NULL;
 m->initialized= 0;
 m->opened= 0;
 m->auto_commit= 1;
 m->fetch_mode= 0;
 m->stq= NULL;
 m->query_token= NULL;
 m->search_where= NULL;
 m->search_stq= NULL;
 m->search_token= NULL;
 m->message= message_pt;

 if(Sregex_string(&(m->hostname),hostname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->username),username,0)<=0)
   goto failed;
 if(Sregex_string(&(m->password),password,0)<=0)
   goto failed;
 if(Sregex_string(&(m->dbname),dbname,0)<=0)
   goto failed;
 return(1);

failed:;
 Sdb_odbc_destroy(sdb,0);
 return(-1);
}


int Sdb_odbc_destroy(sdb,flag)
struct Sdb_odbC **sdb;
int flag;
{
 struct Sdb_odbC *m;

 m= *sdb;
 if(m==NULL)
   return(0);
 Sdb_odbc_close(m,0);
 Sregex_string(&(m->dbname),NULL,0);
 Sregex_string(&(m->password),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->hostname),NULL,0);
 Smem_freE((char *) m);
 *sdb= NULL;
 return(1);
}


int Sdb_odbc_checkerr(sdb,in_oret,errmsg,flag)
struct Sdb_odbC *sdb;
SQLRETURN in_oret;
char errmsg[Sdb_odbc_min_msg_sizE];
int flag;
{
 int ret= 1,i;
 SQLRETURN oret;
 SQLCHAR sql_state[6],sql_msg[SQL_MAX_MESSAGE_LENGTH];
 SQLINTEGER native_error; 
 SQLSMALLINT sql_msg_len;

 switch (in_oret) {
 case SQL_SUCCESS:
   ret= 2;                                             /* indicates no error */
   break;
 case SQL_SUCCESS_WITH_INFO:
   strcpy(errmsg,"ErrorSQL_SUCCESS_WITH_INFO");
   break;
 case SQL_NEED_DATA:
   strcpy(errmsg,"ErrorSQL_NEED_DATA");
   break;
 case SQL_NO_DATA:
   strcpy(errmsg,"ErrorSQL_NO_DATA");
   break;
 case SQL_ERROR:
   errmsg[0]= 0;
   for(i=1;;i++) {
     /* ??? how to do this in ODBC 2.X ??? */
     oret= SQLGetDiagRec(SQL_HANDLE_STMT,sdb->stq->stmt,i,sql_state,
                         &native_error,sql_msg,sizeof(sql_msg),&sql_msg_len);
     if(oret==SQL_NO_DATA)
   break;
     if(sql_msg_len+strlen(errmsg)>=Sdb_odbc_min_msg_sizE)
   break;
     strcpy(errmsg+strlen(errmsg),sql_msg);
   }
   break;
 case SQL_INVALID_HANDLE:
   strcpy(errmsg,"ErrorSQL_INVALID_HANDLE\n");
   break;
 case SQL_STILL_EXECUTING:
   strcpy(errmsg,"ErrorSQL_STILL_EXECUTE\n");
   break;
 default:
   sprintf(errmsg,"Unknown Error #%lu\n",in_oret);
   ret= 0;                                       /* indicates unknown error */
   break;
 }
 return(ret);
}


int Sdb_odbc_init(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 SQLRETURN oret;
 dvoid *dummy;
 int version_value;

 sdb->initialized= 1;

 oret= SQLAllocHandle(SQL_HANDLE_ENV,SQL_NULL_HANDLE,&(sdb->db));
 if(oret != SQL_SUCCESS) {
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 version_value= SQL_OV_ODBC3;
 oret= SQLSetEnvAttr(sdb->db,SQL_ATTR_ODBC_VERSION,&version_value,0);
 if(oret != SQL_SUCCESS) {
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 oret= SQLAllocHandle(SQL_HANDLE_ENV,sdb->db,&(sdb->db));
 if(oret != SQL_SUCCESS) {
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);
   return(0);
 } 
 return(1);
}


int Sdb_odbc_open(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit0= no Sdb_odbc_handle_error() in case of error
*/
{
 SQLRETURN oret;
 int retry_count= 0,ret;
 char errmsg[Sdb_odbc_min_msg_sizE],*datasource= NULL;

 if(Sregex_string(&datasource,sdb->hostname,0)<=0)
   {ret= -1; goto ex;}
 /* with ODBC there seems to be no separate database name */
 if(Sregex_string(&datasource,sdb->dbname,1)<=0)
   {ret= -1; goto ex;}
retry:;

 oret= SQLConnect(sdb->svchp,datasource,strlen(datasource),
                  sdb->username,strlen(sdb->username),
                  sdb->password,strlen(sdb->password));
 if(oret != SQL_SUCCESS) {
   if(!flag&1) {
     ret= Sdb_odbc_handle_error(sdb,0);
     if(ret==1) {
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY)
         goto retry;
     }     
   }     
   ret= 0; goto ex;
 }
 sdb->opened= 1;
 ret= 1;
ex:;
 Sregex_string(&datasource,NULL,0);
 return(1);
}


int Sdb_odbc_close(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 SQLRETURN oret;

 Sdb_odbc_free_result(sdb,0);

 if(!sdb->opened)
   return(2);

 oret= SQLDisconnect(sdb->svchp);
 if(oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);

 oret= SQLFreeHandle(SQL_HANDLE_DBC,sdb->svchp);
 if(oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);

 oret= SQLFreeHandle(SQL_HANDLE_ENV,sdb->db);
 if(oret != SQL_SUCCESS) 
   Sdb_odbc_checkerr(sdb,oret,sdb->message,0);
 sdb->opened= 0;
 return(1);
}


int Sdb_odbc_handle_error(sdb,oret,flag)
struct Sdb_odbC *sdb;
sword oret;
int flag;
/*
 ??? bit2= copy o c i _error() to sdb->message
*/
{
 int ret,siz;
 char errmsg[Sdb_odbc_min_msg_sizE];
 unsigned int merrno= 0;

 Sdb_odbc_checkerr(sdb,oret,errmsg,0);
 if(flag&2) 
   strcpy(sdb->message,errmsg);

 if(merrno==123456789) {                                      /* out of duty */

   fprintf(stderr,"%s\n... trying to reconnect ...\n",errmsg);

   /* shut down */;
   Sdb_odbc_free_result(sdb,0);
   Sdb_odbc_close(sdb,0);
   /* bring up again */;
   ret= Sdb_odbc_init(sdb,0);
   if(ret<=0)
     return(ret);
   /* try to reconnect */;
   ret= Sdb_odbc_open(sdb,1);
   if(ret<=0)
     return(ret);
   return(1);
 }

 return(0);
}


int Sdb_odbc_query(sdb,query_text,query_length,flag)
struct Sdb_odbC *sdb;
char *query_text;
int query_length;
int flag;
/*
 bit0= do send COMMIT WORK by temporary statement
       (usually this is controlled by sdb->auto_commit)
 bit1= do not send query_text (but only COMMIT WORK if bit0 is set)
*/
{
 sword oret;
 char *dummy;
 int retry_count= 0,ret,send_commit= 0;
 struct Sdb_odbc_stQ *commit_stq= NULL;

 if(sdb->opened==0) {
   ret= Sdb_odbc_open(sdb,0);
   if(ret<=0)
     goto ex;
 }

 if(!(flag&2)) {
   Sdb_odbc_stq_destroy(&(sdb->stq),0);
   ret= Sdb_odbc_stq_new(&(sdb->stq),query_text,query_length,0);
   if(ret<=0) 
     goto ex;
   ret= Sdb_odbc_stq_prepare(sdb->stq,0);
   if(ret<=0) 
     goto ex;
   if(sdb->auto_commit) {
     ret= Sdb_odbc_stq_is_commitable(sdb->stq,0);
     if(ret<0)
       goto ex;
     if(ret>0)
       send_commit= 1;
   }
retry:;
   ret= Sdb_odbc_stq_execute(sdb->stq,0);
   if(ret<=0) {
     ret= Sdb_odbc_handle_error(sdb,2);
     if(ret==1) {
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY)
         goto retry;
     }     
     ret= 0; goto ex;
   }
 }
 if(flag&1)
   send_commit= 1;
 if(send_commit) {
   ret= Sdb_odbc_stq_new(&commit_stq,"COMMIT WORK",11,0);
   if(ret<=0) 
     goto ex;
retry_commit:;
   ret= Sdb_odbc_stq_execute(commit_stq,0);
   if(ret<=0) {
     ret= Sdb_odbc_handle_error(sdb,2);
     if(ret==1) {
       if(retry_count++<=Sdb_odbc_reconnect_maxtrY)
         goto retry_commit;
     }
     ret= 0; goto ex;
   }
 }
 ret= 1;
ex:;
 Sdb_odbc_stq_destroy(&(commit_stq),0);
 return(ret);
}


int Sdb_odbc_free_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 ret= Sdb_odbc_stq_destroy(&(sdb->stq),0);
 return(ret);
}


int Sdb_odbc_request_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int retry_count= 0,ret;
 sword oret;
 unsigned long rowcount= 0,rowsize= 0,row_offsets[256],colcount= 0;
 ub2 dtype;
 ub4 dsize;

 ret= Sdb_odbc_stq_rowform(sdb->stq,0);
 if(ret<=0)
   return(ret);
 ret= Sdb_odbc_stq_resultbuf(sdb->stq,(flag&16));
 if(ret<=0)
   return(ret);
 return(1);
}


int Sdb_odbc_get_fieldcount(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int fieldcount;

 fieldcount= Sdb_odbc_stq_column_count(sdb->stq,0);
 return(fieldcount);
}


int Sdb_odbc_get_row(sdb,fields,lengths,flag)
struct Sdb_odbC *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret;
 unsigned int *fieldlengths;

retry:;
 ret= Sdb_odbc_stq_next_row(sdb->stq,fields,lengths,0);
 return(ret);
}


int Sdb_odbc_end_query(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 if(sdb->stq==NULL)
   return(2);
 ret= Sdb_odbc_stq_end_query(sdb->stq,0);
 return(ret);
}



int Sdb_odbc_update_worked(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 ret= Sdb_odbc_stq_update_worked(sdb->stq,0);
 return(ret);
}


int Sdb_odbc_esc_string(sdb,target,source,source_length,flag)
struct Sdb_odbC *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 char inflator= '\'',*rpt,*wpt;

 wpt= target;
 for(rpt= source; *rpt!= 0; rpt++) {
   if((*rpt)==inflator)
     *(wpt++)= inflator;
   *(wpt++)= *rpt;
 } 
 *wpt= 0;
 return(1);
}


/* >>> BAUSTELLE */


int Sdb_odbc_unique_seed(sdb,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_odbC *sdb;
unsigned long *seed;
int flag;
{
 /* >>> */;
 *seed= 0;
 return(1);
}


int Sdb_odbc_get_unique_id(sdb,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_odbC *sdb;
unsigned long *id;
int flag;
{
 /* >>> this only works up to 32 bit, the MySQl type is 64 bit as it seems */
 /* >>> how to use 64 bit in a portable way ? */

 /* >>> *id= odbc_insert_id(sdb->db); */
 return(1);
}


#ifdef Sdb_odbc_gen_apI

/* ------------------ general SDB API (if desired) ---------------- */


int Sdb_new(sdb,hostname,username,password,dbname,message_pt,flag)
struct Sdb_odbC **sdb;
char *hostname;
char *username;
char *password;
char *dbname;
char *message_pt;
int flag;
{
 int ret;
 ret= Sdb_odbc_new(sdb,hostname,username,password,dbname,message_pt,flag);
 return(ret);
}


int Sdb_destroy(sdb,flag)
struct Sdb_odbC **sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_destroy(sdb,flag);
 return(ret);
}


int Sdb_init(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret; 
 ret= Sdb_odbc_init(sdb,flag);
 return(ret);
}


int Sdb_handle_error(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit2= copy odbc_error() to sdb->message
*/
{
 int ret;
 ret= Sdb_odbc_handle_error(sdb,flag);
 return(ret);
}


int Sdb_open(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit0= no Sdb_odbc_handle_error() in case of error
*/
{
 int ret;
 ret= Sdb_odbc_open(sdb,flag);
 return(ret);
}


int Sdb_close(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_close(sdb,flag);
 return(ret);
}


int Sdb_query(sdb,query_text,query_length,flag)
struct Sdb_odbC *sdb;
char *query_text;
int query_length;
int flag;
{
 int ret;
 ret= Sdb_odbc_query(sdb,query_text,query_length,flag);
 return(ret);
}


int Sdb_request_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int ret;
 ret= Sdb_odbc_request_result(sdb,flag);
 return(ret);
}


int Sdb_esc_string(sdb,target,source,source_length,flag)
struct Sdb_odbC *sdb;
char *target;
char *source;
int source_length;
int flag;
{
 int ret;
 ret= Sdb_odbc_esc_string(sdb,target,source,source_length,flag);
 return(ret);
}


int Sdb_end_query(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_end_query(sdb,flag);
 return(ret); 
} 


int Sdb_free_result(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;
 ret= Sdb_odbc_free_result(sdb,flag); 
 return(ret);  
}  


int Sdb_unique_seed(sdb,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct Sdb_odbC *sdb;
unsigned long *seed;
int flag;
{
 int ret; 
 ret= Sdb_odbc_unique_seed(sdb,seed,flag);  
 return(ret);   
}   


int Sdb_get_unique_id(sdb,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct Sdb_odbC *sdb;
unsigned long *id;
int flag;
{
 int ret;  
 ret= Sdb_odbc_get_unique_id(sdb,id,flag);
 return(ret);    
}    


int Sdb_get_fieldcount(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;   
 ret= Sdb_odbc_get_fieldcount(sdb,flag); 
 return(ret);     
}     
 

int Sdb_get_row(sdb,fields,lengths,flag)
struct Sdb_odbC *sdb;
char ***fields;
unsigned long **lengths;
int flag;
{
 int ret;    
 ret= Sdb_odbc_get_row(sdb,fields,lengths,flag);  
 return(ret);      
}      


int Sdb_update_worked(sdb,flag)
struct Sdb_odbC *sdb;
int flag;
{
 int ret;

 ret= Sdb_odbc_update_worked(sdb,flag);
 return(ret);
}


char *Sdb__get_typelist_pt(flag)
int flag;
{
 return(Sdb_odbc_typelisT);
}


#endif /* Sdb_odbc_gen_apI */
