
/*
  cc -g -c sregex.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#ifdef Sregex_use_smeM

#include "smem.h"

#else /* Sregex_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Sregex_use_smeM */

#include "stypes.h"

#include "sregex.h"


#ifdef Stypes_has_regeX

/* --------------------- sregex ---------------------- */


int Sregex_new(re,expr,flag)
struct SregeX **re;
char *expr;
int flag;
{
 struct SregeX *r;
 r= TSOB_FELD(struct SregeX,1);
 if(r==NULL)
   return(-1);
 r->compiled= 0;
 *re= r;
 if(expr[0]!=0){
   if(Sregex_comp(*re,expr,0)<=0)
     return(0); 
 }else{
   if(Sregex_comp(*re,".*",0)<=0)
     return(0);
 }
 return(1);
}


int Sregex_destroy(re,flag)
struct SregeX **re;
int flag;
{
 if(*re==NULL)
   return(0);
 Sregex_free(*re,0);
 Smem_freE((char *) *re);
 *re= NULL;
 return(1);
}


int Sregex_comp(re,expr,flag)
struct SregeX *re;
char *expr;
int flag;
{
 re->compiled= 0;
 strcpy(re->expr,expr);
 if(regcomp(&(re->reg),expr,0)!=0)
   return(0);
 re->compiled= 1;
 return(1);
}


int Sregex_free(re,flag)
struct SregeX *re;
int flag;
{
 if(re->compiled==0)
   return(0);
 regfree(&(re->reg));
 re->compiled= 0;
 return(1);
}


int Sregex_replace(re,expr,flag)
struct SregeX **re;
char *expr;
int flag;
{
 int ret;

 if(*re==NULL){
   ret= Sregex_new(re,expr,0);
 }else{
   Sregex_free(*re,0);
   if(expr[0]!=0)
     ret= Sregex_comp(*re,expr,0);
   else
     ret= 1;
 }
 return(ret);
}


int Sregex_is_match(re,text,flag)
struct SregeX *re;
char *text;
int flag;
{
 if(text[0]==0){
   if(re->expr[0]==0) 
     return(1);
   else
     return(0);
 }
 if(regexec(&(re->reg),text,1,re->match,0)==0) 
   return(1); 
 return(0);
}


int Sregex_is_match_t(expr,text,flag)
char *expr;
char *text;
int flag;
{
 struct SregeX *re;
 int ret;
 
 ret= Sregex_new(&re,expr,0);
 if(ret<=0)
   return(-1);
 ret= Sregex_is_match(re,text,0);
 Sregex_destroy(&re,0);
 return(ret);
}


#endif /* Stypes_has_regeX */


/* ----------- misc help functions ----------- */


int Sregex_string_to_int(text,reply,flag)
char *text;
int *reply;
int flag;
{
 if(text==NULL)
   return(0);
 if(sscanf(text,"%ld",reply)<=0)
   return(0);
 return(1);
}


int Sregex_complines(nakedline_in,dressedline_in,flag)
char *nakedline_in;
char *dressedline_in;
int flag;
/*
 bit0= strncmp()==0 is sufficient
 bit1= compare as lowercase
*/
{
 int l,ret;
 char *nakedline,*dressedline,*nakedlower= NULL,*dressedlower= NULL,*cpt;

 if(flag&2) {
   if(Sregex_string(&nakedlower,nakedline_in,0)<=0)
     {ret= 0; goto ex;}
   if(Sregex_string(&dressedlower,dressedline_in,0)<=0)
     {ret= 0; goto ex;}
   nakedline= nakedlower;
   dressedline= dressedlower;
   for(cpt= nakedline;*cpt!=0;cpt++)
     if(*cpt>='A' && *cpt<='Z')
       *cpt= tolower(*cpt);
   for(cpt= dressedline;*cpt!=0;cpt++)
     if(*cpt>='A' && *cpt<='Z')
       *cpt= tolower(*cpt);
 } else {
   nakedline= nakedline_in;
   dressedline= dressedline_in;
 }
 l= strlen(nakedline);
 if(l==0 || strncmp(nakedline,dressedline,l)==0)
   if((flag&1) || dressedline[l]==0 || dressedline[l]=='\r' 
               || dressedline[l]=='\n')
     {ret= 1; goto ex;}
 ret= 0;
ex:;
 Sregex_string(&nakedlower,NULL,0);
 Sregex_string(&dressedlower,NULL,0);
 return(ret);
}


int Sregex_read_text(line,smark,emark,text,textsiz,flag)
char *line;
char *smark,*emark;
char *text;
int textsiz;
int flag;
/*
 bit0= do not clear text before trying to read
*/
{
 char *rpt,*ept,*wpt,*maxwpt;

 if(!(flag&1))
   text[0]= 0;
 if(smark[0]==0)
   rpt= line;
 else
   rpt= strstr(line,smark);
 if(rpt==NULL)
   return(0);
 rpt+= strlen(smark);
 if(emark[0]==0)
   ept= line+strlen(line);
 else
   ept= strstr(rpt,emark);
 if(ept==NULL)
   return(0);
 wpt= text;
 maxwpt= text+textsiz-1;
 while(rpt<ept && wpt<maxwpt)
   *(wpt++)= *(rpt++);
 *wpt= 0;
 return(1);
}


int Sregex_trimline(line,flag)
/*
  removes line endings as well as leading and trailing blanks
*/
char *line;
int flag;
/*
 bit0= do not remove line end (protects trailing blanks if line end is present)
 bit1= do not remove leading blanks
 bit2= do not remove trailing blanks
 bit3= remove surrounding quotation marks (after removing line end)
*/
{
 char *cpt,*wpt;
 int l;

 if(!(flag&1)){
   l= strlen(line);
   if(l>0) if(line[l-1]=='\r') line[--l]= 0;
   if(l>0) if(line[l-1]=='\n') line[--l]= 0;
   if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 }
 if(flag&3){
   l= strlen(line);
   if(l>1) if(line[0]==34 && line[l-1]==34) {
     wpt= line;
     cpt= wpt+1;
     while(*cpt!=0) 
       *(wpt++)= *(cpt++);
     line[l-2]= 0;
   }
 }
 if(!(flag&2)){
   wpt= cpt= line;
   while(*(cpt)!=0) {
     if(!isspace(*cpt))
   break;
     cpt++;
   }
   while(*(cpt)!=0) 
     *(wpt++)= *(cpt++);
   *wpt= 0;  
 }
 if(!(flag&4)){
   l= strlen(line);
   if(l<=0)
     return(1);
   cpt= line+l;
   while(cpt-->=line){
     if(!isspace(*cpt))
   break;
     *(cpt)= 0;
   }
 }
 return(1);
}


int Sregex_concat_bin(head,head_len,tail,tail_len,flag)
char **head;
int *head_len;
char *tail;
int tail_len;
int flag;
{
 char *handle;

 if(*head_len<0 || tail_len<0 || *head_len+tail_len<=0)
   return(0);
 handle= TSOB_FELD(char,(*head_len+tail_len+1));
 if(handle==NULL)
   return(-1);
 handle[*head_len+tail_len]= 0;  /* for convenience with human readable data */
 if(*head_len>0)
   memcpy(handle,*head,*head_len);
 if(tail_len>0)
   memcpy(handle+*head_len,tail,tail_len);
 if(*head!=NULL)
   Smem_freE(*head);
 *head= handle;
 (*head_len)+= tail_len;
 return(1);
}


int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   if(len>0)
     strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


/* old name of Sregex_string . It was just too long */
int Sregex_set_dyn_string(handle,text,flag)
char **handle;
char *text;
int flag;
{
 int ret;
 ret= Sregex_string(handle,text,flag);
 return(ret);
}


int Sregex_search_mark(data,data_len,mark,mark_len,markpt,flag)
char *data;
int data_len;
char *mark;
int mark_len;
char **markpt;
int flag;
{
 int i;

 if(*markpt==NULL)
   return(0);
 if(*markpt>=mark+mark_len)
   return(1);

 i= 0;
 for(i=0;i<data_len;i++) {
   if(**markpt==data[i]) {
     if(*markpt>=mark+mark_len-1)
       return(1);
     (*markpt)++;
   } else 
     *markpt= mark;
 }
 return(0);
}


int Sregex_closing_bracket(text,bracket,flag)
char *text;
char bracket;
int flag;
/*
 bit0= do only respect the privacy of "..."
*/
{
 int i,j,ret,l,list_l,j_start=0;
 int quote= 0,depth= 0,extendable=1; 
 char *stack= NULL,buf[2];
 static char openers[]={"([{'`\""};
 static char closers[]={")]}'`\""};
 static char nesting[]={"yyynnn"};

 buf[1]= 0;
 list_l= strlen(openers);
 l= strlen(text);
 if(flag&1)
   j_start= list_l-1;

 for(i=0;i<l;i++){
   if(depth>0) {
     if(text[i]==closers[stack[depth-1]-1]) { 
       stack[--depth]= 0;
       if(depth>0)
         extendable= (nesting[stack[depth-1]-1]=='y');
       else
         extendable= 1;
     }
   } else if(text[i]==bracket) {
     ret= i;
     goto ex;
   } else if(extendable) {
     for(j=0;j<list_l;j++) {
       if(text[i]==openers[j]){
         buf[0]= j+1;
         ret= Sregex_string(&stack,buf,1);
         if(ret<0)
           goto ex;
         extendable= (nesting[j]=='y');
         depth++;
     break;
       }
     }
   }
 }
 ret= -1;
ex:
 Sregex_string(&stack,NULL,0);
 return(ret);
}


int Sregex_de_html_char(source,target,flag)
char **source,**target;
int flag;
{
 char *spt;
 
 spt= *source;
 if(*spt=='&') {
   if(*(spt+1)=='l') if(*(spt+2)=='t') if(*(spt+3)==';') {
     *((*target)++)= '<';
     (*source)+= 4;
     return(1);
   }
   if(*(spt+1)=='g') if(*(spt+2)=='t') if(*(spt+3)==';') {
     *((*target)++)= '>';
     (*source)+= 4;
     return(1);
   }
   if(*(spt+1)=='q') if(*(spt+2)=='u') if(*(spt+3)=='o') if(*(spt+4)=='t')
                                                         if(*(spt+5)==';') {
     *((*target)++)= '"';
     (*source)+= 6;
     return(1);
   }
   if(*(spt+1)=='a') if(*(spt+2)=='m') if(*(spt+3)=='p') if(*(spt+4)==';') {
     *((*target)++)= '&';
     (*source)+= 5;
     return(1);
   }
 } else {
   *((*target)++)= *((*source)++);
 }
 return(1);
}


int Sregex_de_html(source,target,flag)
char *source,*target;
int flag;
{
 char *bpt,*rpt,*wpt;
 int ret;

 rpt= source;
 wpt= target;
 while(1) {
   bpt= strchr(rpt,'<');
   if(bpt!=NULL) {
     while(rpt<bpt)
       Sregex_de_html_char(&rpt,&wpt,0);
     ret= Sregex_closing_bracket(bpt,'>',1);
     if(ret<=0)
 break;
     rpt= bpt+ret+1;
   } else {
     while(*rpt!=0)
       Sregex_de_html_char(&rpt,&wpt,0);
 break;
   }  
 }
 *wpt= 0;
 return(1);
}


int Sregex_cut_neck(text,neck,head,body,flag)
char *text;
char neck;
char **head; /* this is a malloced copy */
char **body; /* this is just a pointer within text. */
int flag;
/*
 flag:
   bit0= append to head rather than replace it
   bit1= return(0) if text empty
 return:
   <0 = error
    0 = flag:bit1 && text is empty
    1 = neck found, head and body enclosed
    2 = no neck found, all in head, body points to trailing 0
*/
{
 int ret=0;
 char *cpt;

 if(flag&2)
   if(text[0]==0)
     return(0);
 cpt= strchr(text,neck);
 if(cpt==NULL) {
   if(Sregex_string(head,text,flag&1)<=0)
     {ret= -1; goto ex;}
   (*body)= text+strlen(text);
   return(2);
 }
 if(Sregex_string_cut(head,text,(cpt-text),flag&1)<=0)
   {ret= -1; goto ex;}

 ret= 1;
 (*body)= cpt+1;
ex:;
 return(ret);
}


int Sregex_cut_neck_int(text,neck,head,body,flag)
char *text;
char neck;
int *head;
char **body; /* this is just a pointer within text. */
int flag;
/*
 flag:
    bit8= use %X rather than %ld
 return:
    0 = not a usable number
  others see Sregex_cut_neck()
*/
{
 char *name= NULL;
 int ret,cut_ret;

 ret= Sregex_cut_neck(text,neck,&name,body,0);
 if(ret<=0)
   goto ex;
 cut_ret= ret;
 if(strlen(name)==0)
   {ret= 0; goto ex;}
 if(flag&1)
   ret= sscanf(name,"%X",head);
 else
   ret= sscanf(name,"%ld",head);
 if(ret<=0)
   {ret= 0; goto ex;}
 ret= cut_ret;
ex:;
 Sregex_string(&name,NULL,0);
 return(ret);
}


int Sregex_hex_value(text,value,length,flag)
char *text;
unsigned long *value;
int length;
int flag;
{
 int single,i;

 *value= 0;
 if(length>8)
   return(-1);
 if(length>strlen(text))
   return(0);
 for(i=0;i<length;i++) {
   (*value)<<= 4;
   if(text[i]>='0' && text[i]<='9')
     single= text[i]-'0';
   else if (text[i]>='A' && text[i]<='F')
     single= text[i]-'A'+10;
   else if (text[i]>='a' && text[i]<='f')
     single= text[i]-'a'+10;
   else
     return(0);
   (*value)|= single;
 }
 return(1); 
}


int Sregex_crc32_various(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
 bit8-15= mode 
          0: Craig Bruce
          1: _yenc
*/
{
 int mode,ret= 0;

 mode= (flag>>8)&255;
 switch(mode) {
 case 0:
   ret= Sregex_crc32(data,len,crc,flag&~(255<<8));
 break; case 1:
   ret= Sregex_crc32_yenc(data,len,crc,flag&~(255<<8));
 }
 return(ret);
}
 
  
/* crc32.c by Craig Bruce, 24-May-92 */
/* found 25 December 1999 as
   http://nic.funet.fi/pub/cbm/crossplatform/converters/unix/crc32.c
   data source altered from file to string: 1 September 2000

   seems to comply to CCITT but is not compatible with SFV/CSV files
*/
int Sregex_crc32(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
#define CRC32_GEN       0x04c11db7  /* CCITT standard */
#define CRC32_MSB       0x80000000
       int c, i, crcOverflow, k;
       register unsigned long crcbin;

       if(flag&1)
        crcbin = 0;
       else
        crcbin = *crc;
       for(k=0;k<len;k++) {
                c= data[k];
                for (i=0x80; i; i >>= 1) {
                        crcOverflow = crcbin & CRC32_MSB;
                        crcbin = (crcbin << 1) | (i&c ? 1 : 0);
                        if (crcOverflow) crcbin ^= CRC32_GEN;
                }
       }
 
 *crc= crcbin;
 return(1);
}


/*
  from Juergen Helbing's  yencode-c.txt 
  found at  www.yenc.org  , Feb 4,  2002
  license : "It is public domain - as well as the yEnc specification"
  some alterations made to adapt it to Sregex_crc32() 

  seems to be compatible with SFV/CSV files
*/                       

int Sregex_crc32_yenc(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
 /* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
 static int crc_tab[256] = {
   0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 
   0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
   0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
   0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
   0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
   0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
   0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
   0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
   0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
   0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
   0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
   0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
   0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
   0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
   0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
   0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
   0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
   0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
   0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
   0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
   0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
   0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
   0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
   0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
   0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
   0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
   0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
   0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
   0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
   0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
   0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
   0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
   0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
   0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
   0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
   0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
   0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
   0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
   0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
   0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
   0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
   0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
   0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
   0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
   0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
   0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
   0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
   0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
   0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
   0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
   0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
   0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
   0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
   0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
   0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
   0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
   0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
   0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
   0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
   0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
   0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
   0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
   0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
   0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

 int k;
 unsigned long crc_val,ch1,ch2,cc;

 if(flag&1)
   crc_val = ~0;
 else
   crc_val = ~*crc;

 for(k=0;k<len;k++) {
   cc= data[k] & 0xffL;
   ch1= (crc_val ^ cc) & 0xffL;
   ch1= crc_tab[ch1];
   ch2= (crc_val>>8L) & 0xffffffL;
   crc_val= ch1 ^ ch2;
 }
 *crc= ~crc_val;
 return(1);
}


/* ----------------------- sorting to a permutation vector ---------------- */


int Sregex_qsort_cmp_pt(pair1,pair2)
struct QsortpaiR *pair1;
struct QsortpaiR *pair2;
{
 return(pair1->compar(pair1->data,pair2->data));
}


int Sregex_qsort_pt(base,nmemb,size,compar,permutation,flag)
char *base;
size_t nmemb;
size_t size;
int (*compar)();
int *permutation;            /* pointing to at least nmemb*sizeof(int) bytes */
int flag;
/*
 bit0= return inverse permutation
*/
{
 int i;
 struct QsortpaiR *pairs= NULL;

 if(nmemb==0)
   return(0);

 pairs= TSOB_FELD(struct QsortpaiR,nmemb);
 if(pairs==NULL)
   return(-1);
 for(i=0;i<nmemb;i++) {
   pairs[i].compar= compar;    /* this is quite a waste , but how to avoid ? */
   pairs[i].index= i;
   pairs[i].data= base+i*size;
 }
 qsort(pairs,nmemb,sizeof(struct QsortpaiR),Sregex_qsort_cmp_pt);
 if(flag&1) {
   for(i=0;i<nmemb;i++)
     permutation[pairs[i].index]= i;
 } else {
   for(i=0;i<nmemb;i++)
     permutation[i]= pairs[i].index;
 }
 Smem_free((char *) pairs);
 return(1);
}


/* ---------------------------------- LstrinG --------------------------- */


int Lstring_new_binary(lstring,data,data_len,link,flag)
struct LstrinG **lstring;
char *data;
int data_len;
struct LstrinG *link;
int flag;
/*
 bit0= insert before link rather than after it
 bit1= do not copy data (e.g. because *data is invalid)
*/
{
 int ret;
 struct LstrinG *s;

 s= TSOB_FELD(struct LstrinG,1);
 if(s==NULL)
   return(-1);
 s->text= NULL;
 s->next= s->prev= NULL;

 if(data_len<=0)
   {ret= -1; goto failed;}
 s->text= Smem_malloC(data_len);
 if(s->text==NULL)
   {ret= -1; goto failed;}
 if(!(flag&2))
   memcpy(s->text,data,data_len);

 if(link==NULL) {
   ;
 } else if(flag&1) {
   s->next= link;
   s->prev= link->prev;
   if(link->prev!=NULL) 
     link->prev->next= s;
   link->prev= s;
 } else {
   s->prev= link;
   s->next= link->next;
   if(link->next!=NULL) 
     link->next->prev= s;
   link->next= s;
 }
 *lstring= s;
 return(1);
failed:;
 *lstring= s;
 Lstring_destroy(lstring,0);
 return(-1);
}


int Lstring_new(lstring,text,link,flag)
struct LstrinG **lstring;
char *text;
struct LstrinG *link;
int flag;
/*
 bit0= insert before link rather than after it
*/
{
 int ret;

 ret= Lstring_new_binary(lstring,text,strlen(text)+1,link,flag&1);
 return(ret);
}


int Lstring_destroy(lstring,flag)
struct LstrinG **lstring;
int flag;
/*
 bit0= do not set *lstring to NULL
*/
{
 struct LstrinG *s;

 s= *lstring;
 if(s==NULL)
   return(0);
 if(s->prev!=NULL)
   s->prev->next= s->next;
 if(s->next!=NULL)
   s->next->prev= s->prev;
 if(s->text!=NULL)
   Smem_freE(s->text);
 Smem_freE((char *) s);
 if(!(flag&1))
   *lstring= NULL;
 return(1);
}


int Lstring_destroy_all(lstring,flag)
struct LstrinG **lstring;
int flag;
{
 struct LstrinG *s,*next;

 if((*lstring)==NULL)
   return(0);
 for(s= *lstring; s->prev!=NULL; s= s->prev);
 for(;s!=NULL;s= next){
   next= s->next;
   Lstring_destroy(&s,0);
 }
 *lstring= NULL;
 return(1);
}


int Lstring_find_text(lstring,text,found,flag)
struct LstrinG *lstring;
char *text;
struct LstrinG **found;
int flag;
/*
 bit0= return first lstring which begins with text
*/
{
 struct LstrinG *s;

 if(lstring==NULL)
   return(0);
 for(s= lstring; s->prev!=NULL; s= s->prev);
 if(flag&1) {
   for(;s!=NULL;s= s->next)
     if(strncmp(s->text,text,strlen(text))==0)
       {*found= s; return(1);}
 } else {
   for(;s!=NULL;s= s->next)
     if(strcmp(s->text,text)==0) 
       {*found= s; return(1);}
 }
 return(0);
}


int Lstring_is_listed(lstring,text,flag)
struct LstrinG *lstring;
char *text;
int flag;
{
 struct LstrinG *s;
 int ret;

 ret=  Lstring_find_text(lstring,text,&s,0);
 return(ret);
} 


int Lstring_copy_all(source,target,flag)
struct LstrinG *source;
struct LstrinG **target;
int flag;
{
 struct LstrinG *s,*t;

 if(source==NULL)
   return(0);
 for(s= source; s->prev!=NULL; s= s->prev);
 t= *target;
 if(t!=NULL)
   for(;t->next!=NULL;t=t->next);
 for(;s!=NULL;s= s->next){
   if(Lstring_new(&t,s->text,t,0)<=0)
     return(-1); 
   if((*target)==NULL)
     *target= t;
 }
 return(1);
}


int Lstring_append(entry,text,flag)
struct LstrinG *entry;
char *text;
int flag;
{
 struct LstrinG *target,*newby;

 if(entry!=NULL)
   for(target= entry; target->next!=NULL; target= target->next);
 if(Lstring_new(&newby,text,target,0)<=0)
   return(-1);
 return(1);
}


int Lstring_revert(start,flag)
struct LstrinG **start;
int flag;
{
 struct LstrinG *current,*last,*tr,*next;

 if((*start)==NULL)
   return(0);
 for(current= *start; current!=NULL; current= next) {
   next= current->next;
   last= current;
   tr= current->next;
   current->next= current->prev;
   current->prev= tr;
 }
 last->prev= (*start)->next;
 (*start)->next= NULL;
 *start= last;
 return(1);
}


